/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdarg.h>
#include "common.h"

uint32_t getInstructionSets()
{
#ifdef ARC_x86
	static uint32_t ret = 0xffffffff;

	if (ret == 0xffffffff)
	{
		uint32_t out1, out2, out3, out4, out5;

#ifdef _WIN32
		__asm
		{
			mov   eax,0x80000001
			cpuid
			mov   eax,1
			mov   out3,ecx
			mov   out4,edx
			cpuid
			mov   out1,ecx
			mov   out2,edx
		}
#else
		asm(
			"movl   $2147483649,%%eax\n\t"
			"cpuid\n\t"
			"movl   $1,%%eax\n\t"
			"movl   %%ecx,%2\n\t"
			"movl   %%edx,%3\n\t"
			"cpuid"
			: "=c"(out1), "=d"(out2), "=m"(out3), "=m"(out4) // output
			: // input
			: "eax", "ebx"); // used
#endif
		ret = 0;
		if (out2 & (1 << 23))
		{
			ret |= IS_MMX;
		}
		if (out2 & (1 << 25))
		{
			ret |= IS_SSE;
		}
		if (out2 & (1 << 26))
		{
			ret |= IS_SSE2;
		}
		if (out1 & (1 << 0))
		{
			ret |= IS_SSE3;
		}
		if (out1 & (1 << 9))
		{
			ret |= IS_SSSE3;
		}
		if (out1 & (1 << 19))
		{
			ret |= IS_SSE41;
		}
		if (out1 & (1 << 20))
		{
			ret |= IS_SSE42;
		}
		if (out3 & (1 << 6))
		{
			ret |= IS_SSE4A;
		}
		if (out1 & (1 << 28))
		{
			ret |= IS_AVX;
		}
		if ((ret & IS_AVX) && (out1 & (1 << 27))) // OSXSAVE (XGETBV)
		{
#ifdef _WIN32
			__asm
			{
				xor   ecx,ecx
				xgetbv
				mov   out5,eax
			}
#else
			asm(
				"xor   %%ecx,%%ecx\n\t"
				"xgetbv"
				: "=a"(out5) // output
				: // input
				: "ecx", "edx"); // used
#endif
			if ((out5 & 0x6) == 0x6)
			{
				ret |= IS_OS_YMM;
			}
		}
		if (ret & IS_OS_YMM)
		{
#ifdef _WIN32
			__asm
			{
				mov   eax,7
				xor   ecx,ecx
				cpuid
				mov   out5,ebx
			}
#else
			asm(
				"movl   $7,%%eax\n\t"
				"xor   %%ecx,%%ecx\n\t"
				"cpuid\n\t"
				: "=b"(out5) // output
				: // input
				: "eax", "ecx", "edx"); // used
#endif
			if (out5 & (1 << 5))
			{
				ret |= IS_AVX2;
			}
		}
		if (out3 & (1 << 11))
		{
			ret |= IS_XOP;
		}
		if (out1 & (1 << 23))
		{
			ret |= IS_POPCNT;
		}
#ifdef ARC_x86_32
#ifndef WIN32 // yeah I'm special still can't compile 64 bit Windows code
		if ((out2 & (1 << 30)) || (out4 & (1 << 29)) || (out3 & (1 << 6))) // x86_64, long_mode, long_mode_intel
		{
			reportWarning("Your CPU can run 64 bit which runs faster than the 32 bit code you are currently running.");
		}
#endif
#endif
	}
	return ret;
#else
	return 0;
#endif
}

void reportError(const char *msg, ...)
{
	va_list args;
	char *format;
	size_t len = strlen(msg);

	format = new char[len + 9]; // 9 == sizeof("Error: \n")
	strcpy(format, "Error: ");
	strcpy(format + 7, msg);
	format[len + 7] = '\n';
	format[len + 8] = 0;
	va_start(args, msg);
	vfprintf(stderr, format, args);
	va_end(args);
	delete [] format;
}

void reportWarning(const char *msg, ...)
{
	va_list args;
	char *format;
	size_t len = strlen(msg);

	format = new char[len + 11]; // 11 == sizeof("Warning: \n")
	strcpy(format, "Warning: ");
	strcpy(format + 9, msg);
	format[len + 9] = '\n';
	format[len + 10] = 0;
	va_start(args, msg);
	vfprintf(stderr, format, args);
	va_end(args);
	delete [] format;
}

float reciprocalRoundDown(uint64_t num)
{
	// keySpace = (36 ^ 8 - 36) / 35
	// 1.0f / keySpace = 14,303,679.752109742480909981844964 / 2 ^ 60 = 14,303,680 / 2 ^ 60
	// floor(14,303,680 / 2 ^ 60 * (keySpace - 1)) = floor(1.0000000173181169960123781947914) = 1
	// Very very bad since:
	// (keySpace - 1) - keySpace * 1 = -1 = 0xffffffffffffffff = 18,446,744,073,709,551,615 >= keySpace

	// round_down(1.0f / keySpace) =
	// round_down(14,303,679.752109742480909981844964 / 2 ^ 60) =
	// 14,303,679 / 2 ^ 60

	// keySpaceReciprocal = round_down(1.0f / m_keySpace)
	// Note that setting the rounding mode to round down may not work correctly
	// if the key space is > 2 ^ 54 well unless you can be sure that 

	// This may not work sometimes unless you change the [under|over]41bits functions to [under|over]42bits
	// I think the limit is under/over 2 ^ 40 but made it 2 ^ 41 just incase so then the limit of 2 ^ 41 would work
	// float keySpaceReciprocal = 1.0 / keySpace;
	// uint32_t tmp = (*((uint32_t*) &keySpaceReciprocal) & 0xff800000) - (23 << 23); // :)
	// keySpaceReciprocal -= *((float*) &tmp);

	// I think this should work:
	// #ifdef _WIN32
	//     #include <fenv.h>
	//     #pragma STDC FENV_ACCESS ON
	//     int tmp = fegetround();
	//     fesetround(FE_DOWNWARD);
	//     float keySpaceReciprocal = 1.0 / keySpace;
	//     fesetround(tmp);
	// #else
	//     unsigned int tmp = _control_fp(0, 0);
	//     _control_fp(_RC_DOWN, _MCW_RC);
	//     float keySpaceReciprocal = 1.0 / keySpace;
	//     _control_fp(tmp, _MCW_RC);
	// #endif

	// I totally guarantee this to find the keySpaceReciprocal I'm looking for
	uint64_t q, r;
	float reciprocal;

	if (num == 0)
	{
		//reciprocal = 1.0f / 0.0f;
		reciprocal = 0.0f;
	}
	else if (num == 1)
	{
		reciprocal = 1.0f;
	}
	else
	{
		q = 0xffffffffffffffff / num;
		r = 0xffffffffffffffff % num + 1;
		if (r == num)
		{
			q++;
			r = 0;
		}
		uint32_t power = countLeadingZeros(q) + 1;
		uint32_t highestBit = 65 - power;
		if (highestBit > 24)
		{
			q >>= highestBit - 24;
		}
		else
		{
			while (q <= 8388608)
			{
				q <<= 1;
				r <<= 1;
				if (r >= num)
				{
					q++;
					r -= num;
				}
			}
		}
		*((uint32_t*) &reciprocal) = (((uint32_t) q) & 0x007fffff) + ((127 - power) << 23);
	}
	return reciprocal;
}
