/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef COMMON_H
#define COMMON_H

#include "architecture.h"
#include <stdio.h>
#include <assert.h>

//#define AVX_BROADCAST
#define CHAIN_BUFFER_LENGTH 1048576

#ifdef _WIN32
	#pragma warning(disable:4996)
#endif

#ifndef __inline__
	#if defined(_MSC_VER) || defined(__INTEL_COMPILER)
		#define __inline__  __forceinline
	#elif defined(__GNUC__)
		#define __inline__  __inline __attribute__((__always_inline__))
	#else
		#define __inline__  inline
	#endif
#endif

#define QUOTE_(x) #x
#define QUOTE(x)  QUOTE_(x)

#define __STDC_CONSTANT_MACROS
#define __STDC_FORMAT_MACROS
#define __STDC_LIMIT_MACROS
#ifdef _WIN32
	#include "inttypes.h"
	#include <time.h>
	#include <windows.h>
	typedef LARGE_INTEGER TIMER_TYPE;
	#define TIMER_FUNC(t)             QueryPerformanceCounter(&t)
	__inline__ double TIMER_DIFF(LARGE_INTEGER s, LARGE_INTEGER e)
	{
		LARGE_INTEGER f;
		QueryPerformanceFrequency(&f);
		return ((double) (e.QuadPart - s.QuadPart)) / f.QuadPart;
	}
	#define DIRECTORY_SEPERATOR       "\\"
	#define DIRECTORY_SEPERATOR_CHAR  '\\'

	#define WIN_ALIGN_16 __declspec(align(16))
	#define ALIGN_16
	#define FSEEK64(file,off,origin)  _fseeki64(file,off,origin)
	#define GET_EXE_FULL_NAME(buf,bufSize)  GetModuleFileNameA(NULL, buf, bufSize)
#else
	#include <inttypes.h>
	#include <sys/time.h>
	#include <unistd.h>
	//#include <fcntl.h>
	typedef timeval TIMER_TYPE;
	#define TIMER_FUNC(t)             gettimeofday(&t, NULL)
	#define TIMER_DIFF(s,e)           ((e.tv_sec - s.tv_sec) + (e.tv_usec - s.tv_usec) / (double)1000000.0)
	#define DIRECTORY_SEPERATOR       "/"
	#define DIRECTORY_SEPERATOR_CHAR  '/'

	#ifndef PRIu64
		#define PRIu64 "llu"
	#endif
	#ifndef PRIx64
		#define PRIx64 "llx"
	#endif
	#ifndef UINT64_MAX
		#define UINT64_MAX 0xffffffffffffffffull
	#endif
	#ifndef UINT32_MAX
		#define UINT32_MAX 0xffffffff
	#endif

	#define WIN_ALIGN_16
	#define ALIGN_16 __attribute__((aligned(16)))
	#define FSEEK64(file,off,origin)  fseeko64(file,off,origin)
	#define GET_EXE_FULL_NAME(buf,bufSize)  readlink("/proc/self/exe", buf, bufSize);
#endif

#if defined(_MSC_VER)
	#include <intrin.h>
	__inline__ uint32_t countTrailingZeros(uint32_t val)
	{
		unsigned long out;
		_BitScanForward(&out, val);
		return (uint32_t) out;
	}
	__inline__ uint32_t countLeadingZeros(uint32_t val)
	{
		unsigned long out;
		_BitScanReverse(&out, val);
		return 31 - (uint32_t) out;
	}
	#ifdef ARC_x86_32
		__inline__ uint32_t countTrailingZeros(uint64_t val)
		{
			unsigned long out;
			if ((uint32_t) val == 0)
			{
				_BitScanForward(&out, (uint32_t) (val >> 32));
				return 32 + (uint32_t) out;
			}
			else
			_BitScanForward(&out, (uint32_t) val);
			return (uint32_t) out;
		}
		__inline__ uint32_t countLeadingZeros(uint64_t val)
		{
			unsigned long out;
			if ((uint32_t) (val >> 32) != 0)
			{
				_BitScanReverse(&out, (uint32_t) (val >> 32));
				return 63 - 32 - (uint32_t) out;
			}
			_BitScanReverse(&out, (uint32_t) val);
			return 63 - (uint32_t) out;
		}
	#else
		__inline__ uint32_t countTrailingZeros(uint64_t val)
		{
			unsigned long out;
			_BitScanForward64(&out, val);
			return (uint32_t) out;
		}
		__inline__ uint32_t countLeadingZeros(uint64_t val)
		{
			unsigned long out;
			_BitScanReverse64(&out, val);
			return 63 - (uint32_t) out;
		}
	#endif
#elif defined(__INTEL_COMPILER)
	__inline__ uint32_t countTrailingZeros(uint32_t val) { return (uint32_t) _bit_scan_forward(val); }
	__inline__ uint32_t countLeadingZeros (uint32_t val) { return (uint32_t) _bit_scan_reverse(val); }
	__inline__ uint32_t countTrailingZeros(uint64_t val)
	{
		if ((uint32_t) val == 0)
		{
			return (uint32_t) _bit_scan_forward((uint32_t) (val >> 32)) + 32;
		}
		return (uint32_t) _bit_scan_forward((uint32_t) val);
	}
	__inline__ uint32_t countLeadingZeros(uint64_t val)
	{
		if ((uint32_t) (val >> 32) != 0)
		{
			return 63 - 32 - (uint32_t) _bit_scan_reverse((uint32_t) (val >> 32));
		}
		return 63 - (uint32_t) _bit_scan_reverse((uint32_t) val);
	}
#elif defined(__GNUC__)
	__inline__ uint32_t countTrailingZeros(uint32_t val) { return (uint32_t) __builtin_ctz  (val); }
	__inline__ uint32_t countLeadingZeros (uint32_t val) { return (uint32_t) __builtin_clz  (val); }
	__inline__ uint32_t countTrailingZeros(uint64_t val) { return (uint32_t) __builtin_ctzll(val); }
	__inline__ uint32_t countLeadingZeros (uint64_t val) { return (uint32_t) __builtin_clzll(val); }
#else
	__inline__ uint32_t countTrailingZeros(uint32_t val)
	{
		uint32_t ret = 0;
		if ((val & 0xffff) == 0)
		{
			val >>= 16;
			ret = 16;
		}
		if ((val & 0xff) == 0)
		{
			val >>= 8;
			ret += 8;
		}
		if ((val & 0xf) == 0)
		{
			val >>= 4;
			ret += 4;
		}
		if ((val & 0x3) == 0)
		{
			val >>= 2;
			ret += 2;
		}
		if ((val & 0x1) == 0)
		{
			ret++;
		}
		return ret;
	}
	__inline__ uint32_t countLeadingZeros (uint32_t val)
	{
		uint32_t ret = 0;
		if ((val & 0xffff0000) == 0)
		{
			val <<= 16;
			ret = 16;
		}
		if ((val & 0xff000000) == 0)
		{
			val <<= 8;
			ret += 8;
		}
		if ((val & 0xf0000000) == 0)
		{
			val <<= 4;
			ret += 4;
		}
		if ((val & 0xc0000000) == 0)
		{
			val <<= 2;
			ret += 2;
		}
		if ((val & 0x80000000) == 0)
		{
			ret++;
		}
		return 63 - ret;
	}
	__inline__ uint32_t countTrailingZeros(uint64_t val)
	{
		uint32_t val32 = (uint32_t) val, ret = 0;
		if (val32 == 0)
		{
			val32 = (uint32_t) (val >> 32);
			ret = 32;
		}
		if ((val32 & 0xffff) == 0)
		{
			val32 >>= 16;
			ret += 16;
		}
		if ((val32 & 0xff) == 0)
		{
			val32 >>= 8;
			ret += 8;
		}
		if ((val32 & 0xf) == 0)
		{
			val32 >>= 4;
			ret += 4;
		}
		if ((val32 & 0x3) == 0)
		{
			val32 >>= 2;
			ret += 2;
		}
		if ((val32 & 0x1) == 0)
		{
			ret++;
		}
		return ret;
	}
	__inline__ uint32_t countLeadingZeros(uint64_t val)
	{
		uint32_t val32 = (uint32_t) (val >> 32), ret = 0;
		if (val32 == 0)
		{
			val32 = (uint32_t) val;
			ret = 32;
		}
		if ((val32 & 0xffff0000) == 0)
		{
			val32 <<= 16;
			ret += 16;
		}
		if ((val32 & 0xff000000) == 0)
		{
			val32 <<= 8;
			ret += 8;
		}
		if ((val32 & 0xf0000000) == 0)
		{
			val32 <<= 4;
			ret += 4;
		}
		if ((val32 & 0xc0000000) == 0)
		{
			val32 <<= 2;
			ret += 2;
		}
		if ((val32 & 0x80000000) == 0)
		{
			ret++;
		}
		return 63 - ret;
	}
#endif

#define SWAP_8_LO_BITS_BIT_ENDIAN_24(x)  ((((x) & 0x01) << 23) | (((x) & 0x02) << 22) | (((x) & 0x04) << 21) | (((x) & 0x08) << 20) | \
                                          (((x) & 0x10) << 19) | (((x) & 0x20) << 18) | (((x) & 0x40) << 17) | (((x) & 0x80) << 16))
#define SWAP_8_LO_BITS_BIT_ENDIAN_32(x)  ((((x) & 0x01) << 31) | (((x) & 0x02) << 30) | (((x) & 0x04) << 29) | (((x) & 0x08) << 28) | \
                                          (((x) & 0x10) << 27) | (((x) & 0x20) << 26) | (((x) & 0x40) << 25) | (((x) & 0x80) << 24))
#define SWAP_ENDIAN_32(x)                (((x) >> 24) | ((x) << 24) | (((x) >> 8) & 0xff00) | (((x) << 8) & 0xff0000))
#define SWAP_ENDIAN_16(x)                (((x) >> 8) | ((x) << 8))

#ifdef ARC_LITTLE_ENDIAN
	#define TO_BIG_ENDIAN_32(x)     SWAP_ENDIAN_32(x)
	#define TO_BIG_ENDIAN_16(x)     SWAP_ENDIAN_16(x)
	#define TO_LITTLE_ENDIAN_32(x)  (x)
	#define TO_LITTLE_ENDIAN_16(x)  (x)
#else
	#define TO_BIG_ENDIAN_32(x)     (x)
	#define TO_BIG_ENDIAN_16(x)     (x)
	#define TO_LITTLE_ENDIAN_32(x)  SWAP_ENDIAN_32(x)
	#define TO_LITTLE_ENDIAN_16(x)  SWAP_ENDIAN_16(x)
#endif
#define FROM_BIG_ENDIAN_32(x)     TO_BIG_ENDIAN_32(x)
#define FROM_BIG_ENDIAN_16(x)     TO_BIG_ENDIAN_16(x)
#define FROM_LITTLE_ENDIAN_32(x)  TO_LITTLE_ENDIAN_32(x)
#define FROM_LITTLE_ENDIAN_16(x)  TO_LITTLE_ENDIAN_16(x)

#ifdef ARC_x86
	#ifdef _WIN32
		#define MUL_32_RET_64(a32,b32,r64) \
			{ \
				__asm mov eax,a32 \
				__asm mul b32 \
				__asm mov [DWORD PTR r64],eax \
				__asm mov [DWORD PTR r64+4],edx \
			}

		#define DIV_MOD_64_RET_32_32(n64,d32,r32,q32) \
			{ \
				__asm mov eax,[DWORD PTR n64] \
				__asm mov edx,[DWORD PTR n64+4] \
				__asm div d32 \
				__asm mov q32,eax \
				__asm mov r32,edx \
			}
	#else
		#define MUL_32_RET_64(a32,b32,r64) \
			asm( \
				"mull %2" \
				: "=a"(((uint64u*)&r64)->lo32.u32), "=d"(((uint64u*)&r64)->hi32.u32) \
				: "m"(b32), "a"(a32) );

		#define DIV_MOD_64_RET_32_32(n64,d32,r32,q32) \
			asm( \
				"divl %2" \
				: "=a"(q32), "=d"(r32) \
				: "rm"(d32), "a"(((uint64u*)&n64)->lo32.u32), "d"(((uint64u*)&n64)->hi32.u32) );
	#endif
#else
	#define MUL_32_RET_64(a32,b32,r64) \
		r64 = (uint64_t) a32 * b32;

	#define DIV_MOD_64_RET_32_32(n64,d32,r32,q32) \
		r32 = n64 % d32; \
		q32 = n64 / d32;
#endif

#define pwr2pw(pwr,keySpace,keyBits)           (((pw) * (keySpace)) >> (keyBits))
#define pw2pwr(pw,keySpace,keyBits)  ((uint32_t) (((pw) << (keyBits)) / (keySpace)))

union uint16u
{
	uint16_t u16;
	uint8_t  u8[2];
	struct
	{
#ifdef ARC_LITTLE_ENDIAN
		uint8_t lo8, hi8;
#else
		uint8_t hi8, lo8;
#endif
	};
};

union uint32u
{
	uint32_t u32;
	uint16_t u16[2];
	uint8_t  u8[4];
	struct
	{
#ifdef ARC_LITTLE_ENDIAN
		uint16u lo16, hi16;
#else
		uint16u hi16, lo16;
#endif
	};
};

union uint64u
{
	uint64_t u64;
	uint32_t u32[2];
	uint16_t u16[4];
	uint8_t  u8[8];
	struct
	{
#ifdef ARC_LITTLE_ENDIAN
		uint32u lo32, hi32;
#else
		uint32u hi32, lo32;
#endif
	};
};

struct uint24
{
	uint8_t data[3];
};

struct vector4
{
	WIN_ALIGN_16
	uint32_t data[4] ALIGN_16;
};

struct rtChain
{
	uint64_t startPoint, endPoint;
};

#pragma pack(push)
#pragma pack(1)
struct crtChain
{
	uint64_t startPoint, endPoint;
	uint16_t checkPoints;
};
#pragma pack(pop)

enum instructionSets
{
	IS_MMX     = 0x0001,
	IS_SSE     = 0x0002,
	IS_SSE2    = 0x0004,
	IS_SSE3    = 0x0008,
	IS_SSSE3   = 0x0010, // pshufb xmmd,xmms
	IS_SSE41   = 0x0020, // ptest xmm1,xmm2
	IS_SSE42   = 0x0040,
	IS_SSE4A   = 0x0080,
	IS_AVX     = 0x0100,
	IS_AVX2    = 0x0200,
	IS_OS_YMM  = 0x0400,
	IS_XOP     = 0x0800, // vprotd xmmd,xmms,imm8; vpcmov xmmd,xmms1,xmms2,xmmc
	IS_POPCNT  = 0x1000, // popcnt r,r/m16; popcnt r,r/m32; popcnt r,r/m64
	IS_TZCNT   = 0x2000  // tzcnt  r,r/m16; tzcnt  r,r/m32; tzcnt  r,r/m64
};

enum characterEncoding
{
	CE_ASCII_BIN8     = 1,
	CE_UTF8           = 2,
	CE_UTF16_LE       = 4,
	CE_UTF16_BE       = 8,
	CE_UCS2_LE        = CE_UTF16_LE,
	CE_UCS2_BE        = CE_UTF16_BE,
	CE_UTF32_LE       = 16,
	CE_UTF32_BE       = 32,
	CE_FIX_WIDTH_CHAR = 64,
	CE_NULL_PADDED    = 128,
	CE_80_NULL_PADDED = 256,
	CE_CHARACTER_MASK = CE_ASCII_BIN8 | CE_UTF8 | CE_UTF16_LE | CE_UTF16_BE | CE_UCS2_LE | CE_UCS2_BE | CE_UTF32_LE | CE_UTF32_BE,
	CE_PADDING_MASK   = CE_NULL_PADDED | CE_80_NULL_PADDED
};

uint32_t getInstructionSets();
void reportError(const char *msg, ...);
void reportWarning(const char *msg, ...);
float reciprocalRoundDown(uint64_t num);

#endif
