/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "divsb.h"
#include "reduxfactory.h"
#include "hashfactory.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

uint32_t DivSB::s_inited = DivSB::initFactory();

uint32_t DivSB::initFactory()
{
	uint32_t ret;

	ret =  ReduxFactory::registerRedux("div", "Uses divide",
		CE_ASCII_BIN8 | CE_UTF8 | CE_FIX_WIDTH_CHAR | CE_80_NULL_PADDED | CE_NULL_PADDED, DivSB::create);
	ret += ReduxFactory::registerRedux("divh3", "Uses divide, hybrid3 or multi-masked",
		CE_ASCII_BIN8 | CE_UTF8 | CE_FIX_WIDTH_CHAR | CE_80_NULL_PADDED | CE_NULL_PADDED, DivSB::create);
	return ret;
}

DivSB::DivSB(const keySpace &ks, const char *hashName)
	: ReduxBruteForce(ks, hashName),
	ReduxSingleByte(ks, hashName),
	DivBase(ks, hashName)
{
}

Redux *DivSB::create(const keySpace &ks, const char *hashName)
{
	DivSB *ret;

	ret = new DivSB(ks, hashName);
	if (!ret->isInited())
	{
		delete ret;
		ret = NULL;
	}
	return ret;
}

void DivSB::doLinks(uint64_t *indexes, uint32_t chainPos, uint32_t tableIndex, uint32_t numLinks, uint32_t numIndexes)
{
	vector4 pw[MAX_PW_SIZE];
	vector4 hash[MAX_HASH_SIZE];
	uint32_t pwTmp[MAX_PW_BLOCKS];
	uint64_t keySpace                    = m_keySpace;
	uint64_t *keySpaceLut                = m_keySpaceLut;
	uint64_t *keySpaceBoundaries         = m_keySpaceBoundaries;
	uint8_t **subKeySpaces;
	uint8_t ***subKeySpacesSingleByte    = m_subKeySpacesSingleByte;
	uint32_t *charSetCacheLengths;
	uint32_t *charSetCacheLengthsFPM;
	uint32_t *passwordLengths            = m_passwordLengths;
	uint16_t **subKeySpacesCache;
	uint16_t ***subKeySpacesCacheArr     = m_subKeySpacesCache;
	int *numShortCutCharsArr             = m_numShortCutChars;
	uint32_t **charSetCacheLengthsArr    = m_charSetCacheLengths;
	uint32_t **charSetCacheLengthsFPMArr = m_charSetCacheLengthsFPM;
	uint32_t *divShortCut                = m_divShortCut;
	HASH_FUNC hashFunc                   = m_hashFunc;
	uint32_t keySpaceLutCount            = m_keySpaceLutCount;
	uint32_t numSubKeySpaces             = m_numSubKeySpaces;
	uint32_t charSetLength;
	int a;
	int passwordLength;
	int numShortCutChars;
	uint32_t pwBlocks                    = m_pwBlocks;
	uint32_t vecSize                     = 4 * m_vecSize;
	uint32_t padding                     = 0;
	uint32_t count                       = m_optimalNumIndexes;
	uint32_t startChainPos               = chainPos + (tableIndex << 16);

	numLinks += startChainPos;
	if ((m_characterEncoding & CE_80_NULL_PADDED) != 0)
	{
		padding = 0x80;
	}
	while (numIndexes != 0)
	{
		if (count > numIndexes)
		{
			count = numIndexes;
		}
		numIndexes -= count;
		for (chainPos = startChainPos; chainPos < numLinks; chainPos++)
		{
			// index to pw
			for (uint32_t i = 0; i < count; i++)
			{
				uint64_t index = indexes[i];
				uint32_t curSubKeySpace = 1;

				while (index < keySpaceBoundaries[curSubKeySpace])
				{
					curSubKeySpace++;
				}
				index                 -= keySpaceBoundaries[curSubKeySpace];
				curSubKeySpace         = numSubKeySpaces - curSubKeySpace;
				passwordLength         = (int) passwordLengths[curSubKeySpace];
				numShortCutChars       = numShortCutCharsArr[curSubKeySpace];
				subKeySpaces           = subKeySpacesSingleByte[curSubKeySpace];
				charSetCacheLengths    = charSetCacheLengthsArr[curSubKeySpace];
				charSetCacheLengthsFPM = charSetCacheLengthsFPMArr[curSubKeySpace];
				subKeySpacesCache      = subKeySpacesCacheArr[curSubKeySpace];
				for (a = passwordLength >> 2; a < (int) pwBlocks; a++)
				{
					pwTmp[a] = 0;
				}
				uint32_t quotient = (uint32_t) index;
				a = passwordLength - 2;
				if (numShortCutChars > 0)
				{
					uint32_t remainder, divisor;

					divisor = divShortCut[curSubKeySpace];
					DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

					numShortCutChars = passwordLength - numShortCutChars;
					for (; a >= numShortCutChars; a -= 2)
					{
						uint32_t r, q;

						charSetLength = charSetCacheLengths[a];
						q = (uint32_t) (((uint64_t) remainder * charSetCacheLengthsFPM[a]) >> 32);
						r = remainder - charSetLength * q;
						if (r >= charSetLength)
						{
							q++;
							r -= charSetLength;
						}
						remainder = q;
						*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
					}
					if (a == numShortCutChars - 1)
					{
						a--;
						((uint8_t*) pwTmp)[numShortCutChars] = subKeySpaces[numShortCutChars][remainder];
					}
				}
				for (; a >= 0; a -= 2)
				{
					uint32_t r, q;

					charSetLength = charSetCacheLengths[a];
					q = (uint32_t) (((uint64_t) quotient * charSetCacheLengthsFPM[a]) >> 32);
					r = quotient - charSetLength * q;
					if (r >= charSetLength)
					{
						q++;
						r -= charSetLength;
					}
					quotient = q;
					*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
				}
				if (a == -1)
				{
					((uint8_t*) pwTmp)[0] = subKeySpaces[0][quotient];
				}
				((uint8_t*) pwTmp)[passwordLength] = padding;

				((uint32_t*) pw)[i] = passwordLength;
				for (uint32_t b = 0, c = i + vecSize; b < pwBlocks; b++, c += vecSize)
				{
					((uint32_t*) pw)[c] = pwTmp[b];
				}
			}

			// pw to hash
			hashFunc(pw, hash);

			// Hash to index
			for (uint32_t i = 0; i < count; i++)
			{
				uint64_t tmp = (((uint64_t) hash->data[i + vecSize]) << 32) + hash->data[i] + chainPos;
				uint32_t shift = 56 - 14;

				tmp -= keySpaceLut[tmp >> 56];
				tmp -= keySpaceLut[(tmp >> (56 - 7)) + 256];
				for (uint32_t j = 2; j < keySpaceLutCount; j++)
				{
					tmp -= keySpaceLut[(uint32_t) (tmp >> shift) + 256 * j];
					shift -= 7;
				}
				if (tmp >= keySpace)
				{
					tmp -= keySpace;
				}
				indexes[i] = tmp;
			}
		}
		indexes += count;
	}
}

void DivSB::doPreWorkLinks(uint64_t *indexes, uint32_t firstChainPos, uint32_t tableIndex, uint32_t chainLength, uint32_t numIndexes, const uint32_t *hash2)
{
	//                    +++++++
	//                    ...
	//                    +++++++
	//       *       +    ++++++*
	//      **      ++    +++++**
	//     ***     +++    ++++***
	//    ****    ++++    +++****
	//   *****   +++++    ++*****
	//  ******  ++++++    +******
	// ******* +++++++    *******
	// ...         ...
	// ******* +++++++ vs *******
	// The first one is only a few percent slower in theory but is a
	// lot less code or less complex code so it's about the same speed.
	//
	// chain length    = 9999
	// vector width    = 12
	// cost            = 10000 * 9999 / 2      = 49,995,000
	// blocks          = 9999 / 12             = 833.25
	// waste per block = 12 * 11 / 2           = 66
	// waste           = 833.25 * 66           = 54,994.5
	// percent waste   = 54,994.5 / 49,995,000 = 0.11%
	//
	// chain length    = 9999
	// vector width    = 256  (AVX LM)
	// cost            = 10000 * 9999 / 2         = 49,995,000
	// blocks          = 9999 / 256               = 39.05859375
	// waste per block = 256 * 255 / 2            = 32,640
	// waste           = 39.05859375 * 32,640     = 1,274,872.5
	// percent waste   = 1,274,872.5 / 49,995,000 = 2.55%

	vector4 pw[MAX_PW_SIZE];
	vector4 hash[MAX_HASH_SIZE];
	uint32_t pwTmp[MAX_PW_BLOCKS];
	uint64_t keySpace                    = m_keySpace;
	uint64_t firstIndex;
	uint64_t *keySpaceLut                = m_keySpaceLut;
	uint64_t *keySpaceBoundaries         = m_keySpaceBoundaries;
	uint8_t **subKeySpaces;
	uint8_t ***subKeySpacesSingleByte    = m_subKeySpacesSingleByte;
	uint32_t *charSetCacheLengths;
	uint32_t *charSetCacheLengthsFPM;
	uint32_t *passwordLengths            = m_passwordLengths;
	uint16_t **subKeySpacesCache;
	uint16_t ***subKeySpacesCacheArr     = m_subKeySpacesCache;
	int *numShortCutCharsArr             = m_numShortCutChars;
	uint32_t **charSetCacheLengthsArr    = m_charSetCacheLengths;
	uint32_t **charSetCacheLengthsFPMArr = m_charSetCacheLengthsFPM;
	uint32_t *divShortCut                = m_divShortCut;
	HASH_FUNC hashFunc                   = m_hashFunc;
	uint32_t keySpaceLutCount            = m_keySpaceLutCount;
	uint32_t numSubKeySpaces             = m_numSubKeySpaces;
	uint32_t charSetLength;
	int a;
	int passwordLength;
	int numShortCutChars;
	uint32_t pwBlocks                    = m_pwBlocks;
	uint32_t vecSize                     = 4 * m_vecSize;
	uint32_t padding                     = 0;
	uint32_t count                       = m_optimalNumIndexes;
	uint32_t curCount;

	firstChainPos += tableIndex << 16;
	chainLength   += tableIndex << 16;
	firstIndex = *((uint64_t*) hash2) + firstChainPos;
	firstIndex -= keySpaceLut[ firstIndex >>  56];
	firstIndex -= keySpaceLut[(firstIndex >> (56 - 7)) + 256];
	for (uint32_t j = 2, shift = 56 - 7; j < keySpaceLutCount; j++)
	{
		shift -= 7;
		firstIndex -= keySpaceLut[(uint32_t) (firstIndex >> shift) + 256 * j];
	}
	if (firstIndex >= keySpace)
	{
		firstIndex -= keySpace;
	}
	if ((m_characterEncoding & CE_80_NULL_PADDED) != 0)
	{
		padding = 0x80;
	}
	while (numIndexes != 0)
	{
		if (count > numIndexes)
		{
			count = numIndexes;
		}
		numIndexes -= count;
		for (uint32_t i = 0; i < count; i++)
		{
			indexes[i] = firstIndex;
			firstIndex++;
			if (firstIndex >= keySpace)
			{
				firstIndex = 0;
			}
		}
		curCount = 1;
		for (uint32_t chainPos = firstChainPos; chainPos < chainLength; chainPos++)
		{
			// index to pw
			for (uint32_t i = 0; i < curCount; i++)
			{
				uint64_t index = indexes[i];
				uint32_t curSubKeySpace = 1;

				while (index < keySpaceBoundaries[curSubKeySpace])
				{
					curSubKeySpace++;
				}
				index                 -= keySpaceBoundaries[curSubKeySpace];
				curSubKeySpace         = numSubKeySpaces - curSubKeySpace;
				passwordLength         = (int) passwordLengths[curSubKeySpace];
				numShortCutChars       = numShortCutCharsArr[curSubKeySpace];
				subKeySpaces           = subKeySpacesSingleByte[curSubKeySpace];
				charSetCacheLengths    = charSetCacheLengthsArr[curSubKeySpace];
				charSetCacheLengthsFPM = charSetCacheLengthsFPMArr[curSubKeySpace];
				subKeySpacesCache      = subKeySpacesCacheArr[curSubKeySpace];
				for (a = passwordLength >> 2; a < (int) pwBlocks; a++)
				{
					pwTmp[a] = 0;
				}
				uint32_t quotient = (uint32_t) index;
				a = passwordLength - 2;
				if (numShortCutChars > 0)
				{
					uint32_t remainder, divisor;

					divisor = divShortCut[curSubKeySpace];
					DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

					numShortCutChars = passwordLength - numShortCutChars;
					for (; a >= numShortCutChars; a -= 2)
					{
						uint32_t r, q;

						charSetLength = charSetCacheLengths[a];
						q = (uint32_t) (((uint64_t) remainder * charSetCacheLengthsFPM[a]) >> 32);
						r = remainder - charSetLength * q;
						if (r >= charSetLength)
						{
							q++;
							r -= charSetLength;
						}
						remainder = q;
						*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
					}
					if (a == numShortCutChars - 1)
					{
						a--;
						((uint8_t*) pwTmp)[numShortCutChars] = subKeySpaces[numShortCutChars][remainder];
					}
				}
				for (; a >= 0; a -= 2)
				{
					uint32_t r, q;

					charSetLength = charSetCacheLengths[a];
					q = (uint32_t) (((uint64_t) quotient * charSetCacheLengthsFPM[a]) >> 32);
					r = quotient - charSetLength * q;
					if (r >= charSetLength)
					{
						q++;
						r -= charSetLength;
					}
					quotient = q;
					*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
				}
				if (a == -1)
				{
					((uint8_t*) pwTmp)[0] = subKeySpaces[0][quotient];
				}
				((uint8_t*) pwTmp)[passwordLength] = padding;

				((uint32_t*) pw)[i] = passwordLength;
				for (uint32_t b = 0, c = i + vecSize; b < pwBlocks; b++, c += vecSize)
				{
					((uint32_t*) pw)[c] = pwTmp[b];
				}
			}

			// pw to hash
			hashFunc(pw, hash);

			// Hash to index
			for (uint32_t i = 0; i < count; i++)
			{
				uint64_t index = *((uint64_t*) hash) + chainPos;
				uint32_t shift = 56 - 14;

				index -= keySpaceLut[ index >>  56];
				index -= keySpaceLut[(index >> (56 - 7)) + 256];
				for (uint32_t j = 2; j < keySpaceLutCount; j++)
				{
					index -= keySpaceLut[(uint32_t) (index >> shift) + 256 * j];
					shift -= 7;
				}
				if (index >= keySpace)
				{
					index -= keySpace;
				}
				indexes[i] = index;
			}
			if (curCount < count)
			{
				curCount++;
			}
		}
		firstChainPos += count;
		indexes += count;
	}
}

uint32_t DivSB::falseAlarmChecking(uint64_t *indexes, const uint32_t *hash2, const uint32_t *lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	vector4 pw[MAX_PW_SIZE];
	vector4 hash[MAX_HASH_SIZE];
	uint32_t pwTmp[MAX_PW_BLOCKS];
	uint64_t keySpace                    = m_keySpace;
	uint64_t *curIndexes                 = indexes;
	uint64_t *keySpaceLut                = m_keySpaceLut;
	uint64_t *keySpaceBoundaries         = m_keySpaceBoundaries;
	uint8_t **subKeySpaces;
	uint8_t ***subKeySpacesSingleByte    = m_subKeySpacesSingleByte;
	uint32_t *charSetCacheLengths;
	uint32_t *charSetCacheLengthsFPM;
	uint32_t *passwordLengths            = m_passwordLengths;
	uint16_t **subKeySpacesCache;
	uint16_t ***subKeySpacesCacheArr     = m_subKeySpacesCache;
	int *numShortCutCharsArr             = m_numShortCutChars;
	uint32_t **charSetCacheLengthsArr    = m_charSetCacheLengths;
	uint32_t **charSetCacheLengthsFPMArr = m_charSetCacheLengthsFPM;
	uint32_t *divShortCut                = m_divShortCut;
	HASH_FUNC hashFunc                   = m_hashFunc;
	uint32_t keySpaceLutCount            = m_keySpaceLutCount;
	uint32_t numSubKeySpaces             = m_numSubKeySpaces;
	uint32_t charSetLength;
	int a;
	int passwordLength;
	int numShortCutChars;
	uint32_t pwBlocks                    = m_pwBlocks;
	uint32_t vecSize                     = 4 * m_vecSize;
	uint32_t hashBlocks                  = (m_hashLen + 3) / 4;
	uint32_t padding                     = 0;
	uint32_t count                       = m_optimalNumIndexes;
	uint32_t curCount;
	uint32_t lastChainPos;

	tableIndex <<= 16;
	if ((m_characterEncoding & CE_80_NULL_PADDED) != 0)
	{
		padding = 0x80;
	}
	while (numIndexes != 0)
	{
		if (count > numIndexes)
		{
			count = numIndexes;
		}
		numIndexes -= count;
		curCount = count;
		lastChainPos = lastChainPoses[curCount - 1] + tableIndex;
		for (uint32_t chainPos = tableIndex; curCount != 0; chainPos++)
		{
			// index to pw
			for (uint32_t i = 0; i < curCount; i++)
			{
				uint64_t index = curIndexes[i];
				uint32_t curSubKeySpace = 1;

				while (index < keySpaceBoundaries[curSubKeySpace])
				{
					curSubKeySpace++;
				}
				index                 -= keySpaceBoundaries[curSubKeySpace];
				curSubKeySpace         = numSubKeySpaces - curSubKeySpace;
				passwordLength         = (int) passwordLengths[curSubKeySpace];
				numShortCutChars       = numShortCutCharsArr[curSubKeySpace];
				subKeySpaces           = subKeySpacesSingleByte[curSubKeySpace];
				charSetCacheLengths    = charSetCacheLengthsArr[curSubKeySpace];
				charSetCacheLengthsFPM = charSetCacheLengthsFPMArr[curSubKeySpace];
				subKeySpacesCache      = subKeySpacesCacheArr[curSubKeySpace];
				for (a = passwordLength >> 2; a < (int) pwBlocks; a++)
				{
					pwTmp[a] = 0;
				}
				uint32_t quotient = (uint32_t) index;
				a = passwordLength - 2;
				if (numShortCutChars > 0)
				{
					uint32_t remainder, divisor;

					divisor = divShortCut[curSubKeySpace];
					DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

					numShortCutChars = passwordLength - numShortCutChars;
					for (; a >= numShortCutChars; a -= 2)
					{
						uint32_t r, q;

						charSetLength = charSetCacheLengths[a];
						q = (uint32_t) (((uint64_t) remainder * charSetCacheLengthsFPM[a]) >> 32);
						r = remainder - charSetLength * q;
						if (r >= charSetLength)
						{
							q++;
							r -= charSetLength;
						}
						remainder = q;
						*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
					}
					if (a == numShortCutChars - 1)
					{
						a--;
						((uint8_t*) pwTmp)[numShortCutChars] = subKeySpaces[numShortCutChars][remainder];
					}
				}
				for (; a >= 0; a -= 2)
				{
					uint32_t r, q;

					charSetLength = charSetCacheLengths[a];
					q = (uint32_t) (((uint64_t) quotient * charSetCacheLengthsFPM[a]) >> 32);
					r = quotient - charSetLength * q;
					if (r >= charSetLength)
					{
						q++;
						r -= charSetLength;
					}
					quotient = q;
					*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
				}
				if (a == -1)
				{
					((uint8_t*) pwTmp)[0] = subKeySpaces[0][quotient];
				}
				((uint8_t*) pwTmp)[passwordLength] = padding;

				((uint32_t*) pw)[i] = passwordLength;
				for (uint32_t b = 0, c = i + vecSize; b < pwBlocks; b++, c += vecSize)
				{
					((uint32_t*) pw)[c] = pwTmp[b];
				}
			}

			// pw to hash
			hashFunc(pw, hash);

			// is end of chain
			if (lastChainPos == chainPos)
			{
				do
				{
					uint32_t j = 0;
					for (uint32_t i = curCount - 1; j < hashBlocks; i += vecSize, j++)
					{
						if (hash2[j] != ((uint32_t*) hash)[i])
						{
							break;
						}
					}
					if (j == hashBlocks)
					{
						return ((uint32_t) (curIndexes - indexes)) + curCount;
					}
					curCount--;
				} while (curCount != 0 && lastChainPoses[curCount - 1] + tableIndex == chainPos);
				if (curCount != 0)
				{
					assert(lastChainPoses[curCount - 1] + tableIndex < chainPos);
					lastChainPos = lastChainPoses[curCount - 1] + tableIndex;
				}
			}

			// hash to index
			for (uint32_t i = 0; i < curCount; i++)
			{
				uint64_t index = *((uint64_t*) hash) + chainPos;
				uint32_t shift = 56 - 14;

				index -= keySpaceLut[ index >>  56];
				index -= keySpaceLut[(index >> (56 - 7)) + 256];
				for (uint32_t j = 2; j < keySpaceLutCount; j++)
				{
					index -= keySpaceLut[(uint32_t) (index >> shift) + 256 * j];
					shift -= 7;
				}
				if (index >= keySpace)
				{
					index -= keySpace;
				}
				curIndexes[i] = index;
			}
		}
		curIndexes += count;
		lastChainPoses += count;
	}
	return 0;
}
