/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cuda.h>
#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#ifndef uint8_t
	#define uint8_t unsigned char
#endif
#ifndef uint16_t
	#define uint16_t unsigned short
#endif
#ifndef uint32_t
	#define uint32_t unsigned int
#endif
#ifndef uint64_t
	#define uint64_t unsigned long long int
#endif

typedef void         (*kernel_doLinks)           (int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes);
typedef void         (*kernel_doPreWorkLinks)    (int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first);
typedef unsigned int (*kernel_falseAlarmChecking)(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes);

__constant__ uint64_t c_keySpace;
__constant__ uint64_t c_keySpaceFirst;
__constant__ uint64_t c_keySpaceBoundaries[20];
__constant__ uint64_t c_keySpaceBoundariesMax;
__constant__ float    c_keySpaceReciprocal;
__constant__ float    c_keySpaceReciprocalFirst;
__constant__ uint32_t c_numSubKeySpaceCharSets;
__constant__ uint32_t c_numSubKeySpaces;
__constant__ uint32_t c_charSetLength;

__constant__ float    c_charSetReciprocal;
__constant__ float    c_charSetLen;
__constant__ float    c_ncharSetLen;

__device__ __forceinline__ void md5(uint32_t pw0, uint32_t pw1,                                           uint32_t passwordLength, uint32_t &hash0, uint32_t &hash1)
{
	#define MD5_FF(a, b, c, d, C, x, s) a += (d ^ (b & (c ^ d))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_GG(a, b, c, d, C, x, s) a += (c ^ (d & (b ^ c))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_HH(a, b, c, d, C, x, s) a += (b ^ c ^ d)         + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_II(a, b, c, d, C, x, s) a += (c ^ (b | ~d))      + C + x; a = (a << s) | (a >> (32 - s)); a += b;

	#define x0   pw0
	#define x1   pw1
	#define x2   0
	#define x3   0
	#define x4   0
	#define x5   0
	#define x6   0
	#define x7   0
	#define x8   0
	#define x9   0
	#define x10  0
	#define x11  0
	#define x12  0
	#define x13  0
	#define x14  passwordLength
	#define x15  0

	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476;

	MD5_FF(a, b, c, d, 0xd76aa478,  x0,  7); MD5_FF(d, a, b, c, 0xe8c7b756,  x1, 12); MD5_FF(c, d, a, b, 0x242070db,  x2, 17); MD5_FF(b, c, d, a, 0xc1bdceee,  x3, 22);
	MD5_FF(a, b, c, d, 0xf57c0faf,  x4,  7); MD5_FF(d, a, b, c, 0x4787c62a,  x5, 12); MD5_FF(c, d, a, b, 0xa8304613,  x6, 17); MD5_FF(b, c, d, a, 0xfd469501,  x7, 22);
	MD5_FF(a, b, c, d, 0x698098d8,  x8,  7); MD5_FF(d, a, b, c, 0x8b44f7af,  x9, 12); MD5_FF(c, d, a, b, 0xffff5bb1, x10, 17); MD5_FF(b, c, d, a, 0x895cd7be, x11, 22);
	MD5_FF(a, b, c, d, 0x6b901122, x12,  7); MD5_FF(d, a, b, c, 0xfd987193, x13, 12); MD5_FF(c, d, a, b, 0xa679438e, x14, 17); MD5_FF(b, c, d, a, 0x49b40821, x15, 22);

	MD5_GG(a, b, c, d, 0xf61e2562,  x1,  5); MD5_GG(d, a, b, c, 0xc040b340,  x6,  9); MD5_GG(c, d, a, b, 0x265e5a51, x11, 14); MD5_GG(b, c, d, a, 0xe9b6c7aa,  x0, 20);
	MD5_GG(a, b, c, d, 0xd62f105d,  x5,  5); MD5_GG(d, a, b, c, 0x02441453, x10,  9); MD5_GG(c, d, a, b, 0xd8a1e681, x15, 14); MD5_GG(b, c, d, a, 0xe7d3fbc8,  x4, 20);
	MD5_GG(a, b, c, d, 0x21e1cde6,  x9,  5); MD5_GG(d, a, b, c, 0xc33707d6, x14,  9); MD5_GG(c, d, a, b, 0xf4d50d87,  x3, 14); MD5_GG(b, c, d, a, 0x455a14ed,  x8, 20);
	MD5_GG(a, b, c, d, 0xa9e3e905, x13,  5); MD5_GG(d, a, b, c, 0xfcefa3f8,  x2,  9); MD5_GG(c, d, a, b, 0x676f02d9,  x7, 14); MD5_GG(b, c, d, a, 0x8d2a4c8a, x12, 20);

	MD5_HH(a, b, c, d, 0xfffa3942,  x5,  4); MD5_HH(d, a, b, c, 0x8771f681,  x8, 11); MD5_HH(c, d, a, b, 0x6d9d6122, x11, 16); MD5_HH(b, c, d, a, 0xfde5380c, x14, 23);
	MD5_HH(a, b, c, d, 0xa4beea44,  x1,  4); MD5_HH(d, a, b, c, 0x4bdecfa9,  x4, 11); MD5_HH(c, d, a, b, 0xf6bb4b60,  x7, 16); MD5_HH(b, c, d, a, 0xbebfbc70, x10, 23);
	MD5_HH(a, b, c, d, 0x289b7ec6, x13,  4); MD5_HH(d, a, b, c, 0xeaa127fa,  x0, 11); MD5_HH(c, d, a, b, 0xd4ef3085,  x3, 16); MD5_HH(b, c, d, a, 0x04881d05,  x6, 23);
	MD5_HH(a, b, c, d, 0xd9d4d039,  x9,  4); MD5_HH(d, a, b, c, 0xe6db99e5, x12, 11); MD5_HH(c, d, a, b, 0x1fa27cf8, x15, 16); MD5_HH(b, c, d, a, 0xc4ac5665,  x2, 23);

	MD5_II(a, b, c, d, 0xf4292244,  x0,  6); MD5_II(d, a, b, c, 0x432aff97,  x7, 10); MD5_II(c, d, a, b, 0xab9423a7, x14, 15); MD5_II(b, c, d, a, 0xfc93a039,  x5, 21);
	MD5_II(a, b, c, d, 0x655b59c3, x12,  6); MD5_II(d, a, b, c, 0x8f0ccc92,  x3, 10); MD5_II(c, d, a, b, 0xffeff47d, x10, 15); MD5_II(b, c, d, a, 0x85845dd1,  x1, 21);
	MD5_II(a, b, c, d, 0x6fa87e4f,  x8,  6); MD5_II(d, a, b, c, 0xfe2ce6e0, x15, 10); MD5_II(c, d, a, b, 0xa3014314,  x6, 15); MD5_II(b, c, d, a, 0x4e0811a1, x13, 21);
	MD5_II(a, b, c, d, 0xf7537e82,  x4,  6); MD5_II(d, a, b, c, 0xbd3af235, x11, 10); MD5_II(c, d, a, b, 0x2ad7d2bb,  x2, 15); MD5_II(b, c, d, a, 0xeb86d391,  x9, 21);

	hash0 = a + 0x67452301;
	hash1 = b + 0xefcdab89;

	#undef MD5_FF
	#undef MD5_GG
	#undef MD5_HH
	#undef MD5_II

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15
}

__device__ __forceinline__ void md5(uint32_t pw0, uint32_t pw1, uint32_t pw2,                             uint32_t passwordLength, uint32_t &hash0, uint32_t &hash1)
{
	#define MD5_FF(a, b, c, d, C, x, s) a += (d ^ (b & (c ^ d))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_GG(a, b, c, d, C, x, s) a += (c ^ (d & (b ^ c))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_HH(a, b, c, d, C, x, s) a += (b ^ c ^ d)         + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_II(a, b, c, d, C, x, s) a += (c ^ (b | ~d))      + C + x; a = (a << s) | (a >> (32 - s)); a += b;

	#define x0   pw0
	#define x1   pw1
	#define x2   pw2
	#define x3   0
	#define x4   0
	#define x5   0
	#define x6   0
	#define x7   0
	#define x8   0
	#define x9   0
	#define x10  0
	#define x11  0
	#define x12  0
	#define x13  0
	#define x14  passwordLength
	#define x15  0

	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476;

	MD5_FF(a, b, c, d, 0xd76aa478,  x0,  7); MD5_FF(d, a, b, c, 0xe8c7b756,  x1, 12); MD5_FF(c, d, a, b, 0x242070db,  x2, 17); MD5_FF(b, c, d, a, 0xc1bdceee,  x3, 22);
	MD5_FF(a, b, c, d, 0xf57c0faf,  x4,  7); MD5_FF(d, a, b, c, 0x4787c62a,  x5, 12); MD5_FF(c, d, a, b, 0xa8304613,  x6, 17); MD5_FF(b, c, d, a, 0xfd469501,  x7, 22);
	MD5_FF(a, b, c, d, 0x698098d8,  x8,  7); MD5_FF(d, a, b, c, 0x8b44f7af,  x9, 12); MD5_FF(c, d, a, b, 0xffff5bb1, x10, 17); MD5_FF(b, c, d, a, 0x895cd7be, x11, 22);
	MD5_FF(a, b, c, d, 0x6b901122, x12,  7); MD5_FF(d, a, b, c, 0xfd987193, x13, 12); MD5_FF(c, d, a, b, 0xa679438e, x14, 17); MD5_FF(b, c, d, a, 0x49b40821, x15, 22);

	MD5_GG(a, b, c, d, 0xf61e2562,  x1,  5); MD5_GG(d, a, b, c, 0xc040b340,  x6,  9); MD5_GG(c, d, a, b, 0x265e5a51, x11, 14); MD5_GG(b, c, d, a, 0xe9b6c7aa,  x0, 20);
	MD5_GG(a, b, c, d, 0xd62f105d,  x5,  5); MD5_GG(d, a, b, c, 0x02441453, x10,  9); MD5_GG(c, d, a, b, 0xd8a1e681, x15, 14); MD5_GG(b, c, d, a, 0xe7d3fbc8,  x4, 20);
	MD5_GG(a, b, c, d, 0x21e1cde6,  x9,  5); MD5_GG(d, a, b, c, 0xc33707d6, x14,  9); MD5_GG(c, d, a, b, 0xf4d50d87,  x3, 14); MD5_GG(b, c, d, a, 0x455a14ed,  x8, 20);
	MD5_GG(a, b, c, d, 0xa9e3e905, x13,  5); MD5_GG(d, a, b, c, 0xfcefa3f8,  x2,  9); MD5_GG(c, d, a, b, 0x676f02d9,  x7, 14); MD5_GG(b, c, d, a, 0x8d2a4c8a, x12, 20);

	MD5_HH(a, b, c, d, 0xfffa3942,  x5,  4); MD5_HH(d, a, b, c, 0x8771f681,  x8, 11); MD5_HH(c, d, a, b, 0x6d9d6122, x11, 16); MD5_HH(b, c, d, a, 0xfde5380c, x14, 23);
	MD5_HH(a, b, c, d, 0xa4beea44,  x1,  4); MD5_HH(d, a, b, c, 0x4bdecfa9,  x4, 11); MD5_HH(c, d, a, b, 0xf6bb4b60,  x7, 16); MD5_HH(b, c, d, a, 0xbebfbc70, x10, 23);
	MD5_HH(a, b, c, d, 0x289b7ec6, x13,  4); MD5_HH(d, a, b, c, 0xeaa127fa,  x0, 11); MD5_HH(c, d, a, b, 0xd4ef3085,  x3, 16); MD5_HH(b, c, d, a, 0x04881d05,  x6, 23);
	MD5_HH(a, b, c, d, 0xd9d4d039,  x9,  4); MD5_HH(d, a, b, c, 0xe6db99e5, x12, 11); MD5_HH(c, d, a, b, 0x1fa27cf8, x15, 16); MD5_HH(b, c, d, a, 0xc4ac5665,  x2, 23);

	MD5_II(a, b, c, d, 0xf4292244,  x0,  6); MD5_II(d, a, b, c, 0x432aff97,  x7, 10); MD5_II(c, d, a, b, 0xab9423a7, x14, 15); MD5_II(b, c, d, a, 0xfc93a039,  x5, 21);
	MD5_II(a, b, c, d, 0x655b59c3, x12,  6); MD5_II(d, a, b, c, 0x8f0ccc92,  x3, 10); MD5_II(c, d, a, b, 0xffeff47d, x10, 15); MD5_II(b, c, d, a, 0x85845dd1,  x1, 21);
	MD5_II(a, b, c, d, 0x6fa87e4f,  x8,  6); MD5_II(d, a, b, c, 0xfe2ce6e0, x15, 10); MD5_II(c, d, a, b, 0xa3014314,  x6, 15); MD5_II(b, c, d, a, 0x4e0811a1, x13, 21);
	MD5_II(a, b, c, d, 0xf7537e82,  x4,  6); MD5_II(d, a, b, c, 0xbd3af235, x11, 10); MD5_II(c, d, a, b, 0x2ad7d2bb,  x2, 15); MD5_II(b, c, d, a, 0xeb86d391,  x9, 21);

	hash0 = a + 0x67452301;
	hash1 = b + 0xefcdab89;

	#undef MD5_FF
	#undef MD5_GG
	#undef MD5_HH
	#undef MD5_II

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15
}

__device__ __forceinline__ void md5(uint32_t pw0, uint32_t pw1, uint32_t pw2, uint32_t pw3,               uint32_t passwordLength, uint32_t &hash0, uint32_t &hash1)
{
	#define MD5_FF(a, b, c, d, C, x, s) a += (d ^ (b & (c ^ d))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_GG(a, b, c, d, C, x, s) a += (c ^ (d & (b ^ c))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_HH(a, b, c, d, C, x, s) a += (b ^ c ^ d)         + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_II(a, b, c, d, C, x, s) a += (c ^ (b | ~d))      + C + x; a = (a << s) | (a >> (32 - s)); a += b;

	#define x0   pw0
	#define x1   pw1
	#define x2   pw2
	#define x3   pw3
	#define x4   0
	#define x5   0
	#define x6   0
	#define x7   0
	#define x8   0
	#define x9   0
	#define x10  0
	#define x11  0
	#define x12  0
	#define x13  0
	#define x14  passwordLength
	#define x15  0

	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476;

	MD5_FF(a, b, c, d, 0xd76aa478,  x0,  7); MD5_FF(d, a, b, c, 0xe8c7b756,  x1, 12); MD5_FF(c, d, a, b, 0x242070db,  x2, 17); MD5_FF(b, c, d, a, 0xc1bdceee,  x3, 22);
	MD5_FF(a, b, c, d, 0xf57c0faf,  x4,  7); MD5_FF(d, a, b, c, 0x4787c62a,  x5, 12); MD5_FF(c, d, a, b, 0xa8304613,  x6, 17); MD5_FF(b, c, d, a, 0xfd469501,  x7, 22);
	MD5_FF(a, b, c, d, 0x698098d8,  x8,  7); MD5_FF(d, a, b, c, 0x8b44f7af,  x9, 12); MD5_FF(c, d, a, b, 0xffff5bb1, x10, 17); MD5_FF(b, c, d, a, 0x895cd7be, x11, 22);
	MD5_FF(a, b, c, d, 0x6b901122, x12,  7); MD5_FF(d, a, b, c, 0xfd987193, x13, 12); MD5_FF(c, d, a, b, 0xa679438e, x14, 17); MD5_FF(b, c, d, a, 0x49b40821, x15, 22);

	MD5_GG(a, b, c, d, 0xf61e2562,  x1,  5); MD5_GG(d, a, b, c, 0xc040b340,  x6,  9); MD5_GG(c, d, a, b, 0x265e5a51, x11, 14); MD5_GG(b, c, d, a, 0xe9b6c7aa,  x0, 20);
	MD5_GG(a, b, c, d, 0xd62f105d,  x5,  5); MD5_GG(d, a, b, c, 0x02441453, x10,  9); MD5_GG(c, d, a, b, 0xd8a1e681, x15, 14); MD5_GG(b, c, d, a, 0xe7d3fbc8,  x4, 20);
	MD5_GG(a, b, c, d, 0x21e1cde6,  x9,  5); MD5_GG(d, a, b, c, 0xc33707d6, x14,  9); MD5_GG(c, d, a, b, 0xf4d50d87,  x3, 14); MD5_GG(b, c, d, a, 0x455a14ed,  x8, 20);
	MD5_GG(a, b, c, d, 0xa9e3e905, x13,  5); MD5_GG(d, a, b, c, 0xfcefa3f8,  x2,  9); MD5_GG(c, d, a, b, 0x676f02d9,  x7, 14); MD5_GG(b, c, d, a, 0x8d2a4c8a, x12, 20);

	MD5_HH(a, b, c, d, 0xfffa3942,  x5,  4); MD5_HH(d, a, b, c, 0x8771f681,  x8, 11); MD5_HH(c, d, a, b, 0x6d9d6122, x11, 16); MD5_HH(b, c, d, a, 0xfde5380c, x14, 23);
	MD5_HH(a, b, c, d, 0xa4beea44,  x1,  4); MD5_HH(d, a, b, c, 0x4bdecfa9,  x4, 11); MD5_HH(c, d, a, b, 0xf6bb4b60,  x7, 16); MD5_HH(b, c, d, a, 0xbebfbc70, x10, 23);
	MD5_HH(a, b, c, d, 0x289b7ec6, x13,  4); MD5_HH(d, a, b, c, 0xeaa127fa,  x0, 11); MD5_HH(c, d, a, b, 0xd4ef3085,  x3, 16); MD5_HH(b, c, d, a, 0x04881d05,  x6, 23);
	MD5_HH(a, b, c, d, 0xd9d4d039,  x9,  4); MD5_HH(d, a, b, c, 0xe6db99e5, x12, 11); MD5_HH(c, d, a, b, 0x1fa27cf8, x15, 16); MD5_HH(b, c, d, a, 0xc4ac5665,  x2, 23);

	MD5_II(a, b, c, d, 0xf4292244,  x0,  6); MD5_II(d, a, b, c, 0x432aff97,  x7, 10); MD5_II(c, d, a, b, 0xab9423a7, x14, 15); MD5_II(b, c, d, a, 0xfc93a039,  x5, 21);
	MD5_II(a, b, c, d, 0x655b59c3, x12,  6); MD5_II(d, a, b, c, 0x8f0ccc92,  x3, 10); MD5_II(c, d, a, b, 0xffeff47d, x10, 15); MD5_II(b, c, d, a, 0x85845dd1,  x1, 21);
	MD5_II(a, b, c, d, 0x6fa87e4f,  x8,  6); MD5_II(d, a, b, c, 0xfe2ce6e0, x15, 10); MD5_II(c, d, a, b, 0xa3014314,  x6, 15); MD5_II(b, c, d, a, 0x4e0811a1, x13, 21);
	MD5_II(a, b, c, d, 0xf7537e82,  x4,  6); MD5_II(d, a, b, c, 0xbd3af235, x11, 10); MD5_II(c, d, a, b, 0x2ad7d2bb,  x2, 15); MD5_II(b, c, d, a, 0xeb86d391,  x9, 21);

	hash0 = a + 0x67452301;
	hash1 = b + 0xefcdab89;

	#undef MD5_FF
	#undef MD5_GG
	#undef MD5_HH
	#undef MD5_II

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15
}

__device__ __forceinline__ void md5(uint32_t pw0, uint32_t pw1, uint32_t pw2, uint32_t pw3, uint32_t pw4, uint32_t passwordLength, uint32_t &hash0, uint32_t &hash1)
{
	#define MD5_FF(a, b, c, d, C, x, s) a += (d ^ (b & (c ^ d))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_GG(a, b, c, d, C, x, s) a += (c ^ (d & (b ^ c))) + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_HH(a, b, c, d, C, x, s) a += (b ^ c ^ d)         + C + x; a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_II(a, b, c, d, C, x, s) a += (c ^ (b | ~d))      + C + x; a = (a << s) | (a >> (32 - s)); a += b;

	#define x0   pw0
	#define x1   pw1
	#define x2   pw2
	#define x3   pw3
	#define x4   pw4
	#define x5   0
	#define x6   0
	#define x7   0
	#define x8   0
	#define x9   0
	#define x10  0
	#define x11  0
	#define x12  0
	#define x13  0
	#define x14  passwordLength
	#define x15  0

	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476;

	MD5_FF(a, b, c, d, 0xd76aa478,  x0,  7); MD5_FF(d, a, b, c, 0xe8c7b756,  x1, 12); MD5_FF(c, d, a, b, 0x242070db,  x2, 17); MD5_FF(b, c, d, a, 0xc1bdceee,  x3, 22);
	MD5_FF(a, b, c, d, 0xf57c0faf,  x4,  7); MD5_FF(d, a, b, c, 0x4787c62a,  x5, 12); MD5_FF(c, d, a, b, 0xa8304613,  x6, 17); MD5_FF(b, c, d, a, 0xfd469501,  x7, 22);
	MD5_FF(a, b, c, d, 0x698098d8,  x8,  7); MD5_FF(d, a, b, c, 0x8b44f7af,  x9, 12); MD5_FF(c, d, a, b, 0xffff5bb1, x10, 17); MD5_FF(b, c, d, a, 0x895cd7be, x11, 22);
	MD5_FF(a, b, c, d, 0x6b901122, x12,  7); MD5_FF(d, a, b, c, 0xfd987193, x13, 12); MD5_FF(c, d, a, b, 0xa679438e, x14, 17); MD5_FF(b, c, d, a, 0x49b40821, x15, 22);

	MD5_GG(a, b, c, d, 0xf61e2562,  x1,  5); MD5_GG(d, a, b, c, 0xc040b340,  x6,  9); MD5_GG(c, d, a, b, 0x265e5a51, x11, 14); MD5_GG(b, c, d, a, 0xe9b6c7aa,  x0, 20);
	MD5_GG(a, b, c, d, 0xd62f105d,  x5,  5); MD5_GG(d, a, b, c, 0x02441453, x10,  9); MD5_GG(c, d, a, b, 0xd8a1e681, x15, 14); MD5_GG(b, c, d, a, 0xe7d3fbc8,  x4, 20);
	MD5_GG(a, b, c, d, 0x21e1cde6,  x9,  5); MD5_GG(d, a, b, c, 0xc33707d6, x14,  9); MD5_GG(c, d, a, b, 0xf4d50d87,  x3, 14); MD5_GG(b, c, d, a, 0x455a14ed,  x8, 20);
	MD5_GG(a, b, c, d, 0xa9e3e905, x13,  5); MD5_GG(d, a, b, c, 0xfcefa3f8,  x2,  9); MD5_GG(c, d, a, b, 0x676f02d9,  x7, 14); MD5_GG(b, c, d, a, 0x8d2a4c8a, x12, 20);

	MD5_HH(a, b, c, d, 0xfffa3942,  x5,  4); MD5_HH(d, a, b, c, 0x8771f681,  x8, 11); MD5_HH(c, d, a, b, 0x6d9d6122, x11, 16); MD5_HH(b, c, d, a, 0xfde5380c, x14, 23);
	MD5_HH(a, b, c, d, 0xa4beea44,  x1,  4); MD5_HH(d, a, b, c, 0x4bdecfa9,  x4, 11); MD5_HH(c, d, a, b, 0xf6bb4b60,  x7, 16); MD5_HH(b, c, d, a, 0xbebfbc70, x10, 23);
	MD5_HH(a, b, c, d, 0x289b7ec6, x13,  4); MD5_HH(d, a, b, c, 0xeaa127fa,  x0, 11); MD5_HH(c, d, a, b, 0xd4ef3085,  x3, 16); MD5_HH(b, c, d, a, 0x04881d05,  x6, 23);
	MD5_HH(a, b, c, d, 0xd9d4d039,  x9,  4); MD5_HH(d, a, b, c, 0xe6db99e5, x12, 11); MD5_HH(c, d, a, b, 0x1fa27cf8, x15, 16); MD5_HH(b, c, d, a, 0xc4ac5665,  x2, 23);

	MD5_II(a, b, c, d, 0xf4292244,  x0,  6); MD5_II(d, a, b, c, 0x432aff97,  x7, 10); MD5_II(c, d, a, b, 0xab9423a7, x14, 15); MD5_II(b, c, d, a, 0xfc93a039,  x5, 21);
	MD5_II(a, b, c, d, 0x655b59c3, x12,  6); MD5_II(d, a, b, c, 0x8f0ccc92,  x3, 10); MD5_II(c, d, a, b, 0xffeff47d, x10, 15); MD5_II(b, c, d, a, 0x85845dd1,  x1, 21);
	MD5_II(a, b, c, d, 0x6fa87e4f,  x8,  6); MD5_II(d, a, b, c, 0xfe2ce6e0, x15, 10); MD5_II(c, d, a, b, 0xa3014314,  x6, 15); MD5_II(b, c, d, a, 0x4e0811a1, x13, 21);
	MD5_II(a, b, c, d, 0xf7537e82,  x4,  6); MD5_II(d, a, b, c, 0xbd3af235, x11, 10); MD5_II(c, d, a, b, 0x2ad7d2bb,  x2, 15); MD5_II(b, c, d, a, 0xeb86d391,  x9, 21);

	hash0 = a + 0x67452301;
	hash1 = b + 0xefcdab89;

	#undef MD5_FF
	#undef MD5_GG
	#undef MD5_HH
	#undef MD5_II

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15
}

//#define SM_10
#define SM_20
#ifdef SM_20
	#define USE___fmul_rd
#elif defined(SM_10)
	// none
#endif

// benchmark me
// __shared__ uint32_t shared_hi[32], shared_lo[32];
// uint64_t dataOut;
// *((double*) (&dataOut)) = __hiloint2double(shared_hi[BaseIndex + tid],
//                                            shared_lo[BaseIndex + tid]);

// shared memory 16 KiB for 1.x and 48 KiB for 2.x

extern __shared__ uint32_t s_shared[];

template <int SIZE, bool KS_UNDER_41_BITS>
__global__ static void divSbMd5_doLinks_any(uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
{
/*
	__shared__ uint32_t s_shortCutShift[16];//   = {0, 6, 12, 18, 24, 30}; // ceil(log2(62 ^ pwLen)) - 24
	//                                                             1/62*2^0,              1/3844*2^6,           1/238328*2^12,         1/14776336*2^18,        1/916132832*2^24,       1/56800235584*2^30
	__shared__ float s_shortCutReciprocal[16];// = {8659208.0f/536870912.0f, 8938537.0f/536870912.0f, 9226877.0f/536870912.0f, 9524518.0f/536870912.0f, 9831761.0f/536870912.0f, 10148914.0f/536870912.0f};
	//                                              62^(4-3), 62^(5-3), 62^(6-3), 62^(7-3),  62^(8-3),    62^(9-3)
	__shared__ uint64_t s_shortCut[16];//        = {      62,     3844,   238328, 14776336, 916132832, 56800235584}; // 62 ^ (pwLen - 3)
	__shared__ float s_shortCutf[3];//           = {62.0f, 3844.0f, 238328.0f};
	__shared__ uint8_t s_charSet[256];
/*/
	uint64_t *s_shortCut           = (uint64_t*)  s_shared;
	uint32_t *s_shortCutShift      = (uint32_t*) (s_shortCut           + 16);
	float    *s_shortCutReciprocal = (float*)    (s_shortCutShift      + 16);
	float    *s_shortCutf          = (float*)    (s_shortCutReciprocal + 16);
	uint8_t  *s_charSet            = (uint8_t*)  (s_shortCutf          +  3);
//*/
	const uint32_t THREAD_ID   = blockIdx.x * blockDim.x + threadIdx.x;
	const uint32_t NUM_THREADS = gridDim.x * blockDim.x;

/*
	for (uint32_t i = threadIdx.x; i < 128 + 3; i += blockDim.x)
	{
		if (i < 32)
		{
			((uint32_t*) s_shortCut)[i] = d_shared[i];
		}
		else if (i < 48)
		{
			s_shortCutShift[i - 32] = d_shared[i];
		}
		else if (i < 64)
		{
			((uint32_t*) s_shortCutReciprocal)[i - 48] = d_shared[i];
		}
		else if (i < 128)
		{
			((uint32_t*) s_charSet)[i - 64] = d_shared[i];
		}
		else
		{
			s_shortCutf[i - 128] = (float) d_shared[2 * (i - 128)];
		}
	}
/*/
	for (uint32_t i = threadIdx.x; i < 128 + 3; i += blockDim.x)
	{
		s_shared[i] = d_shared[i];
	}
//*/
	__syncthreads();

	// work
	for (uint32_t i = THREAD_ID; i < numIndexes; i += NUM_THREADS)
	{
		uint32_t pw0, pw1, pw2, pw3, pw4, passwordLength;
		uint32_t hash0, hash1;
		uint64_t index;

		index = d_indexes[i];
		for (uint32_t curChainPos = startChainPos; curChainPos < endChainPos; curChainPos++)
		{
			// Index to pw
			uint64_t keySpaceBoundary = c_keySpaceBoundariesMax;
			passwordLength = c_numSubKeySpaces;
			while (index < keySpaceBoundary)
			{
				passwordLength--;
				keySpaceBoundary = c_keySpaceBoundaries[passwordLength];
			}
			index -= keySpaceBoundary;

			float r, q;
			float index1, index2, index3, index4, index5;

#ifdef USE___fmul_rd
	#define SHORT_CUT(index64, indexOut, shortCutShift, shortCutReciprocal, shortCut64) \
			indexOut = floorf(__fmul_rd((float) ((uint32_t) (index64 >> shortCutShift)), shortCutReciprocal)); \
			index64 -= shortCut64 * ((uint32_t) indexOut); \
			if (index64 >= shortCut64) \
			{ \
				index64 -= shortCut64; \
				indexOut++; \
			}
	#define SHORT_CUT_LAST(index64, indexOut, indexOut2, shortCutShift, shortCutReciprocal, shortCut32, shortCutf) \
			indexOut = floorf(__fmul_rd((float) ((uint32_t) (index64 >> shortCutShift)), shortCutReciprocal)); \
			indexOut2 = (uint32_t) (index64 - (uint64_t) ((uint32_t) shortCut32) * ((uint32_t) indexOut)); \
			if (indexOut2 >= shortCutf) \
			{ \
				indexOut2 -= shortCutf; \
				indexOut++; \
			}
#else
	#define SHORT_CUT(index64, indexOut, shortCutShift, shortCutReciprocal, shortCut64, shortCutf) \
			indexOut = floorf((float) ((uint32_t) (index64 >> shortCutShift)) * shortCutReciprocal); \
			index64 -= shortCut64 * ((uint32_t) indexOut); \
			if (index64 >= 0x8000000000000000) \
			{ \
				index64 += shortCut64; \
				indexOut--; \
			} \
			else if (index64 >= shortCut64) \
			{ \
				index64 -= shortCut64; \
				indexOut++; \
			}
	#define SHORT_CUT_LAST(index64, indexOut, indexOut2, shortCutShift, shortCutReciprocal, shortCut32, shortCutf) \
			indexOut = floorf((float) ((uint32_t) (index64 >> shortCutShift)) * shortCutReciprocal); \
			index32 = (uint32_t) (index64 - (uint64_t) ((uint32_t) shortCut32) * ((uint32_t) indexOut)); \
			if (index32 >= 0x80000000) \
			{ \
				index32 += shortCut32; \
				indexOut--; \
			} \
			indexOut2 = index32; \
			if (indexOut2 >= shortCutf) \
			{ \
				indexOut2 -= shortCutf; \
				indexOut++; \
			}
#endif

			if (SIZE > 12 && passwordLength > 12) // pw3 = (13, 14, 15, ...)
			{
				pw3 = 0x80;
				SHORT_CUT     (index, index1,         s_shortCutShift[passwordLength- 3-1], s_shortCutReciprocal[passwordLength- 3-1], s_shortCut[passwordLength- 3-1]);
				SHORT_CUT     (index, index2,         s_shortCutShift[passwordLength- 6-1], s_shortCutReciprocal[passwordLength- 6-1], s_shortCut[passwordLength- 6-1]);
				SHORT_CUT     (index, index3,         s_shortCutShift[passwordLength- 9-1], s_shortCutReciprocal[passwordLength- 9-1], s_shortCut[passwordLength- 9-1]);
				SHORT_CUT_LAST(index, index4, index5, s_shortCutShift[passwordLength-12-1], s_shortCutReciprocal[passwordLength-12-1], s_shortCut[passwordLength-12-1], s_shortCutf[passwordLength-12-1]);
			}
			else if (SIZE > 9 && passwordLength > 9) // pw2 = (..., 10, 11, 12)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12 && passwordLength == 12)
				{
					pw3 = 0x80;
				}
				else
				{
					if (SIZE >= 12)
					{
						pw3 = 0;
					}
					pw2 = 0x80;
				}
				SHORT_CUT     (index, index1,         s_shortCutShift[passwordLength-3-1], s_shortCutReciprocal[passwordLength-3-1], s_shortCut[passwordLength-3-1]);
				SHORT_CUT     (index, index2,         s_shortCutShift[passwordLength-6-1], s_shortCutReciprocal[passwordLength-6-1], s_shortCut[passwordLength-6-1]);
				SHORT_CUT_LAST(index, index3, index4, s_shortCutShift[passwordLength-9-1], s_shortCutReciprocal[passwordLength-9-1], s_shortCut[passwordLength-9-1], s_shortCutf[passwordLength-9-1]);
			}
			else if (SIZE > 6 && passwordLength > 6) // pw1 = (..., 7, 8), pw2 = (9, ...)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12) pw3 = 0;
				if (SIZE > 7 && passwordLength > 7)
				{
					pw2 = 0x80;
				}
				else
				{
					if (SIZE >= 8)
					{
						pw2 = 0;
					}
					pw1 = 0x80;
				}
				SHORT_CUT     (index, index1,         s_shortCutShift[passwordLength-3-1], s_shortCutReciprocal[passwordLength-3-1], s_shortCut[passwordLength-3-1]);
				SHORT_CUT_LAST(index, index2, index3, s_shortCutShift[passwordLength-6-1], s_shortCutReciprocal[passwordLength-6-1], s_shortCut[passwordLength-6-1], s_shortCutf[passwordLength-6-1]);
			}
			else if (SIZE > 3 && passwordLength > 3) // pw0 = (..., 4), pw1 = (5, 6, ...)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12) pw3 = 0;
				if (SIZE >=  8) pw2 = 0;
				pw1 = 0x80;
				SHORT_CUT_LAST(index, index1, index2, s_shortCutShift[passwordLength-3-1], s_shortCutReciprocal[passwordLength-3-1], s_shortCut[passwordLength-3-1], s_shortCutf[passwordLength-3-1]);
			}
			else // pw0 = (1, 2, 3, ...)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12) pw3 = 0;
				if (SIZE >=  8) pw2 = 0;
				if (SIZE >=  4) pw1 = 0;
				pw0 = 0x80;
				index1 = (uint32_t) index;
			}

#undef SHORT_CUT_NORM
#undef SHORT_CUT


#define GET_CHAR(size, index, pw, charSet, charSetLen, ncharSetLen, charSetReciprocal) \
			if (SIZE >= size && passwordLength >= size) \
			{ \
				q = floorf(index * charSetReciprocal); \
				r = ncharSetLen * q + index; \
				if (r >= charSetLen) \
				{ \
					r -= charSetLen; \
					q++; \
				} \
				index = q; \
				pw = (pw << 8) + charSet[(uint32_t) r]; \
			}
#define GET_CHAR_LAST(size, index, pw, charSet) \
			if (SIZE >= size && passwordLength >= size) \
			{ \
				pw = (pw << 8) + charSet[(uint32_t) index]; \
			}
#define GET_CHAR_NORM(size, index, pw)       GET_CHAR(size, index, pw, s_charSet, c_charSetLen, c_ncharSetLen, c_charSetReciprocal)
#define GET_CHAR_LAST_NORM(size, index, pw)  GET_CHAR_LAST(size,index,pw,s_charSet)

			GET_CHAR_NORM     (15,index5,pw3);
			GET_CHAR_NORM     (14,index5,pw3);
			GET_CHAR_LAST_NORM(13,index5,pw3);
			GET_CHAR_NORM     (12,index4,pw2);
			GET_CHAR_NORM     (11,index4,pw2);
			GET_CHAR_LAST_NORM(10,index4,pw2);
			GET_CHAR_NORM     ( 9,index3,pw2);
			GET_CHAR_NORM     ( 8,index3,pw1);
			GET_CHAR_LAST_NORM( 7,index3,pw1);
			GET_CHAR_NORM     ( 6,index2,pw1);
			GET_CHAR_NORM     ( 5,index2,pw1);
			GET_CHAR_LAST_NORM( 4,index2,pw0);
			GET_CHAR_NORM     ( 3,index1,pw0);
			GET_CHAR_NORM     ( 2,index1,pw0);
			GET_CHAR_LAST_NORM( 1,index1,pw0);

#undef GET_CHAR
#undef GET_CHAR_LAST
#undef GET_CHAR_NORM
#undef GET_CHAR_LAST_NORM

			// Pw to hash
			if (SIZE >= 16)
			{
				md5(pw0, pw1, pw2, pw3, pw4, 8*passwordLength, hash0, hash1);
			}
			else if (SIZE >= 12)
			{
				md5(pw0, pw1, pw2, pw3,      8*passwordLength, hash0, hash1);
			}
			else if (SIZE >= 8)
			{
				md5(pw0, pw1, pw2,           8*passwordLength, hash0, hash1);
			}
			else
			{
				md5(pw0, pw1,                8*passwordLength, hash0, hash1);
			}

			// Hash to index
			index = ((((uint64_t) hash1) << 32) + hash0) + curChainPos;
			if (KS_UNDER_41_BITS)
			{
				// This *should* be faster but is about the same speed:
				// ((uint32_t) (index >> 32)) & 0xffffff00
				// than this:
				// (uint32_t) (index >> 40)
				// because a 64 bit int is just two 32 bit registers so "((uint32_t) (index >> 32))"
				// should just grab the high 32 bit register which costs nothing then and 0xffffff00
				// to it. This can be executed on every stream processor unlike a 32 bit shift which
				// on compute capability 2.0 and 2.1 run on 1/2 and 1/3 respectively of the stream
				// processors. Note that "((uint32_t) (index >> 40))" should be grab the high 32 bit
				// register then right shift by 8.
#ifdef USE___fmul_rd
				uint32_t tmp = floorf(__fmul_rd((float) ((uint32_t) (index >> 40)), c_keySpaceReciprocalFirst));
				index = index - c_keySpaceFirst * tmp;
#else
				uint32_t tmp = floorf((float) ((uint32_t) (index >> 40)) * c_keySpaceReciprocalFirst);
				index = index - c_keySpaceFirst * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpaceFirst;
				}
#endif
#ifdef USE___fmul_rd
				tmp = floorf(__fmul_rd(((float) index) - 131072.0f, c_keySpaceReciprocal));
				index = index - c_keySpace * tmp;
#else
				tmp = floorf((((float) index) - 131072.0f) * c_keySpaceReciprocal);
				index = index - c_keySpace * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpace;
				}
#endif
				if (index >= c_keySpace)
				{
					index -= c_keySpace;
				}
			}
			else
			{
#ifdef USE___fmul_rd
				uint32_t tmp = floorf(__fmul_rd((float) ((uint32_t) (index >> 40)), c_keySpaceReciprocal));
				index = index - c_keySpace * tmp;
#else
				uint32_t tmp = floorf((float) ((uint32_t) (index >> 40)) * c_keySpaceReciprocal);
				index = index - c_keySpace * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpace;
				}
#endif
				if (index >= c_keySpace)
				{
					index -= c_keySpace;
				}
			}
		}
		d_indexes[i] = index;
	}
}

template <int SIZE, bool KS_UNDER_41_BITS>
__global__ static void divSbMd5_doPreWorkLinks_any(uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
{
	__shared__ uint32_t s_hash[2];
/*
	__shared__ uint32_t s_shortCutShift[16];//   = {0, 6, 12, 18, 24, 30}; // ceil(log2(62 ^ pwLen)) - 24
	//                                                             1/62*2^0,              1/3844*2^6,           1/238328*2^12,         1/14776336*2^18,        1/916132832*2^24,       1/56800235584*2^30
	__shared__ float s_shortCutReciprocal[16];// = {8659208.0f/536870912.0f, 8938537.0f/536870912.0f, 9226877.0f/536870912.0f, 9524518.0f/536870912.0f, 9831761.0f/536870912.0f, 10148914.0f/536870912.0f};
	//                                              62^(4-3), 62^(5-3), 62^(6-3), 62^(7-3),  62^(8-3),    62^(9-3)
	__shared__ uint64_t s_shortCut[16];//        = {      62,     3844,   238328, 14776336, 916132832, 56800235584}; // 62 ^ (pwLen - 3)
	__shared__ float s_shortCutf[3];//           = {62.0f, 3844.0f, 238328.0f};
	__shared__ uint8_t s_charSet[256];
/*/
	uint64_t *s_shortCut           = (uint64_t*)  s_shared;
	uint32_t *s_shortCutShift      = (uint32_t*) (s_shortCut           + 16);
	float    *s_shortCutReciprocal = (float*)    (s_shortCutShift      + 16);
	float    *s_shortCutf          = (float*)    (s_shortCutReciprocal + 16);
	uint8_t  *s_charSet            = (uint8_t*)  (s_shortCutf          +  3);
//*/
	const uint32_t THREAD_ID   = blockIdx.x * blockDim.x + threadIdx.x;
	const uint32_t NUM_THREADS = gridDim.x * blockDim.x;

/*
	for (uint32_t i = threadIdx.x; i < 128 + 3 + 2; i += blockDim.x)
	{
		if (i < 32)
		{
			((uint32_t*) s_shortCut)[i] = d_shared[i];
		}
		else if (i < 48)
		{
			s_shortCutShift[i - 32] = d_shared[i];
		}
		else if (i < 64)
		{
			((uint32_t*) s_shortCutReciprocal)[i - 48] = d_shared[i];
		}
		else if (i < 128)
		{
			((uint32_t*) s_charSet)[i - 64] = d_shared[i];
		}
		else if (i < 128 + 3)
		{
			s_shortCutf[i - 128] = (float) d_shared[2 * (i - 128)];
		}
		else if (i == 128 + 3)
		{
			s_hash0 = d_hash[0];
		}
		else
		{
			s_hash1 = d_hash[1];
		}
	}
/*/
	for (uint32_t i = threadIdx.x; i < 128 + 3 + 2; i += blockDim.x)
	{
		if (i < 128 + 3)
		{
			s_shared[i] = d_shared[i];
		}
		else
		{
			s_hash[i - (128 + 3)] = d_hash[i - (128 + 3)];
		}
	}
//*/
	__syncthreads();

	// work
	uint32_t endChainPos2 = endChainPos;
	for (uint32_t i = THREAD_ID; i < numIndexes; i += NUM_THREADS)
	{
		uint32_t pw0, pw1, pw2, pw3, pw4, passwordLength;
		uint32_t hash0, hash1;
		uint32_t curChainPos = firstChainPos + i;
		uint64_t index;

		if (first)
		{
			// Hash to index
			index = ((((uint64_t) s_hash[1]) << 32) + s_hash[0]) + curChainPos;
			if (KS_UNDER_41_BITS)
			{
				// This *should* be faster but is about the same speed:
				// ((uint32_t) (index >> 32)) & 0xffffff00
				// than this:
				// (uint32_t) (index >> 40)
				// because a 64 bit int is just two 32 bit registers so "((uint32_t) (index >> 32))"
				// should just grab the high 32 bit register which costs nothing then and 0xffffff00
				// to it. This can be executed on every stream processor unlike a 32 bit shift which
				// on compute capability 2.0 and 2.1 run on 1/2 and 1/3 respectively of the stream
				// processors. Note that "((uint32_t) (index >> 40))" should be grab the high 32 bit
				// register then right shift by 8.
#ifdef USE___fmul_rd
				uint32_t tmp = floorf(__fmul_rd((float) ((uint32_t) (index >> 40)), c_keySpaceReciprocalFirst));
				index = index - c_keySpaceFirst * tmp;
#else
				uint32_t tmp = floorf((float) ((uint32_t) (index >> 40)) * c_keySpaceReciprocalFirst);
				index = index - c_keySpaceFirst * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpaceFirst;
				}
#endif
#ifdef USE___fmul_rd
				tmp = floorf(__fmul_rd(((float) index) - 131072.0f, c_keySpaceReciprocal));
				index = index - c_keySpace * tmp;
#else
				tmp = floorf((((float) index) - 131072.0f) * c_keySpaceReciprocal);
				index = index - c_keySpace * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpace;
				}
#endif
				if (index >= c_keySpace)
				{
					index -= c_keySpace;
				}
			}
			else
			{
#ifdef USE___fmul_rd
				uint32_t tmp = floorf(__fmul_rd((float) ((uint32_t) (index >> 40)), c_keySpaceReciprocal));
				index = index - c_keySpace * tmp;
#else
				uint32_t tmp = floorf((float) ((uint32_t) (index >> 40)) * c_keySpaceReciprocal);
				index = index - c_keySpace * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpace;
				}
#endif
				if (index >= c_keySpace)
				{
					index -= c_keySpace;
				}
			}
		}
		else
		{
			index = d_indexes[i];
		}
		endChainPos2 = firstEndChainPos + i;
		if (endChainPos2 > endChainPos)
		{
			endChainPos2 = endChainPos;
		}
		for (curChainPos++; curChainPos < endChainPos2; curChainPos++)
		{
			// Index to pw
			uint64_t keySpaceBoundary = c_keySpaceBoundariesMax;
			passwordLength = c_numSubKeySpaces;
			while (index < keySpaceBoundary)
			{
				passwordLength--;
				keySpaceBoundary = c_keySpaceBoundaries[passwordLength];
			}
			index -= keySpaceBoundary;

			float r, q;
			float index1, index2, index3, index4, index5;

#ifdef USE___fmul_rd
	#define SHORT_CUT(index64, indexOut, shortCutShift, shortCutReciprocal, shortCut64) \
			indexOut = floorf(__fmul_rd((float) ((uint32_t) (index64 >> shortCutShift)), shortCutReciprocal)); \
			index64 -= shortCut64 * ((uint32_t) indexOut); \
			if (index64 >= shortCut64) \
			{ \
				index64 -= shortCut64; \
				indexOut++; \
			}
	#define SHORT_CUT_LAST(index64, indexOut, indexOut2, shortCutShift, shortCutReciprocal, shortCut32, shortCutf) \
			indexOut = floorf(__fmul_rd((float) ((uint32_t) (index64 >> shortCutShift)), shortCutReciprocal)); \
			indexOut2 = (uint32_t) (index64 - (uint64_t) ((uint32_t) shortCut32) * ((uint32_t) indexOut)); \
			if (indexOut2 >= shortCutf) \
			{ \
				indexOut2 -= shortCutf; \
				indexOut++; \
			}
#else
	#define SHORT_CUT(index64, indexOut, shortCutShift, shortCutReciprocal, shortCut64, shortCutf) \
			indexOut = floorf((float) ((uint32_t) (index64 >> shortCutShift)) * shortCutReciprocal); \
			index64 -= shortCut64 * ((uint32_t) indexOut); \
			if (index64 >= 0x8000000000000000) \
			{ \
				index64 += shortCut64; \
				indexOut--; \
			} \
			else if (index64 >= shortCut64) \
			{ \
				index64 -= shortCut64; \
				indexOut++; \
			}
	#define SHORT_CUT_LAST(index64, indexOut, indexOut2, shortCutShift, shortCutReciprocal, shortCut32, shortCutf) \
			indexOut = floorf((float) ((uint32_t) (index64 >> shortCutShift)) * shortCutReciprocal); \
			index32 = (uint32_t) (index64 - (uint64_t) ((uint32_t) shortCut32) * ((uint32_t) indexOut)); \
			if (index32 >= 0x80000000) \
			{ \
				index32 += shortCut32; \
				indexOut--; \
			} \
			indexOut2 = index32; \
			if (indexOut2 >= shortCutf) \
			{ \
				indexOut2 -= shortCutf; \
				indexOut++; \
			}
#endif

			if (SIZE > 12 && passwordLength > 12) // pw3 = (13, 14, 15, ...)
			{
				pw3 = 0x80;
				SHORT_CUT     (index, index1,         s_shortCutShift[passwordLength- 3-1], s_shortCutReciprocal[passwordLength- 3-1], s_shortCut[passwordLength- 3-1]);
				SHORT_CUT     (index, index2,         s_shortCutShift[passwordLength- 6-1], s_shortCutReciprocal[passwordLength- 6-1], s_shortCut[passwordLength- 6-1]);
				SHORT_CUT     (index, index3,         s_shortCutShift[passwordLength- 9-1], s_shortCutReciprocal[passwordLength- 9-1], s_shortCut[passwordLength- 9-1]);
				SHORT_CUT_LAST(index, index4, index5, s_shortCutShift[passwordLength-12-1], s_shortCutReciprocal[passwordLength-12-1], s_shortCut[passwordLength-12-1], s_shortCutf[passwordLength-12-1]);
			}
			else if (SIZE > 9 && passwordLength > 9) // pw2 = (..., 10, 11, 12)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12 && passwordLength == 12)
				{
					pw3 = 0x80;
				}
				else
				{
					if (SIZE >= 12)
					{
						pw3 = 0;
					}
					pw2 = 0x80;
				}
				SHORT_CUT     (index, index1,         s_shortCutShift[passwordLength-3-1], s_shortCutReciprocal[passwordLength-3-1], s_shortCut[passwordLength-3-1]);
				SHORT_CUT     (index, index2,         s_shortCutShift[passwordLength-6-1], s_shortCutReciprocal[passwordLength-6-1], s_shortCut[passwordLength-6-1]);
				SHORT_CUT_LAST(index, index3, index4, s_shortCutShift[passwordLength-9-1], s_shortCutReciprocal[passwordLength-9-1], s_shortCut[passwordLength-9-1], s_shortCutf[passwordLength-9-1]);
			}
			else if (SIZE > 6 && passwordLength > 6) // pw1 = (..., 7, 8), pw2 = (9, ...)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12) pw3 = 0;
				if (SIZE > 7 && passwordLength > 7)
				{
					pw2 = 0x80;
				}
				else
				{
					if (SIZE >= 8)
					{
						pw2 = 0;
					}
					pw1 = 0x80;
				}
				SHORT_CUT     (index, index1,         s_shortCutShift[passwordLength-3-1], s_shortCutReciprocal[passwordLength-3-1], s_shortCut[passwordLength-3-1]);
				SHORT_CUT_LAST(index, index2, index3, s_shortCutShift[passwordLength-6-1], s_shortCutReciprocal[passwordLength-6-1], s_shortCut[passwordLength-6-1], s_shortCutf[passwordLength-6-1]);
			}
			else if (SIZE > 3 && passwordLength > 3) // pw0 = (..., 4), pw1 = (5, 6, ...)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12) pw3 = 0;
				if (SIZE >=  8) pw2 = 0;
				pw1 = 0x80;
				SHORT_CUT_LAST(index, index1, index2, s_shortCutShift[passwordLength-3-1], s_shortCutReciprocal[passwordLength-3-1], s_shortCut[passwordLength-3-1], s_shortCutf[passwordLength-3-1]);
			}
			else // pw0 = (1, 2, 3, ...)
			{
				if (SIZE >= 16) pw4 = 0;
				if (SIZE >= 12) pw3 = 0;
				if (SIZE >=  8) pw2 = 0;
				if (SIZE >=  4) pw1 = 0;
				pw0 = 0x80;
				index1 = (uint32_t) index;
			}

#undef SHORT_CUT_NORM
#undef SHORT_CUT


#define GET_CHAR(size, index, pw, charSet, charSetLen, ncharSetLen, charSetReciprocal) \
			if (SIZE >= size && passwordLength >= size) \
			{ \
				q = floorf(index * charSetReciprocal); \
				r = ncharSetLen * q + index; \
				if (r >= charSetLen) \
				{ \
					r -= charSetLen; \
					q++; \
				} \
				index = q; \
				pw = (pw << 8) + charSet[(uint32_t) r]; \
			}
#define GET_CHAR_LAST(size, index, pw, charSet) \
			if (SIZE >= size && passwordLength >= size) \
			{ \
				pw = (pw << 8) + charSet[(uint32_t) index]; \
			}
#define GET_CHAR_NORM(size, index, pw)       GET_CHAR(size, index, pw, s_charSet, c_charSetLen, c_ncharSetLen, c_charSetReciprocal)
#define GET_CHAR_LAST_NORM(size, index, pw)  GET_CHAR_LAST(size,index,pw,s_charSet)

			GET_CHAR_NORM     (15,index5,pw3);
			GET_CHAR_NORM     (14,index5,pw3);
			GET_CHAR_LAST_NORM(13,index5,pw3);
			GET_CHAR_NORM     (12,index4,pw2);
			GET_CHAR_NORM     (11,index4,pw2);
			GET_CHAR_LAST_NORM(10,index4,pw2);
			GET_CHAR_NORM     ( 9,index3,pw2);
			GET_CHAR_NORM     ( 8,index3,pw1);
			GET_CHAR_LAST_NORM( 7,index3,pw1);
			GET_CHAR_NORM     ( 6,index2,pw1);
			GET_CHAR_NORM     ( 5,index2,pw1);
			GET_CHAR_LAST_NORM( 4,index2,pw0);
			GET_CHAR_NORM     ( 3,index1,pw0);
			GET_CHAR_NORM     ( 2,index1,pw0);
			GET_CHAR_LAST_NORM( 1,index1,pw0);

#undef GET_CHAR
#undef GET_CHAR_LAST
#undef GET_CHAR_NORM
#undef GET_CHAR_LAST_NORM

			// Pw to hash
			if (SIZE >= 16)
			{
				md5(pw0, pw1, pw2, pw3, pw4, 8*passwordLength, hash0, hash1);
			}
			else if (SIZE >= 12)
			{
				md5(pw0, pw1, pw2, pw3,      8*passwordLength, hash0, hash1);
			}
			else if (SIZE >= 8)
			{
				md5(pw0, pw1, pw2,           8*passwordLength, hash0, hash1);
			}
			else
			{
				md5(pw0, pw1,                8*passwordLength, hash0, hash1);
			}

			// Hash to index
			index = ((((uint64_t) hash1) << 32) + hash0) + curChainPos;
			if (KS_UNDER_41_BITS)
			{
				// This *should* be faster but is about the same speed:
				// ((uint32_t) (index >> 32)) & 0xffffff00
				// than this:
				// (uint32_t) (index >> 40)
				// because a 64 bit int is just two 32 bit registers so "((uint32_t) (index >> 32))"
				// should just grab the high 32 bit register which costs nothing then and 0xffffff00
				// to it. This can be executed on every stream processor unlike a 32 bit shift which
				// on compute capability 2.0 and 2.1 run on 1/2 and 1/3 respectively of the stream
				// processors. Note that "((uint32_t) (index >> 40))" should be grab the high 32 bit
				// register then right shift by 8.
#ifdef USE___fmul_rd
				uint32_t tmp = floorf(__fmul_rd((float) ((uint32_t) (index >> 40)), c_keySpaceReciprocalFirst));
				index = index - c_keySpaceFirst * tmp;
#else
				uint32_t tmp = floorf((float) ((uint32_t) (index >> 40)) * c_keySpaceReciprocalFirst);
				index = index - c_keySpaceFirst * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpaceFirst;
				}
#endif
#ifdef USE___fmul_rd
				tmp = floorf(__fmul_rd(((float) index) - 131072.0f, c_keySpaceReciprocal));
				index = index - c_keySpace * tmp;
#else
				tmp = floorf((((float) index) - 131072.0f) * c_keySpaceReciprocal);
				index = index - c_keySpace * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpace;
				}
#endif
				if (index >= c_keySpace)
				{
					index -= c_keySpace;
				}
			}
			else
			{
#ifdef USE___fmul_rd
				uint32_t tmp = floorf(__fmul_rd((float) ((uint32_t) (index >> 40)), c_keySpaceReciprocal));
				index = index - c_keySpace * tmp;
#else
				uint32_t tmp = floorf((float) ((uint32_t) (index >> 40)) * c_keySpaceReciprocal);
				index = index - c_keySpace * tmp;
				if (index >= 0x8000000000000000)
				{
					index += c_keySpace;
				}
#endif
				if (index >= c_keySpace)
				{
					index -= c_keySpace;
				}
			}
		}
		d_indexes[i] = index;
	}
}

template <int SIZE, bool KS_UNDER_41_BITS>
__global__ static void divSbMd5_falseAlarmChecking_any(uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	;
}


// **** doLinks ****
void divSbMd5Kernel_doLinks_ksUnder41Bits_any7(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
{
	divSbMd5_doLinks_any<7,true><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
}

void divSbMd5Kernel_doLinks_ksOver41Bits_any7(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
{
	divSbMd5_doLinks_any<7,false><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
}

void divSbMd5Kernel_doLinks_ksUnder41Bits_any11(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
{
	divSbMd5_doLinks_any<11,true><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
}

void divSbMd5Kernel_doLinks_ksOver41Bits_any11(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
{
	divSbMd5_doLinks_any<11,false><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
}

void divSbMd5Kernel_doLinks_ksUnder41Bits_any15(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
{
	divSbMd5_doLinks_any<15,true><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
}

void divSbMd5Kernel_doLinks_ksOver41Bits_any15(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
{
	divSbMd5_doLinks_any<15,false><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
}

//void divSbMd5Kernel_doLinks_ksUnder41Bits_any19(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
//{
//	divSbMd5_doLinks_any<19,true><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
//}

//void divSbMd5Kernel_doLinks_ksOver41Bits_any19(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t startChainPos, uint32_t endChainPos, uint32_t numIndexes)
//{
//	divSbMd5_doLinks_any<19,false><<<blocks, threads, 524>>>(d_shared, d_indexes, startChainPos, endChainPos, numIndexes);
//}

/*
// **** doPreWorkLinks ****
void divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any7(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
{
	divSbMd5_doPreWorkLinks_any<7,true><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
}

void divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any7(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
{
	divSbMd5_doPreWorkLinks_any<7,false><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
}

void divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any11(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
{
	divSbMd5_doPreWorkLinks_any<11,true><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
}

void divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any11(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
{
	divSbMd5_doPreWorkLinks_any<11,false><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
}

void divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any15(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
{
	divSbMd5_doPreWorkLinks_any<15,true><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
}

void divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any15(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
{
	divSbMd5_doPreWorkLinks_any<15,false><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
}

//void divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any19(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
//{
//	divSbMd5_doPreWorkLinks_any<19,true><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
//}

//void divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any19(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, uint32_t firstChainPos, uint32_t firstEndChainPos, uint32_t endChainPos, uint32_t numIndexes, const uint32_t *d_hash, bool first)
//{
//	divSbMd5_doPreWorkLinks_any<19,false><<<blocks, threads, 524>>>(d_shared, d_indexes, firstChainPos, firstEndChainPos, endChainPos, numIndexes, d_hash, first);
//}


// **** falseAlarmChecking ****
unsigned int divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any7(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	divSbMd5_falseAlarmChecking_any<7,true><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
	return 0;
}

unsigned int divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any7(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	divSbMd5_falseAlarmChecking_any<7,false><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
	return 0;
}

unsigned int divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any11(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	divSbMd5_falseAlarmChecking_any<11,true><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
	return 0;
}

unsigned int divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any11(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	divSbMd5_falseAlarmChecking_any<11,false><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
	return 0;
}

unsigned int divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any15(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	divSbMd5_falseAlarmChecking_any<15,true><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
	return 0;
}

unsigned int divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any15(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
	divSbMd5_falseAlarmChecking_any<15,false><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
	return 0;
}

//unsigned int divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any19(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
//{
//	divSbMd5_falseAlarmChecking_any<19,true><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
//	return 0;
//}

//unsigned int divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any19(int blocks, int threads, uint32_t *d_shared, uint64_t *d_indexes, const uint32_t *d_hash, const uint64_t *d_lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
//{
//	divSbMd5_falseAlarmChecking_any<19,false><<<blocks, threads, 524>>>(d_shared, d_indexes, d_hash, d_lastChainPoses, tableIndex, numIndexes);
//	return 0;
//}
*/
extern "C" int divSbMd5_getKernels(uint64_t keySpace, uint32_t m_maxPwLen, kernel_doLinks *doLinks, kernel_doPreWorkLinks *doPreWorkLinks, kernel_falseAlarmChecking *falseAlarmChecking)
{
	if (keySpace <= ((uint64_t) 1) << 41)
	{
		if (m_maxPwLen <= 7)
		{
			*doLinks            = divSbMd5Kernel_doLinks_ksUnder41Bits_any7;
//			*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any7;
//			*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any7;
		}
		else if (m_maxPwLen <= 11)
		{
			*doLinks            = divSbMd5Kernel_doLinks_ksUnder41Bits_any11;
//			*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any11;
//			*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any11;
		}
		else if (m_maxPwLen <= 15)
		{
			*doLinks            = divSbMd5Kernel_doLinks_ksUnder41Bits_any15;
//			*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any15;
//			*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any15;
		}
//		else if (m_maxPwLen <= 19)
//		{
//			*doLinks            = divSbMd5Kernel_doLinks_ksUnder41Bits_any19;
//			*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksUnder41Bits_any19;
//			*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksUnder41Bits_any19;
//		}
		else
		{
			return 1;
		}
	}
	else if (m_maxPwLen <= 7)
	{
		*doLinks            = divSbMd5Kernel_doLinks_ksOver41Bits_any7;
//		*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any7;
//		*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any7;
	}
	else if (m_maxPwLen <= 11)
	{
		*doLinks            = divSbMd5Kernel_doLinks_ksOver41Bits_any11;
//		*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any11;
//		*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any11;
	}
	else if (m_maxPwLen <= 15)
	{
		*doLinks            = divSbMd5Kernel_doLinks_ksOver41Bits_any15;
//		*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any15;
//		*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any15;
	}
//	else if (m_maxPwLen <= 19)
//	{
//		*doLinks            = divSbMd5Kernel_doLinks_ksOver41Bits_any19;
//		*doPreWorkLinks     = divSbMd5Kernel_doPreWorkLinks_ksOver41Bits_any19;
//		*falseAlarmChecking = divSbMd5Kernel_falseAlarmChecking_ksOver41Bits_any19;
//	}
	else
	{
		return 1;
	}
	return 0;
}
