/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "gpudivsb.h"
#include "reduxfactory.h"
#include "hashfactory.h"
#include "cudacommon.h"
#include "divsbmd5.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>

using namespace std;

struct compactedSubKeySpace
{
	uint32_t size, pos;
};

bool compCompactedSubKeySpace(compactedSubKeySpace *i, compactedSubKeySpace *j)
{
	return i->size < j->size;
}

uint32_t GpuDivSB::s_inited = GpuDivSB::initFactory();

uint32_t GpuDivSB::initFactory()
{
	uint32_t ret;

	ret  = ReduxFactory::registerRedux("div", "Uses divide",
		CE_ASCII_BIN8 | CE_UTF8 | CE_FIX_WIDTH_CHAR | CE_80_NULL_PADDED | CE_NULL_PADDED, NULL, GpuDivSB::create);
//	ret += ReduxFactory::registerRedux("divh3", "Uses divide, hybrid3 or multi-masked",
//		CE_ASCII_BIN8 | CE_UTF8 | CE_FIX_WIDTH_CHAR | CE_80_NULL_PADDED | CE_NULL_PADDED, NULL, GpuDivSB::create);
	return ret;
}

GpuDivSB::~GpuDivSB()
{
	cleanUp();
}

GpuDivSB::GpuDivSB(const keySpace &ks, const char *hashName, int gpu)
	: ReduxBruteForce(ks, hashName),
	ReduxSingleByte(ks, hashName),
	DivBase(ks, hashName)
{
	uint32_t numSubKeySpaces = m_numSubKeySpaces;

	m_d_indexes = NULL;
	m_d_lastChainPoses = NULL;
	m_d_hash = NULL;
	m_d_shared = NULL;
	if (isInited())
	{
		uint32_t invalidKeySpace = 0;
		keySpace::const_iterator curSubKeySpace, nextSubKeySpace;
		subKeySpace::const_iterator curCharSet, charSet;

		// Check that the key space is a normal key space
		// EX: charset#a-b
		charSet = ks.begin()->begin();
		for (++(nextSubKeySpace = curSubKeySpace = ks.begin()); nextSubKeySpace != ks.end(); ++curSubKeySpace, ++nextSubKeySpace)
		{
			// Check password length increased by one
			// aaannn0
			// aaannn00
			if (curSubKeySpace->size() + 1 != nextSubKeySpace->size())
			{
				invalidKeySpace = 1;
				break;
			}
			// Check character sets are the same
			for (curCharSet = curSubKeySpace->begin(); curCharSet != curSubKeySpace->end(); ++curCharSet)
			{
				if (!(*curCharSet == *charSet))
				{
					invalidKeySpace = 1;
					break;
				}
			}
		}
		if (ks.size() == 1)
		{
			// Check character sets are the same
			for (curCharSet = curSubKeySpace->begin(); curCharSet != curSubKeySpace->end(); ++curCharSet)
			{
				if (!(*curCharSet == *charSet))
				{
					invalidKeySpace = 1;
					break;
				}
			}
		}
		if (invalidKeySpace)
		{
			reportError("Invalid key space.");
			cleanUp();
			return;
		}
		if (strcmp(hashName, "md5") == 0)
		{
			if (divSbMd5_getKernels(m_keySpace, m_maxPwLen, &m_doLinks, &m_doPreWorkLinks, &m_falseAlarmChecking))
			{
				reportError("Max password length too long.");
				cleanUp();
				return;
			}
		}/*
		else if (strcmp(hashName, "ntlm") == 0)
		{
			if (divSbNtlm_getKernels(m_keySpace, m_maxPwLen, &m_doLinks, &m_doPreWorkLinks, &m_falseAlarmChecking))
			{
				reportError("Max password length too long.");
				cleanUp();
				return;
			}
		}*/
		else
		{
			reportError("No suitable hashing algorithm found for GPU.");
			cleanUp();
			return;
		}
		checkCudaErrors(cudaSetDevice(gpu));
		getDefaultBlocksThreads(gpu, m_blocks, m_threads);
		m_optimalNumIndexes = m_blocks * m_threads;
		checkCudaErrors(cudaMalloc((void**) &m_d_indexes,        sizeof(uint64_t) * 2 * m_optimalNumIndexes));
		checkCudaErrors(cudaMalloc((void**) &m_d_lastChainPoses, sizeof(uint64_t) * 2 * m_optimalNumIndexes));
		checkCudaErrors(cudaMalloc((void**) &m_d_hash,           sizeof(uint32_t) * ((m_hashLen + 3) / 4)));
/* Hybrid stuff:
		// Compact sub key spaces and avoid bank conflicts
		// *
		// **
		// ***
		// ****               Optimal answer                Simple answer
		// *****              maaaaaan00maaaaa              *******MMMMMMMMM
		// ******             ^         ^      Start points ^      ^       ^
		// *******            0000maaaaaaan00*              nnnnnnnnmaaaaa00
		// MMMMMMMM               ^          ^ Start points         ^
		// Mnnnnnnnn          ******maaaaaa000              00maaaaaan00maaa
		// maaaaaaaan     ->        ^      ^   Start points   ^         ^
		// 0000000000         00000000000maaaa              aaaaanmaaaaaa000
		// 00000000000                   ^     Start points       ^
		// 000000000000       aaaanMMMMMMMMnnn              0maaaaaaan00-000
		// 0000000000000           ^      ^    Start points  ^           ^
		// 00000000000000     nnnnn                         00000000000
		// maaaaa0000
		// maaaaaan00         ^   ^^^   ^^^^ ^  Used banks  ^^^   ^^^   ^^ ^
		// maaaaaa0000
		// maaaaaaan00
		//
		// This is an optimal answer
		// maaaaaan00maaaaa0000maaaaaaan00*******maaaaaa00000000000000maaaaaaaanMMMMMMMMnnnnnnnn
		// ^         ^         ^          ^      ^      ^             ^         ^      ^
		//
		// This code does a very simple method which gets this result
		// (wasted 1 spot and didn't compact 5 spots for a total 384 bytes extra)
		// *******MMMMMMMMMnnnnnnnnmaaaaa0000maaaaaan00maaaaaaaanmaaaaaa0000maaaaaaan00-00000000000000
		// ^      ^       ^        ^         ^         ^         ^          ^           ^
		//
		// Totally beats using the CPU version:
		// ***************000000000000000000000******000000000000maaaaa00000000000000000maaaaaan00*******00000000000000maaaaaa0000MnnnnnnnnMMMMMMMMmaaaaaaaanmaaaaaaan00
		// ^^ ^  ^   ^    ^         ^          ^     ^           ^         ^            ^         ^      ^             ^          ^        ^       ^         ^
 
		// Compact sub key spaces
		// To do: Compact partial matches and in the middle matches:
		//        maaaaa0000 and 00000000000000 ->
		//                 maaaaa00000000000000
		//        maaaaa0000 and aaaaa00 -> maaaaa0000  (rare but should do this if it can't be anywhere
		//                                               else that doesn't decrease the number of banks)
		compactedSubKeySpace **compactedSubKeySpaces, *compactedSubKeySpacesIndex;
		uint32_t *compactedInto, i, j, numCompactedSubKeySpaces = 0;

		compactedInto = new uint32_t[numSubKeySpaces];
		compactedSubKeySpaces = new compactedSubKeySpace*[numSubKeySpaces];
		compactedSubKeySpacesIndex = new compactedSubKeySpace[numSubKeySpaces];
		i = numSubKeySpaces - 1;
		for (keySpace::const_reverse_iterator curSubKeySpace = ks.rbegin(); curSubKeySpace != ks.rend(); ++curSubKeySpace, i--)
		{
			compactedInto[i] = i;
			j = numSubKeySpaces - 1;
			for (keySpace::const_reverse_iterator otherSubKeySpace = ks.rbegin(); otherSubKeySpace != curSubKeySpace; ++otherSubKeySpace, j--)
			{
				if (otherSubKeySpace->size() > curSubKeySpace->size())
				{
					subKeySpace::const_iterator curCharSet, otherCharSet;

					curCharSet = curSubKeySpace->begin();
					otherCharSet = otherSubKeySpace->begin();
					for (; curCharSet != curSubKeySpace->end(); ++curCharSet, ++otherCharSet)
					{
						if (!(*curCharSet == *otherCharSet))
						{
							break;
						}
					}
					if (curCharSet == curSubKeySpace->end())
					{
						compactedInto[i] = j;
						break;
					}
				}
			}
		}
		i = 0;
		for (keySpace::const_iterator curSubKeySpace = ks.begin(); i < numSubKeySpaces; i++, ++curSubKeySpace)
		{
			if (compactedInto[i] == i)
			{
				compactedSubKeySpaces[numCompactedSubKeySpaces] = compactedSubKeySpacesIndex + i;
				compactedSubKeySpacesIndex[i].size = curSubKeySpace->size();
				numCompactedSubKeySpaces++;
			}
		}
		// I'm pretty sure this helps
		sort(compactedSubKeySpaces, compactedSubKeySpaces + numCompactedSubKeySpaces, compCompactedSubKeySpace);

		// Avoid bank conflicts
		// To do: Have the smallest sub key spaces conflict (if maxCollisions > 1) as they
		//        are the least likely to be used which means basically no bank conflicts
		// To do: Find a solution that has the optimal size (uncompact if needed)
		uint32_t maxCollisions = (numCompactedSubKeySpaces + 15) / 16;
		uint32_t banks[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
		uint32_t pos = 0;

		for (i = 0; i < numCompactedSubKeySpaces; i++)
		{
			while (banks[pos & 15] == maxCollisions)
			{
				pos++;
			}
			banks[pos & 15]++;
			compactedSubKeySpaces[i]->pos = pos;
			pos += compactedSubKeySpaces[i]->size;
		}
		uint32_t numSubKeySpaceCharSets = pos;

		// To get the offsets for (s_charSets, s_charSetLengths, s_charSetLengthsFPM)
		// into (s_charSetsSubKeySpace, s_charSetLengthsData, s_charSetLengthsFPMData)
		// is "compactedSubKeySpacesIndex[compactedInto[subKeySpaceIndex]].pos"
		// So:
		// s_charSets[i]          = s_charSetsSubKeySpace   + compactedSubKeySpacesIndex[compactedInto[i]].pos;
		// s_charSetLengths[i]    = s_charSetLengthsData    + compactedSubKeySpacesIndex[compactedInto[i]].pos;
		// s_charSetLengthsFPM[i] = s_charSetLengthsFPMData + compactedSubKeySpacesIndex[compactedInto[i]].pos;

		m_sharedSize = sizeof(uint32_t) * (6*numSubKeySpaces + 3*numSubKeySpaceCharSets + (m_charSetLength+3)/4);
		checkCudaErrors(cudaMalloc((void**) &m_d_shared, m_sharedSize));
		uint64_t *shared = new uint64_t[(m_sharedSize + 7) / 8];
/ *
		uint32_t **s_charSetLengths        = new uint32_t*[numSubKeySpaces];// ***
		uint32_t **s_charSetLengthsFPM     = new uint32_t*[numSubKeySpaces];// ***
		uint8_t ***s_charSets              = new uint8_t**[numSubKeySpaces];// ***
		uint8_t  **s_charSetsSubKeySpace   = new uint8_t* [numSubKeySpaceCharSets];
		uint32_t  *s_charSetLengthsData    = new uint32_t [numSubKeySpaceCharSets];
		uint32_t  *s_charSetLengthsFPMData = new uint32_t [numSubKeySpaceCharSets];
		uint32_t  *s_passwordLengths       = new uint32_t [numSubKeySpaces];
		int       *s_numShortCutChars      = new int      [numSubKeySpaces];
		uint32_t  *s_divShortCut           = new uint32_t [numSubKeySpaces];
		uint8_t   *s_charSetsData          = new uint8_t  [m_charSetLength];
* /
		// uint32_t **s_charSetLengths        = new uint32_t*[numSubKeySpaces]; // ***
		// uint32_t **s_charSetLengthsFPM     = new uint32_t*[numSubKeySpaces]; // ***
		// uint8_t ***s_charSets              = new uint8_t**[numSubKeySpaces]; // ***
		uint32_t *p32 = (uint32_t*) shared;
		for (i = 0; i < numSubKeySpaces; i++)
		{
			p32[i] = compactedSubKeySpacesIndex[compactedInto[i]].pos;
		}

		// uint8_t  **s_charSetsSubKeySpace   = new uint8_t* [numSubKeySpaceCharSets];
		// uint32_t  *s_charSetLengthsData    = new uint32_t [numSubKeySpaceCharSets];
		// uint32_t  *s_charSetLengthsFPMData = new uint32_t [numSubKeySpaceCharSets];
		p32 += numSubKeySpaces;
		for (i = 0; i < numCompactedSubKeySpaces; i++)
		{
			uint32_t index = compactedSubKeySpaces[i] - compactedSubKeySpacesIndex;
			uint32_t size = compactedSubKeySpaces[i]->size;

			pos = compactedSubKeySpaces[i]->pos;
			for (j = 0; j < size; j++, pos++)
			{
				p32[pos                           ] = m_subKeySpacesSingleByte[index][j] - m_charSetSingleByte;
				p32[pos +   numSubKeySpaceCharSets] = m_charSetLengths[index][j];
				//p32[pos + 2*numSubKeySpaceCharSets] = m_charSetLengthsFPM[index][j];
				((float*) p32)[pos + 2*numSubKeySpaceCharSets] = 1.0f / m_charSetLengths[index][j];
			}
		}

		// uint32_t  *s_passwordLengths       = new uint32_t [numSubKeySpaces];
		// int       *s_numShortCutChars      = new int      [numSubKeySpaces];
		// uint32_t  *s_divShortCut           = new uint32_t [numSubKeySpaces];
		p32 += 3 * numSubKeySpaceCharSets;
		for (i = 0; i < numSubKeySpaces; i++)
		{
			p32[i                    ] = m_passwordLengths[i];
			p32[i + 1*numSubKeySpaces] = m_numShortCutChars[i];
			p32[i + 2*numSubKeySpaces] = m_divShortCut[i];
		}
*/
		checkCudaErrors(cudaMalloc((void**) &m_d_shared, 268 + 256));
		uint64_t *shared = new uint64_t[(268 + 256 + 7) / 8];

		// for pwLen = 4 to maxPwLen
		//   uint32_t s_shortCutShift[16]   = max(0, ceil(log2(62 ^ pwLen)) - 24);
		//   uint64_t s_shortCut[16]        = 62 ^ (pwLen - 3);
		//   float s_shortCutReciprocal[16] = 2 ^ s_shortCutShift[i] / s_shortCut[i];
		// float s_shortCutf[3]             = 62 ^ (pwLen - 3);
		uint64_t *s_shortCut           = (uint64_t*) shared;
		uint32_t *s_shortCutShift      = (uint32_t*) (s_shortCut           + 16);
		float    *s_shortCutReciprocal = (float*)    (s_shortCutShift      + 16);
		float    *s_shortCutf          = (float*)    (s_shortCutReciprocal + 16);
		uint8_t  *s_charSetsData       = (uint8_t*)  (s_shortCutf          +  3);
		uint32_t charSetLength = m_charSetLengths[0][0];
		uint64_t keySpace = charSetLength * charSetLength * charSetLength;
		uint64_t shortCut = 1;

		for (uint32_t i = 0; i < m_maxPwLen - 3; i++)
		{
			shortCut *= charSetLength;
			keySpace *= charSetLength;
			s_shortCutShift[i] = 64 - countLeadingZeros(keySpace);
			if (s_shortCutShift[i] < 24)
				s_shortCutShift[i]  = 0;
			else
				s_shortCutShift[i] -= 24;
			s_shortCut[i]           = shortCut;
			s_shortCutReciprocal[i] = reciprocalRoundDown(s_shortCut[i]) * ((float) (UINT64_C(1) << s_shortCutShift[i]));
		}
		s_shortCutf[0] = (float) s_shortCut[0];
		s_shortCutf[1] = (float) s_shortCut[1];
		s_shortCutf[2] = (float) s_shortCut[2];
		for (uint32_t i = 0; i < m_charSetLength; i++)
		{
			s_charSetsData[i] = m_charSetSingleByte[i];
		}

		// copy data to d_shared
		checkCudaErrors(cudaMemcpy(m_d_shared, shared, 268 + 256, cudaMemcpyHostToDevice));
		checkCudaErrors(cudaDeviceSynchronize());

		float keySpaceReciprocal = reciprocalRoundDown(m_keySpace);
		uint32_t log2KeySpace = 64 - countLeadingZeros(m_keySpace);

		if (log2KeySpace < 41)
		{
			cudaMemcpyToSymbol("c_keySpace",                &m_keySpace,         sizeof(uint64_t));
			getLastCudaError("c_keySpace");
			cudaMemcpyToSymbol("c_keySpaceReciprocal",      &keySpaceReciprocal, sizeof(float));
			getLastCudaError("c_keySpaceReciprocal");

			uint64_t tmp64 = m_keySpace << (41 - log2KeySpace);
			keySpaceReciprocal *= UINT64_C(1) << (40 - (41 - log2KeySpace));

			cudaMemcpyToSymbol("c_keySpaceFirst",           &tmp64,              sizeof(uint64_t));
			getLastCudaError("c_keySpaceFirst");
			cudaMemcpyToSymbol("c_keySpaceReciprocalFirst", &keySpaceReciprocal, sizeof(float));
			getLastCudaError("c_keySpaceReciprocalFirst");
		}
		else
		{
			keySpaceReciprocal *= UINT64_C(1) << 40;
			cudaMemcpyToSymbol("c_keySpace",                &m_keySpace,         sizeof(uint64_t));
			getLastCudaError("c_keySpace");
			cudaMemcpyToSymbol("c_keySpaceReciprocal",      &keySpaceReciprocal, sizeof(float));
			getLastCudaError("c_keySpaceReciprocal");
		}

		uint64_t keySpaceBoundaries[20] = {0};
		for (uint32_t i = 1; i <= m_numSubKeySpaces; i++)
		{
			keySpaceBoundaries[i] = m_keySpaceBoundaries[m_numSubKeySpaces - i + 1];
		}

		cudaMemcpyToSymbol("c_numSubKeySpaceCharSets", &m_numSubKeySpaces,                           sizeof(uint32_t));
		getLastCudaError("c_numSubKeySpaceCharSets");
		cudaMemcpyToSymbol("c_numSubKeySpaces",        &m_numSubKeySpaces,                           sizeof(uint32_t));
		getLastCudaError("c_numSubKeySpaces");
		cudaMemcpyToSymbol("c_charSetLength",          m_charSetLengths,                             sizeof(uint32_t));
		getLastCudaError("c_charSetLength");
		cudaMemcpyToSymbol("c_keySpaceBoundaries",     keySpaceBoundaries,                      20 * sizeof(uint64_t));
		getLastCudaError("c_keySpaceBoundaries");
		cudaMemcpyToSymbol("c_keySpaceBoundariesMax",  keySpaceBoundaries + m_numSubKeySpaces,       sizeof(uint64_t));
		getLastCudaError("c_keySpaceBoundariesMax");

		float c_charSetLen = reciprocalRoundDown(m_charSetLengths[0][0]);
		cudaMemcpyToSymbol("c_charSetReciprocal",      &c_charSetLen,                                sizeof(float));
		getLastCudaError("c_charSetReciprocal");
		c_charSetLen = (float) m_charSetLengths[0][0];
		cudaMemcpyToSymbol("c_charSetLen",             &c_charSetLen,                                sizeof(float));
		getLastCudaError("c_charSetLen");
		c_charSetLen = -c_charSetLen;
		cudaMemcpyToSymbol("c_ncharSetLen",            &c_charSetLen,                                sizeof(float));
		getLastCudaError("c_ncharSetLen");

		// clean up
/*		delete [] compactedSubKeySpaces;
		delete [] compactedInto;
		delete [] compactedSubKeySpacesIndex;
*/		delete [] shared;
	}
}

Redux *GpuDivSB::create(const keySpace &ks, const char *hashName, int gpu)
{
	GpuDivSB *ret;

	ret = new GpuDivSB(ks, hashName, gpu);
	if (!ret->isInited())
	{
		delete ret;
		ret = NULL;
	}
	return ret;
}

void GpuDivSB::doLinks(uint64_t *indexes, uint32_t chainPos, uint32_t tableIndex, uint32_t numLinks, uint32_t numIndexes)
{
	if (numLinks == 0)
	{
		return;
	}
	uint32_t startChainPos = chainPos + (tableIndex << 16);
	uint32_t endChainPos   = startChainPos + numLinks;
	uint32_t count         = m_optimalNumIndexes;
	uint32_t blocks        = m_blocks;
	uint32_t threads       = m_threads;
	// Lazy, I'm just setting a max linksPerStep of 2000
	uint32_t linksPerStep  = (numLinks + ((numLinks + 1999) / 2000) - 1) / ((numLinks + 1999) / 2000);

	while (numIndexes != 0)
	{
		if (count > numIndexes)
		{
			count = numIndexes;
			threads = (((count + blocks - 1) / blocks) + 15) & ~15;
		}
		checkCudaErrors(cudaMemcpy(m_d_indexes, indexes,     count * sizeof(uint64_t), cudaMemcpyHostToDevice));
		for (startChainPos = chainPos + (tableIndex << 16); startChainPos + linksPerStep <= endChainPos; startChainPos += linksPerStep)
		{
			m_doLinks(blocks, threads, m_d_shared, m_d_indexes, startChainPos, startChainPos + linksPerStep, count);
			getLastCudaError("Failed kernel launch.");
		}
		if (startChainPos <= endChainPos)
		{
			m_doLinks(blocks, threads, m_d_shared, m_d_indexes, startChainPos, endChainPos, count);
			getLastCudaError("Failed kernel launch.");
		}
		checkCudaErrors(cudaMemcpy(indexes, m_d_indexes, count * sizeof(uint64_t), cudaMemcpyDeviceToHost));
		numIndexes -= count;
		indexes += count;
	}
	checkCudaErrors(cudaDeviceSynchronize());
}

void GpuDivSB::doPreWorkLinks(uint64_t *indexes, uint32_t firstChainPos, uint32_t tableIndex, uint32_t chainLength, uint32_t numIndexes, const uint32_t *hash)
{
/*	uint32_t endChainPos, firstEndChainPos;
	uint32_t count       = /*2 * * /m_optimalNumIndexes;
	uint32_t blocks        = m_blocks;
	uint32_t threads       = m_threads;

	firstChainPos += (tableIndex << 16);
	endChainPos = firstChainPos + chainLength;
	checkCudaErrors(cudaMemcpy(m_d_hash, hash, ((m_hashLen + 3) / 4) * sizeof(uint32_t), cudaMemcpyHostToDevice));
	while (numIndexes != 0)
	{
		if (count > numIndexes)
		{
			count = numIndexes;
			threads = (((count + blocks - 1) / blocks) + 15) & ~15;
		}
		for (;;)
		{
			m_doPreWorkLinks(blocks, threads, m_d_shared, m_d_indexes, firstChainPos, firstEndChainPos, endChainPos, count, m_d_hash, true);
		}
		checkCudaErrors(cudaMemcpy(indexes, m_d_indexes, count * sizeof(uint64_t), cudaMemcpyDeviceToHost));
		numIndexes -= count;
		indexes += count;
	}
	checkCudaErrors(cudaDeviceSynchronize());*/
}

uint32_t GpuDivSB::falseAlarmChecking(uint64_t *indexes, const uint32_t *hash, const uint32_t *lastChainPoses, uint32_t tableIndex, uint32_t numIndexes)
{
/*	uint32_t ret;
	uint32_t count = 2 * m_optimalNumIndexes;

	checkCudaErrors(cudaMemcpy(m_d_hash, hash, ((m_hashLen + 3) / 4) * sizeof(uint32_t), cudaMemcpyHostToDevice));
	while (numIndexes != 0)
	{
		if (count > numIndexes)
		{
			count = numIndexes;
		}
		checkCudaErrors(cudaMemcpy(m_d_indexes,        indexes,        count * sizeof(uint64_t), cudaMemcpyHostToDevice));
		checkCudaErrors(cudaMemcpy(m_d_lastChainPoses, lastChainPoses, count * sizeof(uint64_t), cudaMemcpyHostToDevice));
		ret = m_falseAlarmChecking(m_blocks, m_threads, m_d_shared, m_d_indexes, m_d_hash, m_d_lastChainPoses, tableIndex, count);
		checkCudaErrors(cudaMemcpy(indexes,            m_d_indexes,    count * sizeof(uint64_t), cudaMemcpyDeviceToHost));
		if (ret)
		{
			break;
		}
	}
	checkCudaErrors(cudaDeviceSynchronize());
	return ret;*/
	return 0;
}

void GpuDivSB::cleanUp()
{
	if (m_d_indexes != NULL)
	{
		checkCudaErrors(cudaFree(m_d_indexes));
		checkCudaErrors(cudaFree(m_d_lastChainPoses));
		checkCudaErrors(cudaFree(m_d_hash));
		checkCudaErrors(cudaFree(m_d_shared));
		checkCudaErrors(cudaDeviceReset());
		m_d_indexes = NULL;
	}

	DivBase::cleanUp();
	ReduxSingleByte::cleanUp();
	ReduxBruteForce::cleanUp();
}
/*
cudaHostAlloc( ... cudaHostAllocPortable);
cudaHostRegister( ... cudaHostRegisterPortable);


cudaStream_t stream[2];
float* hostPtr;

// create
for (int i = 0; i < 2; ++i)
{
	cudaStreamCreate(&stream[i]);
}
cudaMallocHost(&hostPtr, 2 * size);

// use
for (int i = 0; i < 2; ++i)
{
	cudaMemcpyAsync                     (inputDevPtr  + i * size, hostPtr      + i * size, size, cudaMemcpyHostToDevice, stream[i]);
	MyKernel<<<100, 512, 0, stream[i]>>>(outputDevPtr + i * size, inputDevPtr  + i * size, size);
	cudaMemcpyAsync                     (hostPtr      + i * size, outputDevPtr + i * size, size, cudaMemcpyDeviceToHost, stream[i]);
}

// clean up
for (int i = 0; i < 2; ++i)
{
	cudaStreamDestroy(stream[i]);
}
*/
