/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "hashfactory.h"

using namespace std;

list<hashInfo*> *HashFactory::s_hashes = NULL;

uint32_t HashFactory::getHashCharacterEncoding(const char *name)
{
	list<hashInfo*>::iterator it;

	if (HashFactory::s_hashes != NULL)
	{
		for (it = s_hashes->begin(); it != s_hashes->end(); ++it)
		{
			if ((*it)->name == name)
			{
				return (*it)->hash.characterEncoding;
			}
		}
	}
	return 0;
}

uint32_t HashFactory::getHashInstance(const char *name, uint32_t minPw, uint32_t maxPw, hashAlgorithm &hash)
{
	list<hashInfo*>::iterator it;

	if (HashFactory::s_hashes != NULL)
	{
		for (it = s_hashes->begin(); it != s_hashes->end(); ++it)
		{
			if ((*it)->name == name && (*it)->minPw <= minPw && (*it)->maxPw >= maxPw)
			{
				hash = (*it)->hash;
				return 0;
			}
		}
	}
	hash.func               = NULL;
	hash.len                = 0;
	hash.hashesPerCall      = 0;
	hash.characterEncoding  = 0;
	hash.numPwBlocks        = 0;
	hash.numExtraHashBlocks = 0;
	hash.pwByteSkip         = 0;
	return 1;
}

uint32_t HashFactory::registerHash(const char *name, const char *comment, uint32_t minPw, uint32_t maxPw, HASH_FUNC hashFunc, uint32_t hashLen, uint32_t hashesPerCall, uint32_t characterEncoding, uint32_t numPwBlocks, uint32_t numExtraHashBlocks, uint32_t pwByteSkip)
{
	hashInfo *hi = new hashInfo;

	if ((numPwBlocks + 1) * ((hashesPerCall + 3) / 4) > MAX_PW_SIZE)
	{
		reportError("Exceeded maximum size of password (%u > %u).", (numPwBlocks + 1) * ((hashesPerCall + 3) / 4), MAX_PW_SIZE);
		return 0;
	}
	if (((hashLen + 3) / 4 + numExtraHashBlocks) * ((hashesPerCall + 3) / 4) > MAX_HASH_SIZE)
	{
		reportError("Exceeded maximum size of hash (%u > %u).", ((hashLen + 3) / 4 + numExtraHashBlocks) * ((hashesPerCall + 3) / 4), MAX_HASH_SIZE);
		return 0;
	}
	if (numPwBlocks > MAX_PW_BLOCKS)
	{
		reportError("Exceeded maximum number of password blocks (%u > %u).", numPwBlocks, MAX_PW_BLOCKS);
		return 0;
	}
	hi->hash.func               = hashFunc;
	hi->hash.len                = hashLen;
	hi->hash.hashesPerCall      = hashesPerCall;
	hi->hash.characterEncoding  = characterEncoding;
	hi->hash.numPwBlocks        = numPwBlocks;
	hi->hash.numExtraHashBlocks = numExtraHashBlocks;
	hi->hash.pwByteSkip         = pwByteSkip;
	hi->name                    = name;
	hi->comment                 = comment;
	hi->minPw                   = minPw;
	hi->maxPw                   = maxPw;
	if (HashFactory::s_hashes == NULL)
	{
		HashFactory::s_hashes = new list<hashInfo*>;
	}
	s_hashes->push_back(hi);
	return 1;
}

void HashFactory::printHashes()
{
	list<hashInfo*>::iterator it;
	char *spaces;
	uint32_t nameMaxLen = 0, commentMaxLen = 0, maxLen;

	if (HashFactory::s_hashes != NULL)
	{
		for (it = s_hashes->begin(); it != s_hashes->end(); ++it)
		{
			if (nameMaxLen < (*it)->name.length())
				nameMaxLen = (*it)->name.length();
			if (commentMaxLen < (*it)->comment.length())
				commentMaxLen = (*it)->comment.length();
		}
		maxLen = nameMaxLen > commentMaxLen ? nameMaxLen : commentMaxLen;
		spaces = new char[maxLen + 1];
		for (uint32_t a = 0; a < maxLen; a++)
		{
			spaces[a] = ' ';
		}
		spaces[maxLen] = 0;
		for (it = s_hashes->begin(); it != s_hashes->end(); ++it)
		{
			printf(" * %s%s [%s]%s (lengths %d - %d)\n",
				(*it)->name.c_str(),    spaces + (nameMaxLen    - (*it)->name.length()),
				(*it)->comment.c_str(), spaces + (commentMaxLen - (*it)->comment.length()), (*it)->minPw, (*it)->maxPw);
		}
		delete [] spaces;
	}
	printf("\n");
}
