/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "keyspace.h"
#include <map>
#include <string>
#include <fstream>

using namespace std;

map<string,characterSet> loadCharSetFile();
string getCharSetName(const char **str, uint32_t &min, uint32_t &max);
void readLineTrim(ifstream &fin, string &line);
string hex2bin(const string &hexStr);
void loadKeySpaceFile(keySpace &ks, const char *fileName);

map<string,characterSet> loadCharSetFile()
{
	static map<string,characterSet> charSets;
	ifstream fin;

	if (charSets.size() != 0)
	{
		return charSets;
	}
	fin.open("charset.txt", ifstream::in);
	if (!fin.good())
	{
		char fileName[256];
		int ret = GET_EXE_FULL_NAME(fileName, 256);

		if (ret > 0)
		{
			for (; ret >= 0; ret--)
			{
				if (fileName[ret] == DIRECTORY_SEPERATOR_CHAR)
				{
					break;
				}
			}
			if (ret > 0 && ret < 256 - 12)
			{
				strcpy(fileName + ret + 1, "charset.txt");
				fin.open(fileName, ifstream::in);
			}
		}
	}

	if (fin.good())
	{
		string line, charSetName;
		uint32_t i, equals;
		bool badCharSet = false;

		// Read character sets
		while (!fin.eof())
		{
			// Read line
			readLineTrim(fin, line);
			if ((!fin.good() && !fin.eof()) || // Unable to read line from key space file
			    !fin.is_open())                // Invalid key space file
			{
				badCharSet = true;
				break;
			}
			if (line.length() == 0 || line[0] == '#')
			{
				continue;
			}

			// if (regex(line, "^([a-zA-Z\\-]+)[ \t]*=[ \t]*\\[(.*)\\]$", out) == 0)
			// {
			//     badCharSet = true;
			//     break;
			// }
			// if (charSets.find(out[0]) == charSets.end())
			//     charSets[out[0]] = createCharacterSet(out[1].c_str());
			if (line[line.length() - 1] == ']')
			{
				// line = "...]" therefore I don't need to check if i < line.length()
				for (i = 0; ((line[i] >= 'a' && line[i] <= 'z') ||
								line[i] == '-' ||
								(line[i] >= '0' && line[i] <= '9') ||
								(line[i] >= 'A' && line[i] <= 'Z')); i++)
				{
					if (line[i] >= 'A' && line[i] <= 'Z')
					{
						line[i] = line[i] + ('a' - 'A');
					}
				}
				charSetName = line.substr(0, i);

				// Ignore white space and count '='
				equals = 0;
				while (line[i] == ' ' || line[i] == '\t' || line[i] == '=')
				{
					i++;
					if (line[i] == '=')
					{
						equals++;
					}
				}

				if (equals == 1 && line[i] == '[')
				{
					if (charSets.find(charSetName) == charSets.end())
					{
						charSets[charSetName] = createCharacterSet(line.substr(i + 1, line.length() - i - 2).c_str());
					}
				}
				else
				{
					badCharSet = true;
					break;
				}
			}
			else
			{
				badCharSet = true;
				break;
			}
		}
		if (badCharSet)
		{
			charSets.clear();
		}
	}
	else
	{
		// Load defaults
		charSets["a-z"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz");
		charSets["all"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/");
		charSets["all-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["all-en"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/");
		charSets["all-space"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ");
		charSets["all-space-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/\xC6\xD8\xC5\xC4\xD6\xDC ");
		charSets["alpha"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
		charSets["alpha-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["alpha-num-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["alpha-num-space-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\xC6\xD8\xC5\xC4\xD6\xDC ");
		charSets["alpha-num-sym-14-space-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=\xC6\xD8\xC5\xC4\xD6\xDC ");
		charSets["alpha-num-symbol14"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=");
		charSets["alpha-num-symbol14-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["alpha-numeric"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
		charSets["alpha-numeric-space"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 ");
		charSets["alpha-numeric-symbol-14-space"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+= ");
		charSets["alpha-numeric-symbol14"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=");
		charSets["alpha-numeric-symbol14-space"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+= ");
		charSets["alpha-numeric-symbol32-space"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ");
		charSets["alpha-space"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ ");
		charSets["alpha-space-dg"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xC4\xD6\xDC ");
		charSets["ascii-32-65-123-4"] = createCharacterSet(" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`{|}~");
		charSets["ascii-32-95"] = createCharacterSet(" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~");
		charSets["ascii-97-26-48-10"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789");
		charSets["byte"] = createCharacterSet("");
		charSets["byte-dg"] = createCharacterSet("");
		charSets["dan-ger-lower-alpha"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF");
		charSets["dan-ger-lower-alpha-numeric"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF");
		charSets["dan-ger-mix-alph-num-sym12"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xDF\xC4\xD6\xDC~!@#$^&*().?");
		charSets["dan-ger-mix-alpha"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xE4\xF6\xFC\xDF\xC4\xD6\xDC\xE6\xF8\xE5");
		charSets["dan-ger-upper-alpha"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xDF\xC4\xD6\xDC");
		charSets["dan-ger-upper-alpha-numeric"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xDF\xC4\xD6\xDC" "0123456789");
		charSets["dan-gerall-space-lm"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\xC6\xD8\xC5\xDF\xC4\xD6\xDC!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ");
		charSets["dg-alpha-numeric"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["dg-loweralpha-numeric"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789\xE6\xF8\xE5\xE4\xF6\xFC\xDF");
		charSets["lm-frt-cp437"] = createCharacterSet(" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`{|}~\x80\x8E\x8F\x90\x92\x99\x9A\x9B\x9C\x9D\x9E\x9F\xA5\xE0\xE1\xE2\xE3\xE4\xE6\xE7\xE8\xE9\xEA\xEB\xEE");
		charSets["lm-frt-cp437-850"] = createCharacterSet(" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`{|}~\x80\x8E\x8F\x90\x92\x99\x9A\x9B\x9C\x9D\x9E\x9F\xA5\xB5\xB6\xB7\xBD\xBE\xC7\xCF\xD1\xD2\xD3\xD4\xD5\xD6\xD7\xD8\xDE\xE0\xE1\xE2\xE3\xE4\xE5\xE6\xE7\xE8\xE9\xEA\xEB\xED\xEE\xEF");
		charSets["lm-frt-cp850"] = createCharacterSet(" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`{|}~\x80\x8E\x8F\x90\x92\x99\x9A\x9C\x9D\x9F\xA5\xB5\xB6\xB7\xBD\xBE\xC7\xCF\xD1\xD2\xD3\xD4\xD5\xD6\xD7\xD8\xDE\xE0\xE1\xE2\xE3\xE5\xE6\xE8\xE9\xEA\xEB\xED\xEF");
		charSets["loweralpha"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz");
		charSets["loweralpha-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF");
		charSets["loweralpha-num-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789\xE6\xF8\xE5\xE4\xF6\xFC\xDF");
		charSets["loweralpha-num-all-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE4\xF6\xFC\xDF" "0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/\xE6\xF8\xE5");
		charSets["loweralpha-num-all-space-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE4\xF6\xFC\xDF" "0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/\xE6\xF8\xE5 ");
		charSets["loweralpha-num-space-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789\xE6\xF8\xE5\xE4\xF6\xFC\xDF ");
		charSets["loweralpha-num-symbol14-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE4\xF6\xFC\xDF" "0123456789!@#$%^&*()-_+=\"\xE6\xF8\xE5");
		charSets["loweralpha-numeric"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789");
		charSets["loweralpha-numeric-all"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/");
		charSets["loweralpha-numeric-all-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ");
		charSets["loweralpha-numeric-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789 ");
		charSets["loweralpha-numeric-symbol14"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*()-_+=");
		charSets["loweralpha-numeric-symbol32-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ");
		charSets["loweralpha-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz ");
		charSets["loweralpha-space-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF ");
		charSets["mixalpha"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ");
		charSets["mixalpha-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF" "ABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["mixalpha-num-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF" "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["mixalpha-num-all-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF" "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["mixalpha-num-all-space-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF" "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/\xC6\xD8\xC5\xC4\xD6\xDC ");
		charSets["mixalpha-num-space-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF" "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789\xC6\xD8\xC5\xC4\xD6\xDC ");
		charSets["mixalpha-num-symbol14-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF" "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=\xC6\xD8\xC5\xC4\xD6\xDC");
		charSets["mixalpha-numeric"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
		charSets["mixalpha-numeric-all"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/");
		charSets["mixalpha-numeric-all-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ");
		charSets["mixalpha-numeric-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 ");
		charSets["mixalpha-numeric-symbol14"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=");
		charSets["mixalpha-numeric-symbol32-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()-_+=~`[]{}|\\:;\"'<>,.?/ ");
		charSets["mixalpha-space"] = createCharacterSet("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ ");
		charSets["mixalpha-space-dg"] = createCharacterSet("abcdefghijklmnopqrstuvwxyz\xE6\xF8\xE5\xE4\xF6\xFC\xDF" "ABCDEFGHIJKLMNOPQRSTUVWXYZ\xC6\xD8\xC5\xC4\xD6\xDC ");
		charSets["num-symbol"] = createCharacterSet("1234567890!@#$%^&*()-_+=~`");
		charSets["numeric"] = createCharacterSet("0123456789");
		charSets["numeric-space"] = createCharacterSet("0123456789 ");
		charSets["oracle-alpha-numeric-symbol3"] = createCharacterSet("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789#$_");
		charSets["rus"] = createCharacterSet("\xE9\xF6\xF3\xEA\xE5\xED\xE3\xF8\xF9\xE7\xF5\xFA\xF4\xFB\xE2\xE0\xEF\xF0\xEE\xEB\xE4\xE6\xFD\xFF\xF7\xF1\xEC\xE8\xF2\xFC\xE1\xFE");
		charSets["russian"] = createCharacterSet("\xE9\xF6\xF3\xEA\xE5\xED\xE3\xF8\xF9\xE7\xF5\xFA\xF4\xFB\xE2\xE0\xEF\xF0\xEE\xEB\xE4\xE6\xFD\xFF\xF7\xF1\xEC\xE8\xF2\xFC\xE1\xFE");
	}
	return charSets;
}

characterSet createCharacterSet(const char *charSet)
{
	characterSet ret;
	character    ch;

	ch.size = 1;
	if (*charSet == 0)
	{
		for (uint32_t i = 0; i < 256; i++)
		{
			ch.charData1[0] = i;
			ret.push_back(ch);
		}
	}
	else
	{
		for (; *charSet != 0; charSet++)
		{
			ch.charData1[0] = *charSet;
			ret.push_back(ch);
		}
	}
	return ret;
}

string getCharSetName(const char **str, uint32_t &min, uint32_t &max)
{
	string charSetName;
	const char *p;

	for (p = *str; *p != '#' && *p != 0; p++)
	{
		if (*p >= 'A' && *p <= 'Z')
		{
			charSetName += *p + ('a' - 'A');
		}
		else
		{
			charSetName += *p;
		}
	}
	if (charSetName.length() != 0 && *p == '#')
	{
		uint32_t tmp = 0;

		// get min
		for (p++; *p >= '0' && *p <= '9'; p++)
		{
			tmp = 10 * tmp + *p - '0';
		}
		if (*p == '-' && tmp <= 99)
		{
			min = tmp;

			// get max
			for (tmp = 0, p++; *p >= '0' && *p <= '9'; p++)
			{
				tmp = 10 * tmp + *p - '0';
			}

			if (tmp >= min && tmp <= 99 && tmp != 0)
			{
				max = tmp;
				*str = p;
				return charSetName;
			}
		}
	}
	return "";
}

void readLineTrim(ifstream &fin, string &line)
{
	uint32_t lastNonBlank = 0xffffffff;
	char ch;

	line = "";
	ch = fin.get();
	while (fin.good() && line.length() <= 65536)
	{
		if (ch == '\r')
		{
			if (fin.peek() == '\n')
			{
				fin.get();
			}
			break;
		}
		if (ch == '\n')
		{
			break;
		}
		if (ch == ' ' || ch == '\t')
		{
			if (lastNonBlank != 0xffffffff)
			{
				line += ch;
			}
		}
		else
		{
			line += ch;
			lastNonBlank = line.length();
		}
		ch = fin.get();
	}
	if (line.length() > 65536)
	{
		line = "";
		fin.close();
	}
	if (lastNonBlank != 0xffffffff && lastNonBlank != line.length())
	{
		line = line.substr(0, lastNonBlank);
	}
}

string hex2bin(const string &hexStr)
{
	string ret;
	uint32_t a, len = hexStr.length();
	char ch, ch2;

	if ((len & 1) != 0)
	{
		return "";
	}
	for (a = 0; a < len; a += 2)
	{
		ch = hexStr[a];
		if (ch >= '0' && ch <= '9')
		{
			ch2 = (ch - '0') << 4;
		}
		else if (ch >= 'A' && ch <= 'F')
		{
			ch2 = (ch - 'A' + 10) << 4;
		}
		else if (ch >= 'a' && ch <= 'f')
		{
			ch2 = (ch - 'a' + 10) << 4;
		}
		else
		{
			return "";
		}
		ch = hexStr[a + 1];
		if (ch >= '0' && ch <= '9')
		{
			ch2 |= ch - '0';
		}
		else if (ch >= 'A' && ch <= 'F')
		{
			ch2 |= ch - 'A' + 10;
		}
		else if (ch >= 'a' && ch <= 'f')
		{
			ch2 |= ch - 'a' + 10;
		}
		else
		{
			return "";
		}
		ret += ch2;
	}
	return ret;
}

void loadKeySpaceFile(keySpace &ks, const char *fileName)
{
	ifstream fin(fileName, ifstream::in);
	string line, name, charSet;
	map<string,characterSet> charSets;
	map<string,characterSet>::iterator charSetMapIt;
	subKeySpace curSubKeySpace;
	size_t pos, pos2;
	uint32_t len;
	bool multiCharNames = false;
	char ch[2] = {0, 0};

	ks.clear();
	if (!fin.good())
	{
		// Can't open key space file
		return;
	}

	// Read character sets
	while (1)
	{
		// Read line
		readLineTrim(fin, line);
		if (!fin.good() && !fin.eof())
		{
			// Unable to read line from key space file
			return;
		}
		if (line.length() == 0 || line[0] == '#')
		{
			if (fin.eof() || !fin.is_open())
			{
				// Invalid key space file
				return;
			}
			continue;
		}
		if (line[line.length() - 1] != ']')
		{
			break;
		}

		// Get character set name
		pos = line.find_first_of(" \t", 0, 2);
		if (pos == string::npos)
		{
			break;
		}
		name = line.substr(0, pos);

		// Get character set
		pos = line.find_first_not_of(" \t", pos, 2);
		if (pos == string::npos)
		{
			// This is not even possible... meh
			break;
		}
		if (line[pos] == '[')
		{
			charSet = line.substr(pos + 1, line.length() - pos - 2);
		}
		else if (line[pos] == 'h' && line[pos + 1] == '[')
		{
			charSet = hex2bin(line.substr(pos + 2, line.length() - pos - 3));
			if (charSet.length() == 0 && line.length() - pos - 3 != 0)
			{
				// Invalid character set
				return;
			}
		}
		else
		{
			break;
		}
		if (charSet.length() == 0)
		{
			// Empty character set
			return;
		}

		// Insert into map
		if (charSets.find(name) == charSets.end())
		{
			if (name.length() != 1)
			{
				multiCharNames = true;
			}
			charSets[name] = createCharacterSet(charSet.c_str());
		}
		else
		{
			// Duplicate character set name
			return;
		}
	}
	if (line[line.length() - 1] == ']')
	{
		// Invalid key space file
		return;
	}

	// Read sub key spaces
	do
	{
		if (multiCharNames)
		{
			pos = 0;
			while (pos != string::npos)
			{
				// Get character set name
				pos2 = line.find_first_of(" \t", pos, 2);
				if (pos2 == string::npos)
				{
					pos2 = line.length();
				}
				name = line.substr(pos, pos2 - pos);

				// Get character set
				charSetMapIt = charSets.find(name);
				if (charSetMapIt == charSets.end())
				{
					// Character set name not found
					ks.clear();
					return;
				}
				curSubKeySpace.push_back(charSetMapIt->second);

				if (pos2 == line.length())
				{
					break;
				}
				pos = line.find_first_not_of(" \t", pos2, 2);
			}
		}
		else
		{
			len = line.length();
			for (pos = 0; pos < len; pos++)
			{
				*ch = line[pos];
				if (*ch != ' ' && *ch != '\t')
				{
					charSetMapIt = charSets.find(ch);
					if (charSetMapIt == charSets.end())
					{
						// Character set name not found
						ks.clear();
						return;
					}
					curSubKeySpace.push_back(charSetMapIt->second);
				}
			}
		}
		ks.push_back(curSubKeySpace);
		curSubKeySpace.clear();

		// Read line
		do
		{
			readLineTrim(fin, line);
			if ((!fin.good() && !fin.eof()) || // Unable to read line from key space file
			    !fin.is_open())                // Invalid key space file
			{
				ks.clear();
				return;
			}
		} while ((line.length() == 0 || line[0] == '#') && fin.good());
	} while (line.length() != 0);
}

/*
 * fileName can be the full file name, just the key space part, or a key space file:
 * md5_loweralpha#1-7_0_10000x67108864_00.rt
 * md5_hybrid(loweralpha#7-7,numeric#1-3)#0-0_0_10000x67108864_00.rt
 * md5_hybrid2(loweralpha#7-7,numeric#1-3)#0-0_0_10000x67108864_00.rt
 * md5_hybrid3(omni6.txt)#0-0_0_10000x67108864_00.rt
 * -----
 * loweralpha#1-7
 * hybrid(loweralpha#7-7,numeric#1-3)#0-0
 * hybrid2(loweralpha#7-7,numeric#1-3)#0-0
 * hybrid3(omni6.txt)#0-0
 * -----
 * omni6.txt
 * 
 * Returns 0, 1, or 2 for the three formats (0 - full file name, 1 - just the key space part, 2 - key space file)
 */
uint32_t createKeySpace(keySpace &ks, const char *fileName)
{
	const char *originalFileName = fileName;
	map<string,characterSet> charSets;
	map<string,characterSet>::iterator it;
	subKeySpace curSubKeySpace;
	string charSetName;
	uint32_t hybridVersion = 0, ret = 0;

	charSets = loadCharSetFile();
	ks.clear();

	while (*fileName != 0 && *fileName != '_')
	{
		fileName++;
	}
	if (*fileName == '_')
	{
		fileName++;
	}
	else
	{
		// Reset fileName because it is not a full file name
		fileName = originalFileName;
		ret = 1;
	}
	if (strncmp(fileName, "hybrid", 6) == 0)
	{
		if (fileName[6] == '(')
		{
			fileName += 7;
			hybridVersion = 1;
		}
		else if (fileName[6] == '2' && fileName[7] == '(')
		{
			fileName += 8;
			hybridVersion = 2;
		}
		else if (fileName[6] == '3' && fileName[7] == '(')
		{
			fileName += 8;
			hybridVersion = 3;
			charSetName = fileName;
			uint32_t i = 0;
			while (fileName[i] != ')' && fileName[i] != 0)
			{
				i++;
			}
			if (fileName[i] == ')')
			{
				charSetName[i] = 0;
				loadKeySpaceFile(ks, charSetName.c_str());
			}
		}
	}
	if (hybridVersion != 3)
	{
		while (1)
		{
			uint32_t min, max;

			charSetName = getCharSetName(&fileName, min, max);
			if ((hybridVersion != 0 || (*fileName != '_' && *fileName != 0)) &&
				(hybridVersion == 0 || (*fileName != ',' && *fileName != ')')))
			{
				ks.clear();
				return ret;
			}

			it = charSets.find(charSetName);
			if (it == charSets.end())
			{
				ks.clear();
				return ret;
			}
			for (uint32_t i = 0; i < min; i++)
			{
				curSubKeySpace.push_back(it->second);
			}
			if (hybridVersion == 0 || *fileName == ')')
			{
				ks.push_back(curSubKeySpace);
				for (uint32_t i = min; i < max; i++)
				{
					curSubKeySpace.push_back(it->second);
					ks.push_back(curSubKeySpace);
				}
				break;
			}
			if (min != max)
			{
				ks.clear();
				return ret;
			}
			fileName++;
		}
	}
	if (ks.size() == 0)
	{
		loadKeySpaceFile(ks, originalFileName);
		ret = 2;
	}
	return ret;
}

bool operator==(const characterSet &l, const characterSet &r)
{
	uint32_t end = r.size();

	if (l.size() != end)
	{
		return false;
	}
	for (uint32_t a = 0; a < end; a++)
	{
		if (l[a].charData != r[a].charData || l[a].size != r[a].size)
		{
			return false;
		}
	}
	return true;
}

uint32_t find(const characterSet &hayStack, const characterSet &needle)
{
	uint32_t a, b, needleSize, end;

	needleSize = needle.size();
	end = hayStack.size() - needleSize;
	if (hayStack.size() < needleSize)
	{
		return 0xffffffff;
	}
	for (a = 0; a <= end; a++)
	{
		for (b = 0; b < needleSize; b++)
		{
			if (hayStack[a+b].charData != needle[b].charData || hayStack[a+b].size != needle[b].size)
			{
				break;
			}
		}
		if (b >= needleSize)
		{
			return a;
		}
	}
	return 0xffffffff;
}

character *strstr(/*const*/ character *hayStack, uint32_t hayStackSize, const characterSet &needle)
{
	uint32_t a, b, needleSize, end;

	needleSize = needle.size();
	end = hayStackSize - needleSize;
	if (hayStackSize < needleSize)
	{
		return NULL;
	}
	for (a = 0; a <= end; a++)
	{
		for (b = 0; b < needleSize; b++)
		{
			if (hayStack[a+b].charData != needle[b].charData || hayStack[a+b].size != needle[b].size)
			{
				break;
			}
		}
		if (b >= needleSize)
		{
			return hayStack + a;
		}
	}
	return NULL;
}

uint32_t getMaxLen(const keySpace &ks, uint32_t characterEncoding)
{
	keySpace::const_iterator     curSubKeySpace;
	subKeySpace::const_iterator  curCharSet;
	characterSet::const_iterator curChar;
	uint32_t ret = 0, size, charSize, maxCharSize;

	for (curSubKeySpace = ks.begin(); curSubKeySpace != ks.end(); ++curSubKeySpace)
	{
		size = 0;
		for (curCharSet = curSubKeySpace->begin(); curCharSet != curSubKeySpace->end(); ++curCharSet)
		{
			maxCharSize = 0;
			for (curChar = curCharSet->begin(); curChar != curCharSet->end(); ++curChar)
			{
				switch (curChar->size)
				{
					case 1:
						charSize = 1;
						if ((characterEncoding & CE_UTF16_LE) == CE_UTF16_LE || (characterEncoding & CE_UTF16_BE) == CE_UTF16_BE)
						{
							charSize++; // null pad to 2 bytes
						}
						else if ((characterEncoding & CE_UTF32_LE) == CE_UTF32_LE || (characterEncoding & CE_UTF32_BE) == CE_UTF32_BE)
						{
							charSize += 3; // null pad to 4 bytes
						}
						break;

					case 2:
						charSize = 2;
						if ((characterEncoding & CE_UTF32_LE) == CE_UTF32_LE || (characterEncoding & CE_UTF32_BE) == CE_UTF32_BE)
						{
							charSize += 2; // null pad to 4 bytes
						}
						break;

					case 3:
						charSize = 3;
						if ((characterEncoding & CE_ASCII_BIN8) != CE_ASCII_BIN8 && (characterEncoding & CE_UTF8) != CE_UTF8)
						{
							charSize++; // null pad to 4 bytes
						}
						break;

					case 4:
						charSize = 4;
				}
				if (maxCharSize < charSize)
				{
					maxCharSize = charSize;
				}
			}
			size += maxCharSize;
		}
		if (ret < size)
		{
			ret = size;
		}
	}
	return ret;
}

uint32_t getCharSizes(const keySpace &ks)
{
	keySpace::const_iterator     curSubKeySpace;
	subKeySpace::const_iterator  curCharSet;
	characterSet::const_iterator curChar;
	uint32_t ret = 0;

	for (curSubKeySpace = ks.begin(); curSubKeySpace != ks.end(); ++curSubKeySpace)
	{
		for (curCharSet = curSubKeySpace->begin(); curCharSet != curSubKeySpace->end(); ++curCharSet)
		{
			for (curChar = curCharSet->begin(); curChar != curCharSet->end(); ++curChar)
			{
				if (curChar->size - 1 > 3)
				{
					return 0xffffffff;
				}
				ret |= 1 << curChar->size;
			}
		}
	}
	return ret;
}
