/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "common.h"
#include "reduxfactory.h"
#include "writefactory.h"
#include "generationthread.h"
#include "keyspace.h"
#include "workqueue.h"
#include "atomic.h"
#include <stdlib.h>
#include <errno.h>

#ifdef _WIN32
	#include <string.h>
	#define USE_STR_TO_UL
	#define USE_STR_TO_ULL
	#define STR_TO_ULL(str,endPtr,radix) _strtoui64(str, endPtr, radix)
	#define STR_CMP_ICASE(a,b) strcmpi(a, b)
#else
	#include <strings.h>
	#define USE_STR_TO_UL
	#define USE_STR_TO_ULL
	#define STR_TO_ULL(str,endPtr,radix) strtoull(str, endPtr, radix)
	#define STR_CMP_ICASE(a,b) strcasecmp(a, b)
#endif

void usage(const char *runStr);
int validateInt  (const char *arg, uint32_t &out, bool required, const char *name, uint32_t rangeLo = 1, uint32_t rangeHi = 0, uint32_t defaultVal = 0, char endChar = 0, char **pos = NULL);
int validateInt64(const char *arg, uint64_t &out, bool required, const char *name, uint64_t rangeLo = 1, uint64_t rangeHi = 0, uint64_t defaultVal = 0);

int main(int argc, char *argv[])
{
	#define FILE_FORMAT        1
	#define HASH               2
	#define REDUX              3
	#define KEY_SPACE          4
	#define TABLE_INDEX        5
	#define REAL_CHAIN_LENGTH  6
	#define CHAINS             7
	#define FIRST_START_POINT  8
	#define OUTPUT_DIRECTORY   9
	#define FILE_NAME_COMMENT  10
	#define THREADS            11
	#define GPU_DEVICE_NUM     12

	// *** Get args ***
	uint64_t chains, firstStartPoint;
	char *reduxName;
	const char *fileFormat, *hashName, *keySpaceName;
	const char *outputDirectory, *fileNameComment;
	uint32_t chainLength, tableIndex, threads, gpuDeviceNums[16], gpuDevices = 0;

	threads = getNumCores();
	if (argc < THREADS)
	{
		usage(argv[0]);
		return 1;
	}
	if (STR_CMP_ICASE(argv[FILE_FORMAT], "rt") || STR_CMP_ICASE(argv[HASH], "md5") || STR_CMP_ICASE(argv[REDUX], "div"))
	{
		usage(argv[0]);
		return 1;
	}
	if (validateInt(argv[REAL_CHAIN_LENGTH], chainLength, true, "real-chain-length", 1, 0x80000000))
	{
		return 1;
	}
	if (validateInt(argv[TABLE_INDEX], tableIndex, true, "table-index"))
	{
		return 1;
	}
	if (validateInt64(argv[CHAINS], chains, true, "chains", 1, UINT64_MAX))
	{
		return 1;
	}
	if (validateInt64(argv[FIRST_START_POINT], firstStartPoint, true, "first-start-point"))
	{
		return 1;
	}
	if (argc > THREADS)
	{
		if (validateInt(argv[THREADS], threads, true, "threads", 0, 65536))
		{
			return 1;
		}
		gpuDevices = argc - GPU_DEVICE_NUM;
		if (gpuDevices > 16)
		{
			gpuDevices = 16;
		}
		for (uint32_t i = 0; i < gpuDevices; i++)
		{
			if (validateInt(argv[GPU_DEVICE_NUM + i], gpuDeviceNums[i], true, "gpu-device-num", 0, 65536))
			{
				return 1;
			}
		}
	}
	fileFormat      = argv[FILE_FORMAT];
	hashName        = argv[HASH];
	keySpaceName    = argv[KEY_SPACE];
	outputDirectory = argv[OUTPUT_DIRECTORY];
	fileNameComment = argv[FILE_NAME_COMMENT];

	// *** Init generation ***
	uint64_t lastStartPoint = firstStartPoint + chains - 1, buffer1Start, buffer2Start;
	uint64_t *data;
	rtChain *buffer1, *buffer2, *tmp;
	workQueueNode *node, *bufferNodeHead = NULL, *bufferNodeTail = NULL;
	Redux *redux;
	Write *write;
	generationThreadArgs *genArgs;
	GenerationThread *genThreads;
	WorkQueue inQueue;
	keySpace ks;
	Atomic startPointAtomic(firstStartPoint);

	// 0 - full file name, 1 - just the key space part, 2 - key space file
	uint32_t keySpaceNameFormat = createKeySpace(ks, keySpaceName);
	if (ks.size() == 0)
	{
		reportError("Invalid key space.");
		return 1;
	}

	reduxName = new char[strlen(argv[REDUX]) + 2 + 1];
	strcpy(reduxName, argv[REDUX]);
	// Append "h", "h2", "h3" to reduxName for hybrid, hybrid2, hybrid3
	if (keySpaceNameFormat == 2)
	{
		strcat(reduxName, "h3");
	}
	else
	{
		if (keySpaceNameFormat == 0)
		{
			while (*keySpaceName != '_')
			{
				keySpaceName++;
			}
			keySpaceName++;
		}
		if (strncmp(keySpaceName, "hybrid", 6) == 0)
		{
			if (keySpaceName[6] == '(')
			{
				strcat(reduxName, "h");
			}
			else if (keySpaceName[6] == '2' && keySpaceName[7] == '(')
			{
				strcat(reduxName, "h2");
			}
			else if (keySpaceName[6] == '3' && keySpaceName[7] == '(')
			{
				strcat(reduxName, "h3");
			}
		}
	}
	redux = ReduxFactory::getReduxInstance(reduxName, ks, 0, hashName);
	if (redux == NULL)
	{
		reportError("No redux found.");
		return 1;
	}
	if (lastStartPoint < firstStartPoint || lastStartPoint == UINT64_MAX || lastStartPoint >= redux->getKeySpace())
	{
		reportError("Start point overflow.");
		delete redux;
		return 1;
	}
	delete redux;
	write = WriteFactory::getWriteInstance(fileFormat, outputDirectory, fileNameComment, reduxName, hashName, NULL, keySpaceName, chainLength, tableIndex);
	if (write == NULL)
	{
		reportError("No writer found.");
		return 1;
	}
	// Init threads
	if (threads + gpuDevices == 0)
	{
		threads = 1;
	}
	genArgs = new generationThreadArgs[threads + gpuDevices];
	genThreads = new GenerationThread[threads + gpuDevices];

	TIMER_TYPE s, e;
	TIMER_FUNC(s);
	for (uint32_t i = 0; i < threads; i++)
	{
		genArgs[i].lastStartPoint = lastStartPoint;
		genArgs[i].startPoint     = &startPointAtomic;
		genArgs[i].outWorkQueue   = &inQueue;
		genArgs[i].ks             = &ks;
		genArgs[i].hashName       = hashName;
		genArgs[i].reduxName      = reduxName;
		genArgs[i].chainLength    = chainLength;
		genArgs[i].tableIndex     = tableIndex;
		genArgs[i].gpuId          = 0;
		genThreads[i].start(genArgs + i);
	}
	for (uint32_t i = 0; i < gpuDevices; i++)
	{
		genArgs[threads + i].lastStartPoint = lastStartPoint;
		genArgs[threads + i].startPoint     = &startPointAtomic;
		genArgs[threads + i].outWorkQueue   = &inQueue;
		genArgs[threads + i].ks             = &ks;
		genArgs[threads + i].hashName       = hashName;
		genArgs[threads + i].reduxName      = reduxName;
		genArgs[threads + i].chainLength    = chainLength;
		genArgs[threads + i].tableIndex     = tableIndex;
		genArgs[threads + i].gpuId          = gpuDeviceNums[i] + 1;
		genThreads[threads + i].start(genArgs + (threads + i));
	}
	buffer1 = new rtChain[CHAIN_BUFFER_LENGTH];
	buffer2 = new rtChain[CHAIN_BUFFER_LENGTH];
	buffer1Start = firstStartPoint;
	buffer2Start = firstStartPoint + CHAIN_BUFFER_LENGTH;

	// *** Generation ***
	uint64_t startPoint;
	uint32_t numPws, count, buffer1Count = 0, buffer2Count = 0;
	while (chains != 0)
	{
		node = inQueue.pop();
		assert(node != NULL);
		numPws     = (uint32_t) ((uint64_t*) (node->data))[0];
		startPoint =            ((uint64_t*) (node->data))[1];
		data       =            ((uint64_t*) (node->data)) + 2;
		count      = numPws;

		// does data start in buffer1
		if (startPoint < buffer2Start)
		{
			tmp = buffer1 + (startPoint - buffer1Start);
			if (startPoint + numPws > buffer2Start)
			{
				// spans buffer1 and buffer2
				count = (uint32_t) (buffer2Start - startPoint);
			}

			// copy to buffer1
			for (uint32_t i = 0; i < count; i++)
			{
				tmp[i].startPoint = startPoint + i;
				tmp[i].endPoint   = data[i];
			}
			buffer1Count += count;

			// copy to buffer2
			if (numPws != count)
			{
				data += count;
				startPoint += count;
				for (uint32_t i = 0; i < numPws - count; i++)
				{
					buffer2[i].startPoint = startPoint + i;
					buffer2[i].endPoint   = data[i];
				}
				buffer2Count += numPws - count;
			}
			chains -= numPws;

			// return the node
			node->returnQueue->push(node);

			// is buffer1 full
			assert(buffer1Count <= CHAIN_BUFFER_LENGTH);
			if (buffer1Count == CHAIN_BUFFER_LENGTH)
			{
				if (write->write(buffer1, CHAIN_BUFFER_LENGTH) != CHAIN_BUFFER_LENGTH)
				{
					reportError("Failed to write all of the data.");
					return 1;
				}

				// swap buffer1 and buffer2
				buffer1Count = buffer2Count;
				buffer2Count = 0;
				tmp     = buffer1;
				buffer1 = buffer2;
				buffer2 = tmp;

				// process buffered nodes
				if (bufferNodeHead != NULL)
				{
					inQueue.pushQueue(bufferNodeHead, bufferNodeTail);
					bufferNodeHead = NULL;
					bufferNodeTail = NULL;
				}
			}
		}
		// does data start in buffer2
		else if (startPoint < buffer2Start + CHAIN_BUFFER_LENGTH)
		{
			if (startPoint + numPws > buffer2Start + CHAIN_BUFFER_LENGTH)
			{
				// spans buffer2 and "buffer3"
				// push node into buffered nodes
				if (bufferNodeHead != NULL)
				{
					node->next = NULL;
					bufferNodeTail->next = node;
					bufferNodeTail = node;
				}
				else
				{
					node->next = NULL;
					bufferNodeHead = bufferNodeTail = node;
				}
			}
			else
			{
				// copy to buffer2
				tmp = buffer2 + (startPoint - buffer2Start);
				for (uint32_t i = 0; i < numPws; i++)
				{
					tmp[i].startPoint = startPoint + i;
					tmp[i].endPoint   = data[i];
				}
				buffer2Count += numPws;
				chains -= numPws;

				// return the node
				node->returnQueue->push(node);
			}
		}
		// does data start in "buffer3" or later
		else
		{
			// push node into buffered nodes
			count = 0;
			if (bufferNodeHead != NULL)
			{
				node->next = NULL;
				bufferNodeTail->next = node;
				bufferNodeTail = node;
			}
			else
			{
				node->next = NULL;
				bufferNodeHead = bufferNodeTail = node;
			}
		}
	}
	if (buffer1Count != 0)
	{
		if (write->write(buffer1, buffer1Count) != buffer1Count)
		{
			reportError("Failed to write all of the data.");
			return 1;
		}
	}
	if (buffer2Count != 0)
	{
		if (write->write(buffer2, buffer2Count) != buffer2Count)
		{
			reportError("Failed to write all of the data.");
			return 1;
		}
	}
	TIMER_FUNC(e);
	if (TIMER_DIFF(s, e) < 100.0)
	{
		printf("%" PRIu64 " links in %0.3f seconds @ %0.2f MLinks/sec\n", (lastStartPoint - firstStartPoint + 1) * chainLength, TIMER_DIFF(s, e), ((lastStartPoint - firstStartPoint + 1) * chainLength) / TIMER_DIFF(s, e) / 1000000.0);
	}
	else if (TIMER_DIFF(s, e) < 1000.0)
	{
		printf("%" PRIu64 " links in %0.2f seconds @ %0.2f MLinks/sec\n", (lastStartPoint - firstStartPoint + 1) * chainLength, TIMER_DIFF(s, e), ((lastStartPoint - firstStartPoint + 1) * chainLength) / TIMER_DIFF(s, e) / 1000000.0);
	}
	else if (TIMER_DIFF(s, e) < 10000.0)
	{
		printf("%" PRIu64 " links in %0.1f seconds @ %0.2f MLinks/sec\n", (lastStartPoint - firstStartPoint + 1) * chainLength, TIMER_DIFF(s, e), ((lastStartPoint - firstStartPoint + 1) * chainLength) / TIMER_DIFF(s, e) / 1000000.0);
	}
	else
	{
		printf("%" PRIu64 " links in %0.0f seconds @ %0.2f MLinks/sec\n", (lastStartPoint - firstStartPoint + 1) * chainLength, TIMER_DIFF(s, e), ((lastStartPoint - firstStartPoint + 1) * chainLength) / TIMER_DIFF(s, e) / 1000000.0);
	}

	// Clean up
	delete write;
	delete [] genArgs;
	delete [] genThreads;
	delete [] buffer1;
	delete [] buffer2;
	return 0;
}

int validateInt(const char *arg, uint32_t &out, bool required, const char *name, uint32_t rangeLo, uint32_t rangeHi, uint32_t defaultVal, char endChar, char **pos)
{
#ifdef USE_STR_TO_UL
	char *endPtr;
#endif
	uint32_t val = 0;
	int ret = 0;

	if (arg == NULL)
	{
		out = defaultVal;
		if (required)
		{
			reportError("Missing argument %s.", name);
			ret = 1;
		}
	}
	else
	{
#ifdef USE_STR_TO_UL
		errno = 0;
		out = val = (uint32_t) strtoul(arg, &endPtr, 0);
		if ((*endPtr != endChar && *endPtr != 0) || errno != 0)
		{
			ret = 1;
		}
		if (pos != NULL)
		{
			*pos = endPtr;
		}
#else
		if (*arg == '0')
		{
			if (arg[1] == 'x')
			{
				for (arg += 2; *arg != endChar || *arg != 0; arg++)
				{
					if (*arg >= '0' && *arg <= '9')
					{
						val = 16 * val + (uint32_t) (*arg - '0');
					}
					else if (*arg >= 'a' && *arg <= 'f')
					{
						val = 16 * val + (uint32_t) (*arg - 'a') + 10;
					}
					else if (*arg >= 'A' && *arg <= 'F')
					{
						val = 16 * val + (uint32_t) (*arg - 'A') + 10;
					}
					else
					{
						ret = 1;
						break;
					}
				}
			}
			else
			{
				for (arg++; *arg != endChar || *arg != 0; arg++)
				{
					if (*arg >= '0' && *arg <= '7')
					{
						val = 8 * val + (uint32_t) (*arg - '0');
					}
					else
					{
						ret = 1;
						break;
					}
				}
			}
		}
		else
		{
			for (; *arg != endChar || *arg != 0; arg++)
			{
				if (*arg >= '0' && *arg <= '9')
				{
					val = 10 * val + (uint32_t) (*arg - '0');
				}
				else
				{
					ret = 1;
					break;
				}
			}
		}
		if (pos != NULL)
		{
			*pos = arg;
		}
#endif
		if (rangeLo > rangeHi)
		{
			if (ret)
			{
				reportError("Invalid argument %s must be an integer.", name);
			}
		}
		else if (ret || val < rangeLo || val > rangeHi)
		{
			reportError("Invalid argument %s must be an integer between %u and %u.", name, rangeLo, rangeHi);
			ret = 1;
		}
	}
	return ret;
}

int validateInt64(const char *arg, uint64_t &out, bool required, const char *name, uint64_t rangeLo, uint64_t rangeHi, uint64_t defaultVal)
{
	uint64_t val = 0;
#ifdef USE_STR_TO_ULL
	char *endPtr;
#endif
	int ret = 0;

	if (arg == NULL)
	{
		out = defaultVal;
		if (required)
		{
			reportError("Missing argument %s.", name);
			ret = 1;
		}
	}
	else
	{
#ifdef USE_STR_TO_ULL
		errno = 0;
		out = val = (uint64_t) STR_TO_ULL(arg, &endPtr, 0);
		if (*endPtr != 0 || errno != 0)
		{
			ret = 1;
		}
#else
		if (*arg == '0')
		{
			if (arg[1] == 'x')
			{
				for (arg += 2; *arg != 0; arg++)
				{
					if (*arg >= '0' && *arg <= '9')
					{
						val = 16 * val + (uint32_t) (*arg - '0');
					}
					else if (*arg >= 'a' && *arg <= 'f')
					{
						val = 16 * val + (uint32_t) (*arg - 'a') + 10;
					}
					else if (*arg >= 'A' && *arg <= 'F')
					{
						val = 16 * val + (uint32_t) (*arg - 'A') + 10;
					}
					else
					{
						ret = 1;
						break;
					}
				}
			}
			else
			{
				for (arg++; *arg != 0; arg++)
				{
					if (*arg >= '0' && *arg <= '7')
					{
						val = 8 * val + (uint32_t) (*arg - '0');
					}
					else
					{
						ret = 1;
						break;
					}
				}
			}
		}
		else
		{
			for (; *arg != 0; arg++)
			{
				if (*arg >= '0' && *arg <= '9')
				{
					val = 10 * val + (uint32_t) (*arg - '0');
				}
				else
				{
					ret = 1;
					break;
				}
			}
		}
#endif
		if (rangeLo >= rangeHi)
		{
			if (ret)
			{
				reportError("Invalid argument %s must be an integer.", name);
			}
		}
		else if (ret || val < rangeLo || val > rangeHi)
		{
			reportError("Invalid argument %s must be an integer between %"PRIu64" and %"PRIu64".", name, rangeLo, rangeHi);
			ret = 1;
		}
	}
	return ret;
}

void usage(const char *runStr)
{
	printf("\n"
		"                     **** Advanced RT Generator v0.1a ****\n\n"
		"Generates rainbow tables with advanced options (not really this is a umm\n"
		"pre-release).\n\n"
		"%s file-format hash redux key-space table-index\n"
		"   real-chain-length chains first-start-point output-directory\n"
		"   file-name-comment [threads [gpu-device-num [...]]]\n\n"
		"file-format:\n"
		"  Can be one of the following:\n"
		"  rt\n\n"
		"hash:\n"
		"  Can be one of the following:\n"
		"  md5\n\n"
		"redux:\n"
		"  Can be one of the following:\n"
		"  div\n\n"
		"key-space:\n"
		"  Key space file or string:\n"
		"    loweralpha#1-7\n"
		"    hybrid(loweralpha#7-7,numeric#1-3)#0-0\n"
		"    hybrid2(loweralpha#7-7,numeric#1-3)#0-0\n"
		"    hybrid3(omni6.txt)#0-0\n"
		"    omni6.txt\n\n"
		"real-chain-length:\n"
		"  The cake is a lie. The chain length you are familiar with is actually one\n"
		"  less. So if you want to generate something like\n"
		"  \"md5_loweralpha#1-7_0_10000x67108864_00.rt\" the real-chain-length is 9999\n"
		"  not 10000.\n\n"
		"table-index:\n"
		"  The table index.\n\n"
		"chains:\n"
		"  Number of chains.\n\n"
		"first-start-point:\n"
		"  The first start point.\n\n"
		"output-directory:\n"
		"  Where the rainbow tables will be saved.\n\n"
		"file-name-comment:\n"
		"  With *.rt, *.rtc, *.rti, and *.rti2 (2.0) formats this is the last part of\n"
		"  the file name. With *.grt, *.rti2 (2.1), *.irt, and *.dirt formats this is\n"
		"  the full file name. The file-name-comment is concatenated with the file\n"
		"  number \"_000\":\n"
		"  \"md5_loweralpha#1-7_0_10000x67108864_file-name-comment_000.rt\"\n"
		"  \"file-name-comment_000.dirt\"\n\n"
		"threads:\n"
		"  Number of CPU threads.\n\n"
		"gpu-device-num:\n"
		"  CUDA GPU device number (starts at 0).\n\n"
		"Copyright (c) 2012 Steve Thomas <steve AT tobtu DOT com>\n"
		"  This is free software: you can redistribute it and/or modify it under the\n"
		"  terms of the GNU General Public License as published by the Free Software\n"
		"  Foundation, either version 3 of the License, or (at your option) any later\n"
		"  version. There is NO WARRANTY, to the extent permitted by law.\n", runStr);
}
