/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef REDUX_H
#define REDUX_H

#include "keyspace.h"
#include "hashfactory.h"

// RCrack made a mistake with chain lengths. The real chain length is actully one less.
// This code expects real chain lengths. If you use the classes for reading, writing, and
// searching rainbow tables they will give real chain lengths. The only time you need to
// give a different chain length for doPreWorkLinks() is when the table's end points are
// hashes instead of password indexes. In this case instead of giving doPreWorkLinks()
// the real chain length you give it one less than the real chain length. Then you call 
// indexToHash() to obtain the hash to look up in the table.
class Redux
{
public:
	virtual ~Redux();

	// For (step) generation and pre-work (with lots of hashes)
	virtual void doLinks(uint64_t *indexes, uint32_t chainPos, uint32_t tableIndex, uint32_t numLinks, uint32_t numIndexes) = 0;
	// For candidate generation "pre-work" (optimal numIndexes = 2 * getOptimalNumIndexesh())
	// firstChainPos is the smallest column number
	//   (ex with chainLength = 100 and numIndexes = 10, the first firstChainPos will be 0. This would do the pre computations
	//    for the hash assuming the password is the starting point (column 0) and the nine other columns after it 1 through 9.
	//    Then next firstChainPos would be 10, then 20, ..., 80, and 90. The last one firstChainPos = 90 has numIndexes = 9
	//    because the hash is converted into an index for look up and thus is not a link. Well unless the table format uses
	//    hashes as end points then the hash is left alone.)
	// Always start with firstChainPos = 0 since at the end if (chainLength - 1) / numIndexes / 2 is not an integer then you will
	// only waste at most getOptimalNumIndexesh() unused calculations vs chainLength * (getOptimalNumIndexesh() - 1) calculations.
	virtual void doPreWorkLinks(uint64_t *indexes, uint32_t firstChainPos, uint32_t tableIndex, uint32_t chainLength, uint32_t numIndexes, const uint32_t *hash) = 0;
	// For false alarm checking (optimal numIndexes = 2 * getOptimalNumIndexesh())
	// lastChainPoses needs to be inorder largest to smallest
	// Returns the index+1 into indexes of the found password index (indexes[returned_value - 1])
	// uint32_t ret = redux->doFalseAlarmLinks(indexes, ...);
	// if (ret)
	// {
	//     uint32_t *pw = new uint32_t[redux->getPwBlocks];
	//     redux->indexToPw(indexes[ret - 1], pw);
	//     // report found password "printf("\"%s\"\n", (char*) pw);"
	// }
	virtual uint32_t falseAlarmChecking(uint64_t *indexes, const uint32_t *hash, const uint32_t *lastChainPoses, uint32_t tableIndex, uint32_t numIndexes) = 0;
	// For table formats that end in a hash instead of a password index
	virtual void indexToHash(const uint64_t *indexes, uint32_t **hashes, uint32_t numIndexes) = 0;
	// For the last column of pre-work
	virtual uint64_t hashToIndex(const uint32_t *hash, uint32_t chainPos, uint32_t tableIndex) = 0;
	// For found password
	// Returns the password length
	virtual uint32_t indexToPw(uint64_t index, uint32_t *pw) = 0;

	const __inline__ uint32_t getHashLen()           { return m_hashLen;           }
	const __inline__ uint64_t getKeySpace()          { return m_keySpace;          }
	const __inline__ uint32_t getOptimalNumIndexes() { return m_optimalNumIndexes; }
	const __inline__ uint32_t getMaxPwLen()          { return m_maxPwLen;          }
	const __inline__ uint32_t getPwBlocks()          { return m_pwBlocks;          }

protected:
	Redux(const char *hashFunc, uint32_t maxPwLen);
	const __inline__ uint32_t isInited() { return m_hashFunc != NULL; }
	void cleanUp();
	void copy(const Redux &r);

	uint64_t m_keySpace;
	HASH_FUNC m_hashFunc;
	uint32_t m_hashLen, m_optimalNumIndexes, m_maxPwLen;
	uint32_t m_characterEncoding, m_pwBlocks, m_vecSize;
};

#endif
