/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "reduxbruteforce.h"
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <fstream>
#include <sstream>
#include <map>
#include <list>

using namespace std;

ReduxBruteForce::~ReduxBruteForce()
{
	cleanUp();
}

ReduxBruteForce::ReduxBruteForce(const keySpace &ks, const char *hashName)
	: Redux(hashName, getMaxLen(ks, HashFactory::getHashCharacterEncoding(hashName)))
{
	uint64_t keySpace, lastSubKeySpace = 0, keySpaceBoundary = 0;
	list<characterSet> uniqueCharSets;
	subKeySpace::iterator uniqueCharSetsIt;
	subKeySpace::const_reverse_iterator charSetItRev;
	subKeySpace::const_iterator charSetIt;
	characterSet::const_iterator charIt;
	keySpace::const_iterator curSubKeySpace;
	uint32_t a, b, charSetLen = 0;
	int32_t charSetDiffLen;

	m_keySpaceBoundaries = NULL;
	m_passwordLengths = NULL;
	m_divShortCut = NULL;
	m_charSetLengths = NULL;
	m_charSetLengthsFPM = NULL;
	m_numShortCutChars = NULL;
	m_charSet = NULL;
	m_subKeySpaces = NULL;
	if (isInited())
	{
		if (ks.size() == 0)
		{
			reportError("Invalid key space.");
			cleanUp();
		}
		else
		{
			m_numSubKeySpaces = ks.size();
			m_keySpaceBoundaries = new uint64_t   [m_numSubKeySpaces + 1];
			m_passwordLengths    = new uint32_t   [m_numSubKeySpaces];
			m_divShortCut        = new uint32_t   [m_numSubKeySpaces];
			m_numShortCutChars   = new int        [m_numSubKeySpaces];
			m_charSetLengths     = new uint32_t*  [m_numSubKeySpaces];
			m_charSetLengthsFPM  = new uint32_t*  [m_numSubKeySpaces];
			m_subKeySpaces       = new character**[m_numSubKeySpaces];

			m_keySpaceBoundaries[m_numSubKeySpaces] = 0;
			for (a = 0, curSubKeySpace = ks.begin(); curSubKeySpace != ks.end(); ++curSubKeySpace, a++)
			{
				uint32_t limboCharSetSize;
				keySpace = 1;
				m_passwordLengths[a] = curSubKeySpace->size();
				m_charSetLengths[a] = new uint32_t[curSubKeySpace->size()];
				m_charSetLengthsFPM[a] = new uint32_t[curSubKeySpace->size()];
				for (b = curSubKeySpace->size() - 1, charSetItRev = curSubKeySpace->rbegin(); charSetItRev != curSubKeySpace->rend(); ++charSetItRev, b--)
				{
					// Calculate key space
					keySpace *= charSetItRev->size();
					if (keySpace < UINT64_C(0x100000000))
					{
						m_divShortCut[a] = (uint32_t) keySpace;
						m_numShortCutChars[a] = curSubKeySpace->size() - b;
						limboCharSetSize = charSetItRev->size();
					}

					// Find unique character sets
					m_charSetLengths[a][b] = charSetItRev->size();
					m_charSetLengthsFPM[a][b] = (uint32_t) (UINT64_C(0x100000000) / charSetItRev->size());
					for (uniqueCharSetsIt = uniqueCharSets.begin(); uniqueCharSetsIt != uniqueCharSets.end(); ++uniqueCharSetsIt)
					{
						charSetDiffLen = m_charSetLengths[a][b] - uniqueCharSetsIt->size();
						if (charSetDiffLen > 0)
						{
							if (find(*charSetItRev, *uniqueCharSetsIt) != 0xffffffff)
							{
								*uniqueCharSetsIt = *charSetItRev;
								charSetLen += charSetDiffLen;
								break;
							}
						}
						else if (charSetDiffLen < 0)
						{
							if (find(*uniqueCharSetsIt, *charSetItRev) != 0xffffffff)
							{
								break;
							}
						}
						else if (*uniqueCharSetsIt == *charSetItRev)
						{
							break;
						}
					}
					if (uniqueCharSetsIt == uniqueCharSets.end())
					{
						uniqueCharSets.push_back(*charSetItRev);
						charSetLen += charSetItRev->size();
					}
				}
				if (keySpace == m_divShortCut[a])
				{
					m_numShortCutChars[a] = 0;
				}
				else if (keySpace / m_divShortCut[a] >= UINT64_C(0x100000000))
				{
					m_numShortCutChars[a] = -m_numShortCutChars[a];
				}
				else if ((curSubKeySpace->size() & 1) == 0 && (m_numShortCutChars[a]  & 1) == 1 && // limbo char set will make a difference
							(keySpace / m_divShortCut[a]) * limboCharSetSize < UINT64_C(0x100000000)) // limbo char set can be swapped
				{
					m_divShortCut[a] /= limboCharSetSize;
					m_numShortCutChars[a]--;
				}
				keySpaceBoundary += keySpace;
				m_keySpaceBoundaries[m_numSubKeySpaces - a - 1] = keySpaceBoundary;
				if (lastSubKeySpace > keySpace)
				{
					// I had this as a warning but it might get ignored. There is no excuse
					// for this since it makes it faster being in order with no down side.
					reportError("Key space is improperly formed (sub key spaces are not in ascending order).");
					cleanUp();
					return;
				}
				lastSubKeySpace = keySpace;
			}
			m_keySpace = keySpaceBoundary;

			m_charSetLength = charSetLen;
			m_charSet = new character[charSetLen];
			a = 0;
			for (uniqueCharSetsIt = uniqueCharSets.begin(); uniqueCharSetsIt != uniqueCharSets.end(); ++uniqueCharSetsIt)
			{
				uint32_t curCharSetLen = uniqueCharSetsIt->size();
				for (b = 0; b < curCharSetLen; a++, b++)
				{
					m_charSet[a] = (*uniqueCharSetsIt)[b];
				}
			}
			uniqueCharSets.clear();
			for (a = 0, curSubKeySpace = ks.begin(); curSubKeySpace != ks.end(); ++curSubKeySpace, a++)
			{
				m_subKeySpaces[a] = new character*[curSubKeySpace->size()];
				for (b = 0, charSetIt = curSubKeySpace->begin(); charSetIt != curSubKeySpace->end(); ++charSetIt, b++)
				{
					m_subKeySpaces[a][b] = strstr(m_charSet, charSetLen, *charSetIt);
				}
			}

			uint32_t invalidCharSet = 0;

			if ((m_characterEncoding & CE_UTF16_LE) != 0)
			{
				for (a = 0; a < charSetLen; a++)
				{
					switch (m_charSet[a].size)
					{
						case 1: // NULL pad
							m_charSet[a].size = 2;

						case 2: case 4:
							break;

						default:
							invalidCharSet = 1;
					}
				}
			}
			else if ((m_characterEncoding & CE_UTF16_BE) != 0)
			{
				for (a = 0; a < charSetLen; a++)
				{
					switch (m_charSet[a].size)
					{
						case 1: // NULL pad
#ifdef ARC_LITTLE_ENDIAN
							m_charSet[a].charData <<= 8;
#else
							m_charSet[a].charData >>= 8;
#endif
							m_charSet[a].size = 2;

						case 2: case 4:
							break;

						default:
							invalidCharSet = 1;
					}
				}
			}
			else if ((m_characterEncoding & CE_UTF32_LE) != 0)
			{
				for (a = 0; a < charSetLen; a++)
				{
					switch (m_charSet[a].size)
					{
						case 1: case 2: case 3: // NULL pad
							m_charSet[a].size = 4;

						case 4:
							break;

						default:
							invalidCharSet = 1;
					}
				}
			}
			else if ((m_characterEncoding & CE_UTF32_BE) != 0)
			{
				for (a = 0; a < charSetLen; a++)
				{
					switch (m_charSet[a].size)
					{
						case 1: case 2: case 3: // NULL pad
#ifdef ARC_LITTLE_ENDIAN
							m_charSet[a].charData <<= m_charSet[a].size << 3;
#else
							m_charSet[a].charData >>= m_charSet[a].size << 3;
#endif
							m_charSet[a].size = 4;

						case 4:
							break;

						default:
							invalidCharSet = 1;
					}
				}
			}
			if (invalidCharSet)
			{
				reportError("Invalid character set.");
				cleanUp();
			}
		}
	}
}

uint32_t ReduxBruteForce::indexToPw(uint64_t index, uint32_t *pw)
{
	character **subKeySpaces;
	uint64_t *keySpaceBoundaries;
	uint32_t *charSetLengths, *charSetLengthsFPM;
	uint32_t curSubKeySpace, charSetLength, quotient, r, q;
	int a, t, passwordLength, numShortCutChars;

	curSubKeySpace = 1;
	keySpaceBoundaries = m_keySpaceBoundaries;
	while (index < keySpaceBoundaries[curSubKeySpace])
	{
		curSubKeySpace++;
	}
	index             -= keySpaceBoundaries[curSubKeySpace];
	curSubKeySpace     = m_numSubKeySpaces - curSubKeySpace;
	passwordLength     = (int) m_passwordLengths[curSubKeySpace];
	numShortCutChars   = (int) m_numShortCutChars[curSubKeySpace];
	charSetLengths     = m_charSetLengths[curSubKeySpace];
	charSetLengthsFPM  = m_charSetLengthsFPM[curSubKeySpace];
	subKeySpaces       = m_subKeySpaces[curSubKeySpace];
	t = (int) m_pwBlocks;
	for (a = passwordLength >> 2; a < t; a++)
	{
		pw[a] = 0;
	}
	quotient = (uint32_t) index;
	a = passwordLength - 1;
	if ((m_characterEncoding & (CE_ASCII_BIN8 | CE_UTF8)) != 0)
	{
		// 8 bit
		if (numShortCutChars > 0)
		{
			uint32_t divisor, remainder;

			divisor = m_divShortCut[curSubKeySpace];
			DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

			numShortCutChars = passwordLength - numShortCutChars;
			for (; a >= numShortCutChars; a--)
			{
				charSetLength = charSetLengths[a];
				q = (uint32_t) (((uint64_t) remainder * charSetLengthsFPM[a]) >> 32);
				r = remainder - charSetLength * q;
				if (r >= charSetLength)
				{
					q++;
					r -= charSetLength;
				}
				remainder = q;
				((uint8_t*) pw)[a] = subKeySpaces[a][r].charData1[0];
			}
		}
		for (; a > 0; a--)
		{
			charSetLength = charSetLengths[a];
			q = (uint32_t) (((uint64_t) quotient * charSetLengthsFPM[a]) >> 32);
			r = quotient - charSetLength * q;
			if (r >= charSetLength)
			{
				q++;
				r -= charSetLength;
			}
			quotient = q;
			((uint8_t*) pw)[a] = subKeySpaces[a][r].charData1[0];
		}
		((uint8_t*) pw)[0] = subKeySpaces[0][quotient].charData1[0];
	}
	else if ((m_characterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
	{
		// 16 bit
		if (numShortCutChars > 0)
		{
			uint32_t divisor, remainder;

			divisor = m_divShortCut[curSubKeySpace];
			DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

			numShortCutChars = passwordLength - numShortCutChars;
			for (; a >= numShortCutChars; a--)
			{
				charSetLength = charSetLengths[a];
				q = (uint32_t) (((uint64_t) remainder * charSetLengthsFPM[a]) >> 32);
				r = remainder - charSetLength * q;
				if (r >= charSetLength)
				{
					q++;
					r -= charSetLength;
				}
				remainder = q;
				((uint16_t*) pw)[a] = subKeySpaces[a][r].charData2[0];
			}
		}
		for (; a > 0; a--)
		{
			charSetLength = charSetLengths[a];
			q = (uint32_t) (((uint64_t) quotient * charSetLengthsFPM[a]) >> 32);
			r = quotient - charSetLength * q;
			if (r >= charSetLength)
			{
				q++;
				r -= charSetLength;
			}
			quotient = q;
			((uint16_t*) pw)[a] = subKeySpaces[a][r].charData2[0];
		}
		((uint16_t*) pw)[0] = subKeySpaces[0][quotient].charData2[0];
	}
	else //if ((m_characterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
	{
		// 32 bit
		if (numShortCutChars > 0)
		{
			uint32_t divisor, remainder;

			divisor = m_divShortCut[curSubKeySpace];
			DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

			numShortCutChars = passwordLength - numShortCutChars;
			for (; a >= numShortCutChars; a--)
			{
				charSetLength = charSetLengths[a];
				q = (uint32_t) (((uint64_t) remainder * charSetLengthsFPM[a]) >> 32);
				r = remainder - charSetLength * q;
				if (r >= charSetLength)
				{
					q++;
					r -= charSetLength;
				}
				remainder = q;
				((uint32_t*) pw)[a] = subKeySpaces[a][r].charData;
			}
		}
		for (; a > 0; a--)
		{
			charSetLength = charSetLengths[a];
			q = (uint32_t) (((uint64_t) quotient * charSetLengthsFPM[a]) >> 32);
			r = quotient - charSetLength * q;
			if (r >= charSetLength)
			{
				q++;
				r -= charSetLength;
			}
			quotient = q;
			((uint32_t*) pw)[a] = subKeySpaces[a][r].charData;
		}
		((uint32_t*) pw)[0] = subKeySpaces[0][quotient].charData;
	}
	return (uint32_t) passwordLength;
}

void ReduxBruteForce::cleanUp()
{
	if (m_charSet != NULL)
	{
		for (uint32_t a = 0; a < m_numSubKeySpaces; a++)
		{
			delete [] m_subKeySpaces[a];
			delete [] m_charSetLengths[a];
			delete [] m_charSetLengthsFPM[a];
		}
		delete [] m_keySpaceBoundaries;
		delete [] m_passwordLengths;
		delete [] m_divShortCut;
		delete [] m_numShortCutChars;
		delete [] m_charSetLengths;
		delete [] m_charSetLengthsFPM;
		delete [] m_charSet;
		delete [] m_subKeySpaces;
		m_charSet = NULL;
	}

	Redux::cleanUp();
}

void ReduxBruteForce::copy(const ReduxBruteForce &r)
{
	Redux::copy(r);

	m_keySpaceBoundaries = NULL;
	m_passwordLengths = NULL;
	m_divShortCut = NULL;
	m_charSetLengths = NULL;
	m_charSetLengthsFPM = NULL;
	m_numShortCutChars = NULL;
	m_charSet = NULL;
	m_subKeySpaces = NULL;
	if (r.m_hashFunc != NULL)
	{
		m_numSubKeySpaces = r.m_numSubKeySpaces;
		m_charSetLength   = r.m_charSetLength;

		m_keySpaceBoundaries = new uint64_t   [m_numSubKeySpaces + 1];
		m_passwordLengths    = new uint32_t   [m_numSubKeySpaces];
		m_divShortCut        = new uint32_t   [m_numSubKeySpaces];
		m_numShortCutChars   = new int        [m_numSubKeySpaces];
		m_charSetLengths     = new uint32_t*  [m_numSubKeySpaces];
		m_charSetLengthsFPM  = new uint32_t*  [m_numSubKeySpaces];
		m_subKeySpaces       = new character**[m_numSubKeySpaces];
		m_charSet            = new character  [m_charSetLength];
		memcpy(m_keySpaceBoundaries, r.m_keySpaceBoundaries, sizeof(uint64_t)  * (m_numSubKeySpaces + 1));
		memcpy(m_passwordLengths,    r.m_passwordLengths,    sizeof(uint32_t)  * m_numSubKeySpaces);
		memcpy(m_divShortCut,        r.m_divShortCut,        sizeof(uint32_t)  * m_numSubKeySpaces);
		memcpy(m_numShortCutChars,   r.m_numShortCutChars,   sizeof(uint32_t)  * m_numSubKeySpaces);
		memcpy(m_charSet,            r.m_charSet,            sizeof(character) * m_charSetLength);

		for (uint32_t a = 0; a < m_numSubKeySpaces; a++)
		{
			m_charSetLengths[a]    = new uint32_t[m_passwordLengths[a]];
			m_charSetLengthsFPM[a] = new uint32_t[m_passwordLengths[a]];
			memcpy(m_charSetLengths[a],    r.m_charSetLengths[a],    sizeof(uint32_t) * m_passwordLengths[a]);
			memcpy(m_charSetLengthsFPM[a], r.m_charSetLengthsFPM[a], sizeof(uint32_t) * m_passwordLengths[a]);
			m_subKeySpaces[a] = new character*[m_passwordLengths[a]];
			for (uint32_t b = 0; b < m_passwordLengths[a]; b++)
			{
				m_subKeySpaces[a][b] = m_charSet + (r.m_subKeySpaces[a][b] - r.m_charSet);
			}
		}
	}
}
