/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "reduxfactory.h"
#include "hashfactory.h"

using namespace std;

list<reduxInfo*> *ReduxFactory::s_reduxes = NULL;

Redux* ReduxFactory::getReduxInstance(const char *name, const keySpace &ks, int gpu, const char *hashName)
{
	if (s_reduxes != NULL)
	{
		uint32_t charSizes = getCharSizes(ks);

		if (charSizes != 0xffffffff) // Valid character set
		{
			list<reduxInfo*>::iterator it;
			string sname = name;
			uint32_t charEncoding = HashFactory::getHashCharacterEncoding(hashName);

			// Check for fixed character width key spaces
			if (((charEncoding & (CE_ASCII_BIN8 | CE_UTF8))     != 0 && charSizes <= 1 << 1) || // 1 byte/character
			    ((charEncoding & (CE_UTF16_BE   | CE_UTF16_LE)) != 0 && charSizes <= 1 << 2) || // 2 bytes/character
			    ( charEncoding & (CE_UTF32_BE   | CE_UTF32_LE)) != 0) // 4 bytes/character
			{
				charEncoding |= CE_FIX_WIDTH_CHAR;
				for (it = s_reduxes->begin(); it != s_reduxes->end(); ++it)
				{
					if ((*it)->name == sname &&
					    ((gpu != 0) == ((*it)->createReduxGpu != NULL)) &&
					    (charEncoding & (*it)->charEncodings) == charEncoding &&
					    ((*it)->charEncodings & CE_FIX_WIDTH_CHAR) != 0)
					{
						if (gpu)
						{
							return (*it)->createReduxGpu(ks, hashName, gpu - 1);
						}
						return (*it)->createRedux(ks, hashName);
					}
				}
				charEncoding ^= CE_FIX_WIDTH_CHAR;
			}

			// Check for varying character width key spaces
			for (it = s_reduxes->begin(); it != s_reduxes->end(); ++it)
			{
				if ((*it)->name == sname &&
				    ((gpu != 0) == ((*it)->createReduxGpu != NULL)) &&
				    (charEncoding & (*it)->charEncodings) == charEncoding)
				{
					if (gpu)
					{
						return (*it)->createReduxGpu(ks, hashName, gpu - 1);
					}
					return (*it)->createRedux(ks, hashName);
				}
			}
		}
	}
	return NULL;
}

uint32_t ReduxFactory::registerRedux(const char *name, const char *comment, uint32_t charEncodings, CREATE_REDUX createRedux, CREATE_REDUX_GPU createReduxGpu)
{
	reduxInfo *ri = new reduxInfo;

	ri->createRedux    = createRedux;
	ri->createReduxGpu = createReduxGpu;
	ri->name           = name;
	ri->comment        = comment;
	ri->charEncodings  = charEncodings;
	if (s_reduxes == NULL)
	{
		s_reduxes = new list<reduxInfo*>;
	}
	s_reduxes->push_back(ri);
	return 1;
}

void ReduxFactory::printReduxes()
{
	list<reduxInfo*>::iterator it;
	char *spaces;
	uint32_t maxLen = 0;

	if (s_reduxes != NULL)
	{
		// Find max name length
		for (it = s_reduxes->begin(); it != s_reduxes->end(); ++it)
		{
			if (maxLen < (*it)->name.length())
				maxLen = (*it)->name.length();
		}

		// Fill spaces with max name length number of spaces ' '
		spaces = new char[maxLen + 1];
		for (uint32_t a = 0; a < maxLen; a++)
		{
			spaces[a] = ' ';
		}
		spaces[maxLen] = 0;

		// Print
		for (it = s_reduxes->begin(); it != s_reduxes->end(); ++it)
		{
			printf(" * %s:%s %s (%s)\n",
				(*it)->name.c_str(), spaces + (maxLen - (*it)->name.length()),
				(*it)->comment.c_str(), (*it)->createReduxGpu != NULL ? "GPU" : "CPU");
		}
		delete [] spaces;
	}
	printf("\n");
}
