/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "reduxsinglebyte.h"
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <fstream>
#include <sstream>
#include <map>
#include <list>

using namespace std;

ReduxSingleByte::~ReduxSingleByte()
{
	cleanUp();
}

ReduxSingleByte::ReduxSingleByte(const keySpace &ks, const char *hashName)
	: ReduxBruteForce(ks, hashName)
{
	m_charSetSingleByte = NULL;
	m_subKeySpacesSingleByte = NULL;
	m_charSetCacheLengths = NULL;
	m_charSetCache = NULL;
	m_subKeySpacesCache = NULL;
	m_charSetCacheLengthsFPM = NULL;
	if (isInited())
	{
		// **** Not Cached ****
		m_subKeySpacesSingleByte = new uint8_t** [m_numSubKeySpaces];
		m_charSetSingleByte      = new uint8_t   [m_charSetLength];

		for (uint32_t a = 0; a < m_numSubKeySpaces; a++)
		{
			m_subKeySpacesSingleByte[a] = new uint8_t* [m_passwordLengths[a]];
			for (uint32_t b = 0; b < m_passwordLengths[a]; b++)
			{
				m_subKeySpacesSingleByte[a][b] = m_charSetSingleByte + (m_subKeySpaces[a][b] - m_charSet);
			}
		}

		for (uint32_t a = 0; a < m_charSetLength; a++)
		{
			m_charSetSingleByte[a] = m_charSet[a].charData1[0];
			if (m_charSet[a].size != 1)
			{
				// This should never happen if you use ReduxFactory
				reportError("Invalid character set.");
				cleanUp();
				ReduxBruteForce::cleanUp();
				return;
			}
		}

		// **** Cache ****
		struct charSetCache
		{
			uint8_t *a, *b;
			uint32_t aLen, bLen;
			uint32_t offset, size;
		};
		list<charSetCache> uniqueCharSetsCache;
		list<charSetCache>::iterator curUniqueCharSetsCache;
		charSetCache curCharSetCache;

		m_charSetCacheLengths    = new uint32_t* [m_numSubKeySpaces];
		m_subKeySpacesCache      = new uint16_t**[m_numSubKeySpaces];
		m_charSetCacheLengthsFPM = new uint32_t* [m_numSubKeySpaces];

		// Find unique cache
		uint32_t charSetCacheSize = 0;
		for (uint32_t a = 0; a < m_numSubKeySpaces; a++)
		{
			int skipChar = m_passwordLengths[a] - m_numShortCutChars[a] - 1;

			if (m_numShortCutChars[a] < 0)
			{
				skipChar = m_passwordLengths[a] + m_numShortCutChars[a] - 1;
			}
			for (int b = m_passwordLengths[a] - 2; b >= 0; b -= 2)
			{
				if (b == skipChar)
				{
					// With a 4, 5 short cut split one character will not be cached
					// ** **|* ** **
					// cc cc|* cc cc
					//     ^-- skipChar
					b--;
				}
				curCharSetCache.a    = m_subKeySpacesSingleByte[a][b];
				curCharSetCache.b    = m_subKeySpacesSingleByte[a][b + 1];
				curCharSetCache.aLen = m_charSetLengths[a][b];
				curCharSetCache.bLen = m_charSetLengths[a][b + 1];
				for (curUniqueCharSetsCache = uniqueCharSetsCache.begin(); curUniqueCharSetsCache != uniqueCharSetsCache.end(); ++curUniqueCharSetsCache)
				{
					if (curUniqueCharSetsCache->a    == curCharSetCache.a &&
					    curUniqueCharSetsCache->b    == curCharSetCache.b &&
					    curUniqueCharSetsCache->aLen == curCharSetCache.aLen &&
					    curUniqueCharSetsCache->bLen == curCharSetCache.bLen)
					{
						break;
					}
				}
				if (curUniqueCharSetsCache == uniqueCharSetsCache.end())
				{
					curCharSetCache.offset = charSetCacheSize;
					curCharSetCache.size = curCharSetCache.aLen * curCharSetCache.bLen;
					charSetCacheSize += curCharSetCache.size;
					uniqueCharSetsCache.push_back(curCharSetCache);
				}
			}
		}

		// Create cache
		m_charSetCacheLength = charSetCacheSize;
		m_charSetCache = new uint16_t[charSetCacheSize];

		uint16_t *charSetCache = m_charSetCache;

		for (curUniqueCharSetsCache = uniqueCharSetsCache.begin(); curUniqueCharSetsCache != uniqueCharSetsCache.end(); ++curUniqueCharSetsCache)
		{
			curCharSetCache = *curUniqueCharSetsCache;
			for (uint32_t a = 0; a < curCharSetCache.aLen; a++)
			{
				for (uint32_t b = 0; b < curCharSetCache.bLen; b++)
				{
					((uint8_t*) charSetCache)[0] = curCharSetCache.a[a];
					((uint8_t*) charSetCache)[1] = curCharSetCache.b[b];
					charSetCache++;
				}
			}
		}

		// Point to cache
		for (uint32_t a = 0; a < m_numSubKeySpaces; a++)
		{
			m_charSetCacheLengths[a]    = new uint32_t [m_passwordLengths[a]];
			m_subKeySpacesCache[a]      = new uint16_t*[m_passwordLengths[a]];
			m_charSetCacheLengthsFPM[a] = new uint32_t [m_passwordLengths[a]];

			int skipChar = m_passwordLengths[a] - m_numShortCutChars[a] - 1;

			if (m_numShortCutChars[a] < 0)
			{
				skipChar = m_passwordLengths[a] + m_numShortCutChars[a] - 1;
			}
			int b;
			for (b = m_passwordLengths[a] - 2; b >= 0; b -= 2)
			{
				m_subKeySpacesCache     [a][b + 1] = NULL;
				m_charSetCacheLengths   [a][b + 1] = 0;
				m_charSetCacheLengthsFPM[a][b + 1] = 0;
				if (b == skipChar)
				{
					// With a 4, 5 short cut split one character will not be cached
					// ** **|* ** **
					// cc cc|* cc cc
					//     ^-- skipChar
					m_subKeySpacesCache     [a][b] = NULL;
					m_charSetCacheLengths   [a][b] = 0;
					m_charSetCacheLengthsFPM[a][b] = 0;
					b--;
				}
				curCharSetCache.a    = m_subKeySpacesSingleByte[a][b];
				curCharSetCache.b    = m_subKeySpacesSingleByte[a][b + 1];
				curCharSetCache.aLen = m_charSetLengths[a][b];
				curCharSetCache.bLen = m_charSetLengths[a][b + 1];
				for (curUniqueCharSetsCache = uniqueCharSetsCache.begin(); curUniqueCharSetsCache != uniqueCharSetsCache.end(); ++curUniqueCharSetsCache)
				{
					if (curUniqueCharSetsCache->a    == curCharSetCache.a &&
					    curUniqueCharSetsCache->b    == curCharSetCache.b &&
					    curUniqueCharSetsCache->aLen == curCharSetCache.aLen &&
					    curUniqueCharSetsCache->bLen == curCharSetCache.bLen)
					{
						m_subKeySpacesCache     [a][b] = m_charSetCache + curUniqueCharSetsCache->offset;
						m_charSetCacheLengths   [a][b] = curUniqueCharSetsCache->size;
						m_charSetCacheLengthsFPM[a][b] = (uint32_t) (UINT64_C(0x100000000) / curUniqueCharSetsCache->size);
						break;
					}
				}
			}
			if (b == -1)
			{
				m_subKeySpacesCache     [a][0] = NULL;
				m_charSetCacheLengths   [a][0] = 0;
				m_charSetCacheLengthsFPM[a][0] = 0;
			}
		}
	}
}

void ReduxSingleByte::indexToHash(const uint64_t *indexes, uint32_t **hashes, uint32_t numIndexes)
{
	vector4 pw[MAX_PW_SIZE];
	vector4 hash[MAX_HASH_SIZE];
	uint32_t pwTmp[MAX_PW_BLOCKS];
	uint64_t keySpace                    = m_keySpace;
	uint64_t *keySpaceBoundaries         = m_keySpaceBoundaries;
	uint8_t **subKeySpaces;
	uint8_t ***subKeySpacesSingleByte    = m_subKeySpacesSingleByte;
	uint32_t *charSetCacheLengths;
	uint32_t *charSetCacheLengthsFPM;
	uint32_t *passwordLengths            = m_passwordLengths;
	uint16_t **subKeySpacesCache;
	uint16_t ***subKeySpacesCacheArr     = m_subKeySpacesCache;
	int *numShortCutCharsArr             = m_numShortCutChars;
	uint32_t **charSetCacheLengthsArr    = m_charSetCacheLengths;
	uint32_t **charSetCacheLengthsFPMArr = m_charSetCacheLengthsFPM;
	uint32_t *divShortCut                = m_divShortCut;
	HASH_FUNC hashFunc                   = m_hashFunc;
	uint32_t numSubKeySpaces             = m_numSubKeySpaces;
	uint32_t charSetLength;
	int a;
	int passwordLength;
	int numShortCutChars;
	uint32_t pwBlocks                    = m_pwBlocks;
	uint32_t vecSize                     = 4 * m_vecSize;
	uint32_t hashBlocks                  = (m_hashLen + 3) / 4;
	uint32_t padding                     = 0;
	uint32_t count                       = m_optimalNumIndexes;

	if ((m_characterEncoding & CE_80_NULL_PADDED) != 0)
	{
		padding = 0x80;
	}
	while (numIndexes != 0)
	{
		if (numIndexes < count)
		{
			count = numIndexes;
		}
		numIndexes -= count;

		// index to pw
		for (uint32_t i = 0; i < count; i++)
		{
			uint64_t index = indexes[i];
			uint32_t curSubKeySpace = 1;

			while (index < keySpaceBoundaries[curSubKeySpace])
			{
				curSubKeySpace++;
			}
			index                 -= keySpaceBoundaries[curSubKeySpace];
			curSubKeySpace         = numSubKeySpaces - curSubKeySpace;
			passwordLength         = (int) passwordLengths[curSubKeySpace];
			numShortCutChars       = numShortCutCharsArr[curSubKeySpace];
			subKeySpaces           = subKeySpacesSingleByte[curSubKeySpace];
			charSetCacheLengths    = charSetCacheLengthsArr[curSubKeySpace];
			charSetCacheLengthsFPM = charSetCacheLengthsFPMArr[curSubKeySpace];
			subKeySpacesCache      = subKeySpacesCacheArr[curSubKeySpace];
			for (a = passwordLength >> 2; a < (int) pwBlocks; a++)
			{
				pwTmp[a] = 0;
			}
			uint32_t quotient = (uint32_t) index;
			a = passwordLength - 2;
			if (numShortCutChars > 0)
			{
				uint32_t remainder, divisor;

				divisor = divShortCut[curSubKeySpace];
				DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

				numShortCutChars = passwordLength - numShortCutChars;
				for (; a >= numShortCutChars; a -= 2)
				{
					uint32_t r, q;

					charSetLength = charSetCacheLengths[a];
					q = (uint32_t) (((uint64_t) remainder * charSetCacheLengthsFPM[a]) >> 32);
					r = remainder - charSetLength * q;
					if (r >= charSetLength)
					{
						q++;
						r -= charSetLength;
					}
					remainder = q;
					*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
				}
				if (a == numShortCutChars - 1)
				{
					a--;
					((uint8_t*) pwTmp)[numShortCutChars] = subKeySpaces[numShortCutChars][remainder];
				}
			}
			for (; a >= 0; a -= 2)
			{
				uint32_t r, q;

				charSetLength = charSetCacheLengths[a];
				q = (uint32_t) (((uint64_t) quotient * charSetCacheLengthsFPM[a]) >> 32);
				r = quotient - charSetLength * q;
				if (r >= charSetLength)
				{
					q++;
					r -= charSetLength;
				}
				quotient = q;
				*((uint16_t*) (((uint8_t*) pwTmp) + a)) = subKeySpacesCache[a][r];
			}
			if (a == -1)
			{
				((uint8_t*) pwTmp)[0] = subKeySpaces[0][quotient];
			}
			((uint8_t*) pwTmp)[passwordLength] = padding;

			((uint32_t*) pw)[i] = passwordLength;
			for (uint32_t b = 0, c = i + vecSize; b < pwBlocks; b++, c += vecSize)
			{
				((uint32_t*) pw)[c] = pwTmp[b];
			}
		}

		// pw to hash
		hashFunc(pw, hash);

		// Copy hashes
		for (uint32_t i = 0; i < count; i++)
		{
			for (uint32_t j = 0; j < hashBlocks; j++)
			{
				hashes[i][j] = ((uint32_t*) hash)[j * vecSize + i];
			}
		}

		indexes += count;
		hashes  += count;
	}
}

uint32_t ReduxSingleByte::indexToPw(uint64_t index, uint32_t *pw)
{
	uint8_t **subKeySpaces;
	uint64_t *keySpaceBoundaries;
	uint32_t *charSetCacheLengths;
	uint32_t *charSetCacheLengthsFPM;
	uint16_t **subKeySpacesCache;
	uint32_t curSubKeySpace;
	uint32_t charSetLength;
	uint32_t quotient;
	int a;
	int passwordLength;
	int numShortCutChars;

	curSubKeySpace = 1;
	keySpaceBoundaries = m_keySpaceBoundaries;
	while (index < keySpaceBoundaries[curSubKeySpace])
	{
		curSubKeySpace++;
	}
	index                 -= keySpaceBoundaries[curSubKeySpace];
	curSubKeySpace         = m_numSubKeySpaces - curSubKeySpace;
	passwordLength         = (int) m_passwordLengths[curSubKeySpace];
	numShortCutChars       = (int) m_numShortCutChars[curSubKeySpace];
	subKeySpaces           = m_subKeySpacesSingleByte[curSubKeySpace];
	charSetCacheLengths    = m_charSetCacheLengths[curSubKeySpace];
	charSetCacheLengthsFPM = m_charSetCacheLengthsFPM[curSubKeySpace];
	subKeySpacesCache      = m_subKeySpacesCache[curSubKeySpace];
	for (a = passwordLength >> 2; a < (int) m_pwBlocks; a++)
	{
		pw[a] = 0;
	}
	quotient = (uint32_t) index;
	a = passwordLength - 2;
	if (numShortCutChars > 0)
	{
		uint32_t remainder, divisor;

		divisor = m_divShortCut[curSubKeySpace];
		DIV_MOD_64_RET_32_32(index, divisor, remainder, quotient);

		numShortCutChars = passwordLength - numShortCutChars;
		for (; a >= numShortCutChars; a -= 2)
		{
			uint32_t r, q;

			charSetLength = charSetCacheLengths[a];
			q = (uint32_t) (((uint64_t) remainder * charSetCacheLengthsFPM[a]) >> 32);
			r = remainder - charSetLength * q;
			if (r >= charSetLength)
			{
				q++;
				r -= charSetLength;
			}
			remainder = q;
			*((uint16_t*) (((uint8_t*) pw) + a)) = subKeySpacesCache[a][r];
		}
		if (a == numShortCutChars - 1)
		{
			a--;
			((uint8_t*) pw)[numShortCutChars] = subKeySpaces[numShortCutChars][remainder];
		}
	}
	for (; a >= 0; a -= 2)
	{
		uint32_t r, q;

		charSetLength = charSetCacheLengths[a];
		q = (uint32_t) (((uint64_t) quotient * charSetCacheLengthsFPM[a]) >> 32);
		r = quotient - charSetLength * q;
		if (r >= charSetLength)
		{
			q++;
			r -= charSetLength;
		}
		quotient = q;
		*((uint16_t*) (((uint8_t*) pw) + a)) = subKeySpacesCache[a][r];
	}
	if (a == -1)
	{
		((uint8_t*) pw)[0] = subKeySpaces[0][quotient];
	}
	return (uint32_t) passwordLength;
}

void ReduxSingleByte::cleanUp()
{
	if (m_charSetSingleByte != NULL)
	{
		for (uint32_t a = 0; a < m_numSubKeySpaces; a++)
		{
			if (m_charSetCacheLengths != NULL)
			{
				delete [] m_charSetCacheLengths[a];
				delete [] m_subKeySpacesCache[a];
			}
			delete [] m_subKeySpacesSingleByte[a];
		}
		if (m_charSetCacheLengths != NULL)
		{
			delete [] m_charSetCacheLengths;
			delete [] m_subKeySpacesCache;
			delete [] m_charSetCacheLengthsFPM;
			delete [] m_charSetCache;
		}
		delete [] m_subKeySpacesSingleByte;
		delete [] m_charSetSingleByte;
		m_charSetSingleByte = NULL;
	}

	// Don't do this because of multiple virtual inheritance (although it don't really matter here)
	// ReduxBruteForce::cleanUp();
}

void ReduxSingleByte::copy(const ReduxSingleByte &r)
{
	// Don't do this because of multiple virtual inheritance
	// ReduxBruteForce::copy(r);

	m_subKeySpacesSingleByte = NULL;
	m_charSetSingleByte = NULL;
	m_charSetCacheLengths = NULL;
	m_charSetCache = NULL;
	m_subKeySpacesCache = NULL;
	if (r.m_hashFunc != NULL)
	{
		m_subKeySpacesSingleByte = new uint8_t**[m_numSubKeySpaces];
		m_charSetSingleByte      = new uint8_t  [m_charSetLength];
		m_charSetCacheLengths = new uint32_t* [m_numSubKeySpaces];
		m_subKeySpacesCache   = new uint16_t**[m_numSubKeySpaces];
		m_charSetCache        = new uint16_t  [m_charSetCacheLength];
		memcpy(m_charSetSingleByte, r.m_charSetSingleByte, sizeof(uint8_t) * m_charSetLength);
		memcpy(m_charSetCache, r.m_charSetCache, sizeof(uint16_t) * m_charSetCacheLength);

		for (uint32_t a = 0; a < m_numSubKeySpaces; a++)
		{
			m_subKeySpacesSingleByte[a] = new uint8_t*[m_passwordLengths[a]];
			m_charSetCacheLengths[a] = new uint32_t [m_passwordLengths[a] / 2];
			m_subKeySpacesCache[a]   = new uint16_t*[m_passwordLengths[a] / 2];
			memcpy(m_charSetCacheLengths[a], r.m_charSetCacheLengths[a], sizeof(uint32_t) * (m_passwordLengths[a] / 2));
			for (uint32_t b = 0; b < m_passwordLengths[a]; b++)
			{
				m_subKeySpacesSingleByte[a][b] = m_charSetSingleByte + (r.m_subKeySpacesSingleByte[a][b] - r.m_charSetSingleByte);
				m_subKeySpacesCache[a][b] = m_charSetCache + (r.m_subKeySpacesCache[a][b] - r.m_charSetCache);
			}
		}
	}
}
