/*
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This file is part of a set of programs called Advanced TMTO.

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "writert.h"
#include "writefactory.h"
#include "keyspace.h"
#include <iostream>
#include <iomanip>
#include <sstream>
#include <stdio.h>

#define CHAINS_PER_FILE 67108864

using namespace std;

uint32_t WriteRT::s_inited = WriteRT::initFactory();

uint32_t WriteRT::initFactory()
{
	uint32_t ret;

	ret =  WriteFactory::registerWrite("rt", "Basic RT format", WriteRT::create);
	return ret;
}

WriteRT::~WriteRT()
{
	cleanUp();
}

WriteRT::WriteRT(const char *folder, const char *fileNameComment, const char *reduxName, const char *hashName, const char *salt, const char *keySpaceName, uint32_t chainLength, uint32_t tableIndex)
	: Write(reduxName, hashName, salt, chainLength, tableIndex)
{
	uint32_t keySpaceNameLen = 0;
	keySpace ks;
	stringstream fileName(stringstream::out);

	m_fileChains = 0;
	m_fileIndex = 0;

	fileName << folder;
	if (folder[strlen(folder) - 1] != DIRECTORY_SEPERATOR_CHAR)
	{
		fileName << DIRECTORY_SEPERATOR_CHAR;
	}
	fileName << hashName << '_';

	// 0 - full file name, 1 - just the key space part, 2 - key space file
	uint32_t keySpaceNameFormat = createKeySpace(ks, keySpaceName);
	if (ks.size() == 0)
	{
		cleanUp();
		return;
	}
	if (keySpaceNameFormat == 0)
	{
		while (*keySpaceName != '_')
		{
			keySpaceName++;
		}
		keySpaceName++;
		while (keySpaceName[keySpaceNameLen] == '_' && keySpaceName[keySpaceNameLen] != 0)
		{
			keySpaceNameLen++;
		}
		fileName.write(keySpaceName, keySpaceNameLen);
	}
	else if (keySpaceNameFormat == 1)
	{
		while (keySpaceName[keySpaceNameLen] != '_' && keySpaceName[keySpaceNameLen] != 0)
		{
			keySpaceNameLen++;
		}
		fileName.write(keySpaceName, keySpaceNameLen);
	}
	else
	{
		fileName << "hybrid3(" << keySpaceName << ")#0-0";
	}
	if (keySpaceNameFormat != 2 &&
	    strncmp(keySpaceName, "hybrid", 6) == 0 &&
	    ( keySpaceName[6] == '(' ||
	     (keySpaceName[6] == '2' && keySpaceName[7] == '(') ||
	     (keySpaceName[6] == '3' && keySpaceName[7] == '(')) &&
	    keySpaceName[keySpaceNameLen - 1] == ')')
	{
		fileName << "#0-0";
	}
	fileName << '_' << tableIndex << '_' << (chainLength + 1) << 'x';
	m_fileName = fileName.str();
	m_fileName2 = fileNameComment;
	m_fileName2 = '_' + m_fileName2 + '_';
	fileName << CHAINS_PER_FILE << m_fileName2 << setfill('0') << setw(3) << m_fileIndex << setw(0) << ".rt";
	m_fout.open(fileName.str(), ios_base::binary | ios_base::trunc);
	if (m_fout.fail())
	{
		reportError("Failed to open file for output '%s'.", fileName.str().c_str());
		exit(1);
	}
}

Write *WriteRT::create(const char *folder, const char *fileNameComment, const char *reduxName, const char *hashName, const char *salt, const char *keySpaceName, uint32_t chainLength, uint32_t tableIndex)
{
	Write *ret;

	ret = new WriteRT(folder, fileNameComment, reduxName, hashName, salt, keySpaceName, chainLength, tableIndex);
	if (!((WriteRT*) ret)->isInited())
	{
		delete ret;
		ret = NULL;
	}
	return ret;
}

uint32_t WriteRT::write(rtChain *chains, uint32_t count)
{
	uint32_t count2 = 0;

	m_fileChains += count;
	if (m_fileChains > CHAINS_PER_FILE)
	{
		m_fileChains -= count;
		count2 = count;
		count = (uint32_t) (CHAINS_PER_FILE - m_fileChains);
	}
	if (count > 0)
	{
		m_fout.write((char*) chains, 16 * count);
		if (m_fout.bad())
		{
			reportError("Failed to write data.");
			exit(1);
		}
	}
	if (count2 > 0)
	{
		nextFile();
		m_fout.write((char*) (chains + count), 16 * count2);
		if (m_fout.bad())
		{
			reportError("Failed to write data.");
			exit(1);
		}
	}
	return count + count2;
}

uint32_t WriteRT::write(crtChain *chains, uint32_t count)
{
	uint32_t count2 = 0;

	m_fileChains += count;
	if (m_fileChains > CHAINS_PER_FILE)
	{
		m_fileChains -= count;
		count2 = count;
		count = (uint32_t) (CHAINS_PER_FILE - m_fileChains);
	}
	if (count > 0)
	{
		for (uint32_t i = 0; i < count; i++)
		{
			m_fout.write((char*) (chains + i), 16);
			if (m_fout.bad())
			{
				reportError("Failed to write data.");
				exit(1);
			}
		}
	}
	if (count2 > 0)
	{
		nextFile();
		chains += count;
		for (uint32_t i = 0; i < count2; i++)
		{
			m_fout.write((char*) (chains + i), 16);
			if (m_fout.bad())
			{
				reportError("Failed to write data.");
				exit(1);
			}
		}
	}
	return count + count2;
}

void WriteRT::done()
{
	closeFile();
}

void WriteRT::cleanUp()
{
	closeFile();
	Write::cleanUp();
}

void WriteRT::nextFile()
{
	stringstream fileName(stringstream::out);

	closeFile();
	m_fileChains = 0;
	m_fileIndex++;
	fileName << m_fileName << CHAINS_PER_FILE << m_fileName2 << setfill('0') << setw(3) << m_fileIndex << setw(0) << ".rt";
	m_fout.open(fileName.str(), ios_base::binary | ios_base::trunc);
	if (m_fout.fail())
	{
		reportError("Failed to open file for output '%s'.", fileName.str().c_str());
		exit(1);
	}
}

void WriteRT::closeFile()
{
	if (m_fout.is_open())
	{
		m_fout.close();
		if (m_fileChains != CHAINS_PER_FILE)
		{
			stringstream fileName(stringstream::out);
			string oldName, newName;

			fileName << m_fileName << CHAINS_PER_FILE << m_fileName2 << setfill('0') << setw(3) << m_fileIndex << setw(0) << ".rt";
			oldName = fileName.str();
			fileName.str("");

			fileName << m_fileName << m_fileChains << m_fileName2 << setfill('0') << setw(3) << m_fileIndex << setw(0) << ".rt";
			newName = fileName.str();
			if (rename(oldName.c_str(), newName.c_str()) != 0)
			{
				perror("Error renaming file");
				reportError("Renameing '%s' to '%s'.", oldName.c_str(), newName.c_str());
				exit(1);
			}
		}
	}
}
